{--------------------------------------------------------------}
{                                                              }
{  DsMiniMm.inc - mini memory manager for DLLs                 }
{                                                              }
{  Author:     Semi                                            }
{  Started:    131001 , previous DataSys.pas 2002-2013         }
{                                                              }
{                                                              }
{                                                              }
{--------------------------------------------------------------}

{$ifdef UNIT}
{$undef UNIT}
unit DsMiniMm;
interface uses Windows; {$define DSMINIMM_END}
{$endif UNIT}

//!!!TODO_200830 - yet more check FreePageSched and make ListEntry list inside the pages
//                 scheduled for release - there is usually free space near page-start...

//? {$i M:\Units\DsTypes.inc}

//{$define MINIMM_DBG}
{$define MINIMM_PAGEALIGNED} // enable using page-aligned 4kb buffers (is slower, but less wasting memory, and allows allocating page-aligned buffers optimally. is available for g_MainMmPool only and cannot be used accross DLL boundaries...)

//-------------------------------------------------------------
{$ifndef DSMINIMM_HDR}
{$define DSMINIMM_HDR}

{$i DsIMemPool.inc}  //New171113 - IMemPool interface, TIMemPoolVmt

type
  TMmPtrAlloc=function(Arg: Pointer; Size: integer): Pointer;
  
  PMmVirtPager=^TMmVirtPager;
  TMmVirtPager=record
    AllocRawPage:  function(Arg: Pointer; Size: integer): Pointer register;
    ReleaseRawPage:function(Arg: Pointer; Page: Pointer; Size: integer): Boolean register;
    Arg:           Pointer;
    ReInit:        procedure(Pager: PMmVirtPager; AllocProc: TMmPtrAlloc; AllocArg: Pointer; lpLockRtl: Pointer) register; // called when allocated a new pool, so that VirtPool can re-allocate its "Arg" from the new Pool..
    //ReAllocRawPage:function(Arg: Pointer; var Page: Pointer; OldSize,NewCapacity,NewCommit: integer): Boolean register;
    Shutdown:      function(Arg,AllocArg: Pointer): Boolean register;
  end;

type
  PMmPool=^TMmPool;
    PMmType=^TMmType;
      PMmPage=^TMmPage;
  
  TMmPage=record case Byte of
  0: (
    Signature:     Cardinal; // = cMmPageSign
    NextFreePage:  Pointer;
    Typ:           PMmType;
    ItemSize:      Byte; // shl 4...
    FreeCount:     Byte;
    FreeList:      Byte;
    MaxFreeCount:  Byte; // detect completelly empty pages...
    //
    //New200831 - most pages have free space here anyway, just type 1 was using this place... now it wastes one more tile... :
    FreeLink:      TListEntry; // pages scheduled for removal are linked here...
    SmallPageLink: TListEntry; //New200907 - all small pages list...
    //
  );
  1: (
    Bytes:         array[0..4095] of Byte;
  );
  end;
  
  TCheckAlignedProc=function(P: Pointer): Pointer; // return NIL if not aligned, can use ECX, preserve all other registers...
  
  TMmType=record
    Signature:     Cardinal;
    FreePages:     Pointer;
    CheckAligned:  TCheckAlignedProc;
    Pool:          PMmPool;
    //
    ItemSize:      Byte; // shl 4
    DefFreeCount:  Byte;
    DefFreeList:   Byte;
    MaxFreeCount:  Byte;
  end;
  
  PMmLargePage=^TMmLargePage;
  TMmLargePage=record //!!!TODO: consider allocating LargeBlocks at page-start...
    Signature:     Cardinal; // = cMmLargePageSign
    Size:          Longint;
    Capacity:      Longint;
    Pool:          PMmPool;
    Data:          record end;
  end;
  
  PMmMidType=^TMmMidType;
  TMmMidType=record
    OffsNext:      Longint;
    OffsPrior:     Longint;
    MinSize:       Longint;
    dwPad:         Longint;
  end;
  PMmMidTypeTable=^TMmMidTypeTable;
  TMmMidTypeTable=array[0..7] of TMmMidType; // count is variable...
  
  TMmPool=record    // (TMmPool)
    Signature:     Cardinal;
    VirtPool:      TMmVirtPager;
    SmallTypes:    array[0..16] of PMmType;
    MidLock:       TRtlCriticalSection;
    AllocProc:     Pointer; // either MmAlloc or MmAllocStartup
    MidPage:       Pointer;
    MidTypes:      PMmMidTypeTable;
    MidTypeCount:  integer;
    MidTypes_MinSize:integer;
    MidMax_MinSize:integer;
    MidMax_OffsNext:Longint;
    MidMax_OffsPrior:Longint;
    AlignedList:   Pointer;
    AlignedCount:  integer;
    Owner:         Pointer;
    //x200822: bnCollectNeeded:Boolean; bPad:          array[1..3] of Byte;
    iCollectNeeded:integer; //New200822
    //
    Vmt:           Pointer; //New171113 IMemPool vmt...
    RefCount:      integer; //New171113 IMemPool (interface) RefCount
    //
    FreeSchedCount:integer; //New200822
    FreePageList:  TListEntry; //New200831 - list of small pages scheduled for removal
    SmallPageList: TListEntry; //New200907 - list of all small pages
    //x200831: FreePageSched: Pointer; //New200822
    hCollectThread:THandle; //New200822
    idCollectThread:Cardinal; //New200822
    //
    Vmt_IMalloc:   Pointer; //New200825 - IMalloc vmt...
  end;
  
  PMmMiddlePrefix=^TMmMiddlePrefix;
  TMmMiddlePrefix=packed record case Byte of
  2: (
    PriorSizeFlags:Word;
    NextSizeFlags: Word;
    dwTag:         DWORD; // for 8-byte alignment...
    Data:          record end;
  );
  1: (
    PriorSizeLo:   Byte;
    PriorFlags:    Byte;
    NextSizeLo:    Byte;
    NextFlags:     Byte;
  );
  end;
  
  PMmMiddlePageHdr=^TMmMiddlePageHdr;
  TMmMiddlePageHdr=record
    Zero:          Word;
    NextSizeFlags: Word;
    Signature:     Cardinal; // = cMmMiddlePageSign
  end;
  
  PMmMiddlePageTail=^TMmMiddlePageTail;
  TMmMiddlePageTail=record
    PriorSizeFlags:Word;
    Zero:          Word;
    Pool:          PMmPool;
  end;
  
  PMmMiddleFree=^TMmMiddleFree;
  TMmMiddleFree=record
    dwPad:         DWORD;
    // RelListEntry:
    OffsNext:      Longint;
    OffsPrior:     Longint;
    //
    Signature:     Cardinal; // = cMmMiddleFreeSign
  end;
  
  PMmByteArray=^TMmByteArray; //New191121
  TMmByteArray=array[0..4095] of Byte;

const
  cMmPageSign=$A96750B5;
  cMmTypeSign=$A970DEB5;
  cMmMiddlePageSign=$A970B1B5;
  cMmLargePageSign=$A950BBB5;
  cMmPoolSign=$A970AAB5;

  cMmMiddleFreeSign=$0000AB01;
  cMmCurAllocSign=$0000BB03;
  
  cMmMaxFreeSmallPages = 32; // maximum count of small free pages before releasing them...

function MmAlloc(Pool: PMmPool; Size: integer): Pointer; register; forward;
function MmFree(Pool: PMmPool; Item: Pointer): Boolean; register; forward;
function MmQuerySizeFast(Item: Pointer): integer; register; forward;
function MmQuerySizeCheck(Item: Pointer): integer; register; forward;
function MmGetItemPool(Item: Pointer): PMmPool; register; forward;
function MmReAlloc(Pool: PMmPool; var P: Pointer; NewSize: integer): Boolean; register; forward;
function MmAllocWithTag(Pool: PMmPool; Size: integer; Tag: Cardinal): Pointer; register; forward;

function MmIMemPool(Pool: PMmPool; out IMemPool): Boolean; register; forward;

  function Mm_IMemPool_Vmt(): Pointer; forward; //New171113
  //
  // Middle-only IMemPool - conserves memory, is little slower...
  // (replace Pool.Vmt with this and use QueryInterface to initialize - before any small alloc...)
  function Mm_Middle_IMemPool_Vmt(): Pointer; forward; //New171114

var
  g_MainMmPool: PMmPool;
  bnMmWasInit: Boolean; //New200822

// TMemoryManager handlers (for g_MainMmPool)
function MmGetMem(Size: integer): Pointer; register; forward;
function MmFreeMem(Ptr: Pointer): integer; register; forward;
function MmReallocMem(P: Pointer; Size: integer): Pointer; register; forward;
function MmGetMemoryManager(var Mmgr: TMemoryManager; lpVirtPool: PMmVirtPager=nil): Boolean; forward;
procedure MmShutdownMainPool; forward;
  procedure MmGetDefVirtPager(VirtPool: PMmVirtPager); forward;

function MmCreatePool(VirtPool: PMmVirtPager; Owner: Pointer): PMmPool; forward;
function MmDestroyPool(Pool: PMmPool; Owner: Pointer): Boolean; forward;

{$endif !DSMINIMM_HDR}
{$ifndef INTERFACE}{$ifndef DSMINIMM_INC}{$define DSMINIMM_INC}
//-------------------------------------------------------------
{$ifdef DSMINIMM_END}
implementation
{$endif DSMINIMM_END}

type
  PMmFreeItem=^TMmFreeItem;
  TMmFreeItem=record
    dwPad0:        DWORD;
    dwPad1:        DWORD;
    bPad2:         Byte;
    NextFree:      Byte;
    //x200822: bSign:         Byte; bPad2b:        Byte;
    wSign:         Word; // same $00AB (AB in third byte of third dword)
    dwPad3:        DWORD;
  end;

const
  cMmFreeSign=$00AB;
  cMmFreeSign_DW=cMmFreeSign shl 16;
  cMmNextFreeMask_DW=$0000FF00;
  cMmMaxSmallSize=256;

function MmAcquirePage(Pool: PMmPool; Typ: PMmType): Pointer; register; forward;
function MmOutOfMemoryError(Size: integer; Typ: PMmType): Pointer; register; forward;
function MmAllocMiddle(Pool: PMmPool; Size: integer): Pointer; register; forward;
function MmMidPageAlloc(var Space: Pointer; Size: integer): Pointer; register; forward;
function MmVerifySmallPage(BadItem: Pointer): Pointer; register; forward;

procedure MmInitializeListHead(Base: PListEntry); assembler; forward;
function MmInsertHeadList(Base,Entry: PListEntry): Pointer; assembler; forward;
function MmInsertTailList(Base,Entry: PListEntry): Pointer; assembler; forward;
function MmRemoveEntryListAndClear(Entry: PListEntry): Pointer; assembler; forward;

{$ifdef MINIMM_DBG}
function DbgBreak2(Param: Pointer): Pointer;
begin
  if (DebugHook<>0) then begin
    try
      asm
        db $CC
      end;
      if (Param<>nil) then ;;;
    except ; end;
  end;
  Result:=Param;
end;

function DbgBreak(Param: Pointer): Pointer;
asm
  push edx
  push ecx
  call DbgBreak2
  pop  ecx
  pop  edx
end;

function MmDbg_CheckSmallPage(Item: Pointer): Pointer;
asm //                        EAX           ->EAX
  mov  edx,eax
  and  edx,$FFFFF000
  cmp  [edx].TMmPage.Signature,cMmPageSign
  jne  @damaged
  ret
@damaged:
  call DbgBreak
end;

function MmDbg_CheckMiddlePage(Item: Pointer): Pointer;
asm //                         EAX           ->EAX
  push eax
  push edx
  push ecx
  mov  edx,eax //New240220
  and  edx,$FFFFF000 //zmena240220: eax->edx, there was error on [eax-8] below !
  cmp  [edx].TMmMiddlePageHdr.Signature,cMmMiddlePageSign
  jne  @damaged_page
  movzx edx,[eax-8].TMmMiddlePrefix.NextSizeFlags
  mov  ecx,edx
  and  edx,$1FF
  shl  edx,3
  cmp  [eax+edx].TMmMiddlePrefix.PriorSizeFlags,cx
  jne  @damaged_item
  //
  nop
  jmp  @out
  //
@damaged_page:
  call DbgBreak
  jmp  @out
  //
@damaged_item:
  call DbgBreak
  nop
  //jmp  @out
@out:
  pop  ecx
  pop  edx
  pop  eax
end;
{$endif MINIMM_DBG}

function MmAlloc(Pool: PMmPool; Size: integer): Pointer; assembler;
asm //           EAX            EDX           ->EAX
  test edx,edx
  jle  @fail
  //x:add  edx,$F
  //x:and  edx,not $F
  cmp  edx,cMmMaxSmallSize
  //x:mov  ecx,edx
  ja   MmAllocMiddle//@middle
  //
  add edx,$F
  //xd1: and edx,not $F
  //xd1: mov ecx,edx
  //
  // get the "type"
  shr  edx,4 //xd1: shr  ecx,4
  mov  ecx,dword ptr [eax].TMmPool.SmallTypes[edx*4] //xd1: mov  ecx,dword ptr [eax].TMmPool.SmallTypes[ecx*4]
  // get the free page:
@retry_type:
  mov  edx,ecx
  mov  ecx,[ecx].TMmType.FreePages
  test ecx,ecx
  jz   @need_page
  //New240215 - if there is pointer to MtType, it's not a free page !
  cmp  ecx,edx
  je   @need_page
  //
  // alloc block:
@retry_page:
  movzx eax,[ecx].TMmPage.FreeList
  test eax,eax
  jz   @no_free
  //
  mov  edx,eax
  shl  edx,4
  add  edx,ecx
  {?xd2 + see also below!
  push edx
  mov  dl,[edx].TMmFreeItem.NextFree
  lock cmpxchg [ecx].TMmPage.FreeList,dl
  jne  @thread_problem
  //
  //mov  dl,$FF
  //xadd [ecx].TMmPage.FreeCount,dl
  dec  [ecx].TMmPage.FreeCount
  //
  pop  eax
  {}//new:xd2
  mov  ah,[edx].TMmFreeItem.NextFree
  lock cmpxchg [ecx].TMmPage.FreeList,ah
  jne  @thread_problem
  dec  [ecx].TMmPage.FreeCount
  mov  eax,edx
  {}
  //
  cmp  [eax].TMmFreeItem.wSign,cMmFreeSign //x200822: cmp  [eax].TMmFreeItem.bSign,cMmFreeSign
  jne  @check_free
  xor  edx,edx
  mov  [eax+8],edx
  {$ifdef MINIMM_DBG}
  call MmDbg_CheckSmallPage
  {$endif MINIMM_DBG}
  ret
  //
@thread_problem:
  db $F3; nop
  //xd2: add  esp,4
  jmp  @retry_page
  //
  //zmena240218 - simpler & safer:
@no_free:
    push  edi
    mov   edx,[ecx].TMmPage.Typ
    mov   edi,[ecx].TMmPage.NextFreePage
    xor   eax,eax
    cmp   edi,edx
    db 0Fh,44h,F8h ;//cmovz edi,eax
    mov   eax,ecx
    lock cmpxchg [edx].TMmType.mt_FreePages,edi
    jne   .pd_restart
    xor   eax,eax
    mov   [ecx].TMmPage.NextFreePage,eax
    pop  edi
    mov  ecx,edx
    jmp short .retry_type
  .pd_restart:
    pop  edi
    mov  ecx,edx
    jmp short .retry_type
    //
          ;////x240218 - unsafe!
          ;//@no_free:
          ;//  // no free items in this page - get another page:  ECX=Page
          ;//  mov  eax,[ecx].TMmPage.NextFreePage
          ;//  xor  edx,edx
          ;//  lock cmpxchg [ecx].TMmPage.NextFreePage,edx
          ;//  jne  @unlink_failed
          ;//  //
          ;//  cmp  eax,[ecx].TMmPage.Typ //x200906b: test eax,eax -- test order changed, more probable first...
          ;//  je   @no_next_page
          ;//  test eax,eax //x200906b: cmp  eax,[ecx].TMmPage.Typ
          ;//  jz   @no_next_page
          ;//  //
          ;//  xchg eax,ecx
          ;//  mov  edx,[ecx].TMmPage.Typ
          ;//  lock cmpxchg [edx].TMmType.FreePages,ecx
          ;//  je   @retry_page
          ;//  //
          ;//@relink_failed:
          ;//  // another thread linked another page to Typ.FreePages:
          ;//  push ebx
          ;//  mov  ebx,[edx].TMmType.FreePages
          ;//  test ebx,ebx
          ;//  jz   @relink_noNext
          ;//  xor  eax,eax
          ;//  lock cmpxchg [ecx].TMmPage.NextFreePage,ebx
          ;//  jne  @relink2_failed_skip
          ;//@relink1b:
          ;//  mov  eax,ebx
          ;//  lock cmpxchg [edx].TMmType.FreePages,ecx
          ;//  jne  @relink2_failed_retry
          ;//@relink2_failed_skip:
          ;//  pop  ebx
          ;//  mov  ecx,[ecx].TMmPage.Typ
          ;//  jmp  @retry_type
          ;//@relink2_failed_retry:
          ;//  db $F3; nop
          ;//  pop  ebx
          ;//  jmp  @relink_failed
          ;//  //
          ;//@relink_noNext:
          ;//  xor  eax,eax
          ;//  lock cmpxchg [ecx].TMmPage.NextFreePage,edx
          ;//  jne  @relink2_failed_skip
          ;//  //
          ;//  mov  eax,ebx
          ;//  jmp  @relink1b
          ;//  //
          ;//@unlink_failed:
          ;//  mov  ecx,[ecx].TMmPage.Typ
          ;//  jmp  @retry_type
  //
@no_next_page:
  mov  edx,[ecx].TMmPage.Typ
  //New200906b - also in this case unlink full page from Typ.FreePages:
  mov  eax,ecx
  xor  ecx,ecx
  lock cmpxchg [edx].TMmType.FreePages,ecx
  //
@need_page:
  // need a new free page:  EDX=Type, ECX=0
  push edx
  mov  eax,[edx].TMmType.Pool
  call MmAcquirePage
  pop  ecx
  test eax,eax
  jnz  @retry_type
  // Out of memory?
  xor  eax,eax
  mov  edx,ecx
  call MmOutOfMemoryError
  ret
  //
@check_free:
  // item is either not free, or its free-sign was damaged... Item=EAX
  call MmVerifySmallPage
  test eax,eax
  jz   @fail
  and  eax,$FFFFF000
  mov  ecx,[eax].TMmPage.Typ
  jmp  @retry_type
  //
@fail:
  xor  eax,eax
  ret
  //
@middle:
  jmp MmAllocMiddle
end;

const
  //cDefMinSizes: array[0..15] of Word=(256,272,288,320,352,384,416,448,496,544,592,640,704,768,832,896);
  cDefMinSizes: array[0..24] of Word=(16,64,96,128,256,272,288,320,352,384,416,448,496,544,592,640,704,768,832,896,960,1024,1152,1280,1536);

procedure MmInitMidTypes(Pool: PMmPool);
var Count,i: integer;
    MidTypes: PMmMidTypeTable;
    MTyp: PMmMidType;
begin
  Count:=Length(cDefMinSizes);
  MidTypes:=MmMidPageAlloc(Pool.MidPage,Count*SizeOf(TMmMidType)+16);
  // prevent freeing:
  PLongint(Longint(MidTypes)-8)^:=-1;
  //x?: MidTypes:=Pointer((Longint(MidTypes)+15) and (not 15)); // 16-byte aligned...
  Pool.MidTypes:=MidTypes;
  Pool.MidTypeCount:=Count;
  MTyp:=@MidTypes[0];
  i:=0;
  Pool.MidTypes_MinSize:=cDefMinSizes[0];
  while (i<Count) do begin
    MTyp.OffsNext:=0;
    MTyp.OffsPrior:=0;
    MTyp.MinSize:=cDefMinSizes[i];
    MTyp.dwPad:=0;
    inc(i);
    inc(Longint(MTyp),SizeOf(TMmMidType));
  end;
  dec(Longint(MTyp),SizeOf(TMmMidType));
  Pool.MidMax_MinSize:=MTyp.MinSize+64;
end;

function MmInternalVerifySmallPage(Page: PMmPage): Pointer;
var Offs,fc,ItemSize,NextOffs: integer;
    NewIdx,PrevIdx: Byte;
    Where: PByte;
Label _Retry;
begin
  if (Page.Signature=cMmPageSign) then begin
    ItemSize:=Page.ItemSize shl 4; //x240212: Page.ItemSize shl 3;
    if (ItemSize>0) then begin
      Page.MaxFreeCount:=(4096 div ItemSize)-1;
      //New240212
      if(ItemSize<32) then
        dec(Page.MaxFreeCount);
      //
    _Retry:
      fc:=0;
      Where:=@Page.FreeList;
      Offs:=Page.FreeList shl 4;
      while (Offs<>0) do begin
        if (Page.Bytes[Offs+10]<>cMmFreeSign)
        or (Page.Bytes[Offs+11]<>0) //New200822
        then begin
          // Bad-item suspected...
          NextOffs:=Page.Bytes[Offs+9] shl 4;
          if (NextOffs<>0)
          and (Page.Bytes[NextOffs+10]=cMmFreeSign)
          and (Page.Bytes[NextOffs+11]=0) //New200822
          then begin
            NewIdx:=NextOffs shr 4;
          end else begin
            NewIdx:=0;
          end;
          PrevIdx:=Offs shr 4;
          if (Where^=PrevIdx) then
            Where^:=NewIdx;
          Offs:=NewIdx shl 4;
          continue;
        end;
        inc(fc);
        if (fc>=255) then begin
          // Circular free-list...
          if (Where^<>0)
          and (PByte(Longint(Where)+1)^=cMmFreeSign)
          then begin
            Where^:=0;
            goto _Retry;
          end;
          break;
        end;
        Where:=@Page.Bytes[Offs+9];
        Offs:=integer(Page.Bytes[Offs+9]) shl 4; //x171113: Offs:=Page.Bytes[Offs+9];
      end;
      Page.FreeCount:=fc;
      Result:=Page;
      exit;
    end;
  end;
  Result:=nil;
end;

function MmRemoveTopSmallPage(Page: PMmPage; Typ: PMmType): Boolean; assembler;
asm //                        EAX            EDX          ->AL
  mov  ecx,[eax].TMmPage.NextFreePage
  lock cmpxchg [edx].TMmType.FreePages,ecx
  //x240215: setz al
  //New240215
  jnz  @fail
  and  [eax].TMmPage.NextFreePage,0
  mov  al,1
  jmp  @out  ;// =ret
@fail:
  xor  eax,eax
@out:
end;

function MmAddTopSmallPage(Page: PMmPage; Typ: PMmType): Boolean; assembler;
asm //                     EAX            EDX          ->AL
  push ebx
  mov  ebx,eax
  mov  ecx,[edx].TMmType.FreePages
  xor  eax,eax
  lock cmpxchg [ebx].TMmPage.NextFreePage,ecx
  jnz  @fail
  mov  eax,ecx
  lock cmpxchg [edx].TMmType.FreePages,ebx
  jnz  @fail
  mov  eax,1
  pop  ebx
  ret
@fail:
  xor  eax,eax
  pop  ebx
  //ret
end;

function MmVerifySmallPage(BadItem: Pointer): Pointer; register;
var Page: PMmPage;
    Pool: PMmPool;
    Typ: PMmType;
begin
  {$ifdef MINIMM_DBG}
  DbgBreak(BadItem);
  MmDbg_CheckSmallPage(BadItem);
  {$endif MINIMM_DBG}
  Page:=Pointer(Cardinal(BadItem) and $FFFFF000);
  if (Page.Signature=cMmPageSign) then begin
    Typ:=Page.Typ;
    if (Typ<>nil)
    and (Typ.Signature=cMmTypeSign)
    then begin
      Pool:=Typ.Pool;
      if (Pool<>nil)
      and (Pool.Signature=cMmPoolSign)
      then begin
        //
        // get the page from top-list, so that other thread does not allocate there:
        if not MmRemoveTopSmallPage(Page,Typ) then begin
          Result:=Page;
          exit;
        end;
        if IsMultiThread then
          Sleep(0);
        //
        // Prevent multiple threads repairing the page:
        EnterCriticalSection(Pool.MidLock);
        try
          Result:=MmInternalVerifySmallPage(Page);
        finally
          LeaveCriticalSection(Pool.MidLock);
        end;
        //New240215
        if (Result=nil) then begin
          Result:=MmAcquirePage(Pool,Typ);
          exit;
        end;
        //
        if (Result<>nil)
        and (Page.FreeCount>4) // otherwise will be added when FreeCount gets =4 ...
        then begin
          // Return page to free-list:
          MmAddTopSmallPage(Page,Typ);
        end;
        //
        exit;
      end;
    end;
  end;
  Result:=nil;
end;

const
  cMmMaxMiddleSize=4096-256-16;

procedure MmLinkMiddleFree(Pool: PMmPool; FreeItem: Pointer); register;
asm //                     EAX            EDX
  push ebx
  push edi
  push esi
  lea  edx,[edx].TMmMiddleFree.OffsNext
  push edx // [esp]=^FreeItem.OffsNext
  //
  cmp  [eax].TMmPool.Signature,cMmPoolSign
  jne  @bad_pool
  mov  esi,eax
  // get size:
  movsx edx,[edx-8-offset TMmMiddleFree.OffsNext].TMmMiddlePrefix.NextSizeFlags
  test edx,edx
  jnl  @not_free
  and  edx,$1FF
  shl  edx,3
  //
  // need at least 16 bytes to link:
  cmp  edx,[eax].TMmPool.MidTypes_MinSize//16
  jb   @too_small
  //
  cmp  edx,[eax].TMmPool.MidMax_MinSize
  jae  @MidMax0
  //
  // find slot:
  mov  ecx,[eax].TMmPool.MidTypeCount
  mov  ebx,[eax].TMmPool.MidTypes
  //mov  esi,eax
  xor  edi,edi // l:=0
  dec  ecx     // h:=MidTypeCount-1
  // while (l<=h)
  cmp  edi,ecx
  mov  eax,edi
  jnl  @before_first//@break
@cy:
  lea  eax,[edi+ecx] // i:=(l+h) shr 1
  and  eax,not 1 // here eax=2*i
  cmp  [ebx+8*eax].TMmMidType.MinSize,edx
  jb   @not_before // if MidTypes[i].MinSize<ThisSize then go not_before
@not_after:
  // here MidTypes[i].MinSize>=ThisSize
  je   @exact // for =MinSize, found exact
  // h:=i-1;
  shr  eax,1
  dec  eax
  mov  ecx,eax
  // while (l<=h)
  cmp  edi,ecx
  jle  @cy
  //x: after this slot:
  //x: lea  eax,[edi+1]
  //x: in this slot:
  mov  eax,edi
  // before this slot:
  sub  eax,1
  jmp  @found
@not_before:
  shr  eax,1
  // l:=i+1;
  inc  eax
  mov  edi,eax
  // while (l<=h)
  cmp  edi,ecx
  jle  @cy
  //x: at this slot:
  // before this slot:
  sub  eax,1 //dec  eax
  jmp  @found
@exact:
  shr  eax,1
//@break:
@found:
  jl   @before_first
  cmp  eax,[esi].TMmPool.MidTypeCount
  mov  edx,[esp] // ^FreeItem.OffsNext
  jae  @MidMax
  // Link to this slot:
  add  eax,eax
  lea  eax,[ebx+8*eax]
@rel_link: // EAX=@Slot,EDX=@This
  mov  ecx,[eax] // OffsNext
  lea  edi,[eax+ecx] // Next (possibly =Slot)...
  neg  edi
  add  edi,edx // This-Next
  mov  ebx,eax
  sub  ebx,edx // Slot-This
  mov  [eax+ecx+4],edi // Next.Prior:=This-Next;
  neg  edi
  mov  [edx+4],ebx // This.Prior:=Slot-This;
  neg  ebx
  mov  [edx],edi // This.Next:=Next-This;
  mov  [eax],ebx // Slot.Next:=This-Slot;
  mov  dword ptr [edx+8],cMmMiddleFreeSign
  jmp  @out
  //
@MidMax0:
  cmp  edx,4096-16
  jae  @free_page
@MidMax:
  mov  edx,[esp]
  // link to Pool.MidMax_OffsNext,MidMax_OffsPrior instead...
  lea  eax,[esi].TMmPool.MidMax_OffsNext
  jmp  @rel_link
  //
@free_page:
  // whole page is empty...
  jne  @not_exact
  mov  edx,[esp]
  sub  edx,offset TMmMiddleFree.OffsNext
  //and  edx,$FFF
  //cmp  edx,8
  //jne  @MidMax
  test edx,$FF7
  jnz  @MidMax
  //
  mov  edx,[esp]
  and  edx,$FFFFF000
  mov  ecx,$1000
  mov  ebx,eax
  mov  eax,[eax].TMmPool.VirtPool.Arg
  call [ebx].TMmPool.VirtPool.ReleaseRawPage
  //
@before_first:
@bad_pool:
@not_free:
@too_small:
@not_exact:
@out:
  add  esp,4
  pop  esi
  pop  edi
  pop  ebx
  //ret
end;

function MmMidPageAlloc(var Space: Pointer; Size: integer): Pointer; assembler;
asm //                     [EAX]            EDX           ->EAX
  // edx already 8-byte round, max-size checked...
  // pool already locked if appropriate...
  push ebx
  push esi
  mov  ebx,eax
  // "better safe than sorry" - check Size aligned...
  xor  esi,esi
  add  edx,7
  and  edx,not 7
  jle  @size_invalid
  //mov  eax,[eax]
  xor  eax,eax
  xchg [ebx],eax //lock xchg [ebx],eax  -- pool already locked, no need for "lock" here...
  shr  edx,3
  //
  test eax,$FFF
  jz   @none
  test eax,7
  jnz  @bad
  movsx ecx,[eax-8].TMmMiddlePrefix.NextSizeFlags
  test ecx,ecx
  jnl  @not_free
  and  ecx,$1FF
  cmp  ecx,edx
  jne  @not_exact
@alloc_exact:
  shl  ecx,3
  // item fills exactly this space:
  cmp  [eax+ecx].TMmMiddlePrefix.PriorSizeLo,dl
  jne  @damaged
  and  [eax-8].TMmMiddlePrefix.NextFlags,$0F
  and  [eax+ecx].TMmMiddlePrefix.PriorFlags,$0F
  xor  edx,edx
  test eax,$FF7
  jz   @skip_clear_tag1
  mov  [eax-8].TMmMiddlePrefix.dwTag,edx
@skip_clear_tag1:
  {$ifdef MINIMM_DBG}
  call MmDbg_CheckMiddlePage
  {$endif MINIMM_DBG}
  pop  esi
  pop  ebx
  ret
  //
@not_exact:
  jb   @no_space
  // there is some rest:
  mov  esi,ecx
  sub  esi,edx
  dec  esi
  cmp  esi,5//x: cMmMaxSmallSize/8 //32 // = 256/8
  jb   @too_few_to_retain
  or   esi,$8000
  shl  ecx,3
  // patch this prefix:
  mov  [eax-8].TMmMiddlePrefix.NextSizeFlags,dx
  // make middle prefix:
  mov  [eax+8*edx].TMmMiddlePrefix.PriorSizeFlags,dx
  mov  [eax+8*edx].TMmMiddlePrefix.NextSizeFlags,si
  // patch following prefix:
  mov  [eax+ecx].TMmMiddlePrefix.PriorSizeFlags,si
  // clear dwTag:
  xor  ecx,ecx
  test eax,$FF7 // must NOT clear Page.Signature (with EAX and $FFF=8) !
  jz   @skip_clear_tag2
  mov  [eax-8].TMmMiddlePrefix.dwTag,ecx
@skip_clear_tag2:
  //
  // return rest of space:
  mov  esi,eax
  lea  edx,[eax+8*edx+8]
  mov  [edx].TMmMiddleFree.Signature,cMmCurAllocSign
  xor  eax,eax
  cmpxchg [ebx],edx //lock cmpxchg [ebx],edx -- pool is locked, no need for "lock" here...
  jne  @thread_problem
  //
@size_invalid:
@done_ok:
  mov  eax,esi
  //
  {$ifdef MINIMM_DBG}
  call MmDbg_CheckMiddlePage
  {$endif MINIMM_DBG}
  pop  esi
  pop  ebx
  ret
  //
@too_few_to_retain:
  mov  edx,ecx
  jmp  @alloc_exact
  //
@thread_problem:
  // cannot return rest of space to Pool.MidPage (other thread assigned some value)
  // must link the rest space into free-list...
  mov  [edx].TMmMiddleFree.Signature,ecx  // clear sign...
  mov  eax,ebx
  sub  eax,offset TMmPool.MidPage
  call MmLinkMiddleFree
  jmp  @done_ok
  //
@no_space:
  // return MidPage:
  mov  edx,eax
  xor  eax,eax
  cmpxchg [ebx],edx //lock cmpxchg [ebx],edx -- thread is locked, no need for "lock" here...
  //
@damaged:
@bad:
@not_free:
@none:
  xor  eax,eax
  pop  esi
  pop  ebx
end;

const
  cMmMidPageTailOffset=$FF8;

function MmFormatMidPage(Page: Pointer; Pool: PMmPool): Pointer;
asm //                   EAX            EDX           ->EAX
  test eax,$7FFF0000
  jz   @bad_page
  cmp  dword ptr [eax],0
  jnz  @bad_page
  test eax,$FFF
  jnz  @not_aligned
  //
  mov  dword ptr [eax],(((4096-16)/8) or $8000) shl 16 // Zero:=0; NextSizeFlags:=(4096-16)/8 or $8000;
  mov  dword ptr [eax+4],cMmMiddlePageSign
  mov  dword ptr [eax+cMmMidPageTailOffset].TMmMiddlePageTail.Pool,edx
  mov  dword ptr [eax+cMmMidPageTailOffset].TMmMiddlePageTail.PriorSizeFlags,((4096-16)/8) or $8000
  //
  // return start of free space:
  add  eax,8
  //
  ret
  //
@bad_page:
@not_aligned:
  xor  eax,eax
end;

function MmInternalAllocMiddle(Pool: PMmPool; Size: integer): Pointer; assembler;
asm //                         EAX            EDX           ->EAX
  add  edx,7
  and  edx,not 7
  //
  push ebx
  push edi
  push esi
  push edx // [esp]=Size
  mov  ebx,eax
  // From FreeList:
  cmp  edx,[eax].TMmPool.MidMax_MinSize
  jae  @cur_Alloc
  //
  mov  ecx,[eax].TMmPool.MidTypeCount
  mov  eax,[eax].TMmPool.MidTypes // eax=@MidTypes[0]
  // find suitable MidType:
  xor  esi,esi // esi=l:=0
  dec  ecx     // ecx=h:=MidTypeCount-1
  // while (l<=h)
  cmp  esi,ecx
  jnl  @cur_Alloc
@mt_cy:
  // i:=(l+h) shr 1;
  lea  edi,[esi+ecx]
  and  edi,not 1
  // if MidTypes[i].MinSize<Size then go not_before
  cmp  [eax+8*edi].TMmMidType.MinSize,edx
  jb   @not_before
  // if =Size then go found
  je   @found_2x
  // here MidTypes[i].MinSize>Size
  // h:=i-1
  shr  edi,1
  dec  edi
  mov  ecx,edi
  // while (l<=h)
  cmp  esi,ecx
  jle  @mt_cy
  // use MidTypes[i]
  inc  edi
  jmp  @found
@not_before:
  // l:=i+1;
  shr  edi,1
  inc  edi
  mov  esi,edi
  // while (l<=h)
  cmp  esi,ecx
  jle  @mt_cy
  // here use i+1
@found:
  cmp  edi,[ebx].TMmPool.MidTypeCount
  jae  @cur_alloc
  add  edi,edi
  lea  edi,[eax+8*edi]
  cmp  [edi].TMmMidType.OffsNext,0
  jz   @cur_Alloc
  //
  // alloc from this free-list slot:
  // get first item:
  mov  eax,[edi] // This:=Slot+Slot.OffsNext;
  add  eax,edi
  cmp  dword ptr [eax+8],cMmMiddleFreeSign
  jne  @damaged
  // unlink:
  mov  edx,[eax] // OffsNext
  add  edx,eax // Next:=This+OffsNext
  mov  ecx,edi
  sub  ecx,edx // Slot-Next
  mov  [edx+4],ecx // Next.Prior:=Slot-Next;
  neg  ecx
  mov  [edi],ecx // Slot.Next:=Next-Slot;
  //
  xor  ecx,ecx
  mov  [eax+8],ecx // MUST clear FreeSign...
  mov  [eax],ecx; mov [eax+4],ecx // clear OffsNext&OffsPrior too...
  // alloc:
  sub  eax,offset TMmMiddleFree.OffsNext
  movsx ecx,[eax-8].TMmMiddlePrefix.NextSizeFlags
  test ecx,ecx
  jnl  @not_free
  and  ecx,$1FF
  shl  ecx,3
  cmp  ecx,[esp]
  jb   @not_large_enough
  and  [eax-8].TMmMiddlePrefix.NextFlags,$7F
  and  [eax+ecx].TMmMiddlePrefix.PriorFlags,$7F
  // done:
  {$ifdef MINIMM_DBG}
  call MmDbg_CheckMiddlePage
  {$endif MINIMM_DBG}
  jmp  @done
  //
@found_2x:
  shr  edi,1
  jmp  @found
  //
@not_large_enough:
  mov  edx,eax
  mov  eax,ebx
  call MmLinkMiddleFree
  mov  edx,[esp]
  jmp  @cur_Alloc
@damaged:
  //!!! some repair of damaged MidType free-list ?
  jmp  @cur_Alloc
  //
@not_free:
  mov  edx,[esp]
@cur_Alloc:
  // From CurAlloc: (Pool.MidPage)
  lea  eax,[ebx].TMmPool.MidPage
  call MmMidPageAlloc
  test eax,eax
  jz   @no_midPage
@done:
  {$ifdef MINIMM_DBG}
  call MmDbg_CheckMiddlePage
  {$endif MINIMM_DBG}
  add  esp,4
  pop  esi
  pop  edi
  pop  ebx
  ret
  //
@no_midPage:
  // check re-using some MidMax for curAlloc, if allocating small size...
  mov  edx,[esp]
  cmp  edx,[ebx].TMmPool.MidMax_MinSize
  ja   @new_page
  lea  eax,[ebx].TMmPool.MidMax_OffsNext
  cmp  dword ptr [eax],0
  jz   @new_page
  // unlink first item:
  //push esi
  mov  edx,eax
  add  edx,[eax] // Item:=Slot.Next
  cmp  dword ptr [edx+8],cMmMiddleFreeSign
  jne  @new_page
  mov  ecx,edx
  add  ecx,[edx] // Item.Next  - which may be Slot...
  //x: cmp  dword ptr [ecx+8],cMmMiddleFreeSign
  //x: jne  @new_page_p
  mov  esi,eax
  sub  esi,ecx // Slot-Item.Next
  mov  [ecx+4],esi // Item.Next.Prior:=Slot-Item.Next
  neg  esi
  mov  [eax],esi // Slot.Next:=Item.Next
  //pop  esi
  //
  xor  ecx,ecx
  mov  [edx+8],ecx
  //
  sub  edx,offset TMmMiddleFree.OffsNext
  xchg [ebx].TMmPool.MidPage,edx
  test edx,edx
  jnz  @release_prev_MidPage
  //
  mov  edx,[esp]
  jmp  @cur_Alloc
  //
@release_prev_MidPage:
  mov  eax,ebx
  {$ifdef MINIMM_DBG}
  call MmDbg_CheckMiddlePage
  {$endif MINIMM_DBG}
  call MmLinkMiddleFree
  mov  edx,[esp]
  jmp  @cur_Alloc
  //
@new_page_p:
  //pop  esi
  //
@new_page:
  cmp  dword ptr [esp],cMmMaxMiddleSize
  ja   @too_large
  //
  // start a new MidPage
  mov  edx,$1000
  mov  eax,[ebx].TMmPool.VirtPool.Arg
  call [ebx].TMmPool.VirtPool.AllocRawPage
  mov  edx,ebx
  call MmFormatMidPage
  test eax,eax
  jz   @OutOfMemory
  //
  lock xchg [ebx].TMmPool.MidPage,eax
  test eax,eax
  jnz  @release_prev
  mov  edx,[esp]
  jmp  @cur_Alloc
  //
@release_prev:
  {$ifdef MINIMM_DBG}
  call MmDbg_CheckMiddlePage
  {$endif MINIMM_DBG}
  mov  edx,eax
  mov  eax,ebx
  call MmLinkMiddleFree
  mov  edx,[esp]
  jmp  @cur_Alloc
  //
@too_large:
@OutOfMemory:
  pop  eax
  mov  edx,ebx
  pop  esi
  pop  edi
  pop  ebx
  //jmp MmOutOfMemoryError
  call MmOutOfMemoryError
  //ret
end;

function MmAllocLarge(Pool: PMmPool; Size: integer): Pointer; register; forward;
procedure MmPoolMidLock_Try(Pool: PMmPool; BailProc: Pointer; FailResult: Longint); register; forward;
procedure MmTry_Bailout(ExcFrame: Pointer); forward;

{function MmAllocMiddle(Pool: PMmPool; Size: integer): Pointer; register;
begin
  // Size is already checked >256 and aligned to 16 bytes...
  if (Pool.Signature=cMmPoolSign) then begin
    if (Size<cMmMaxMiddleSize) then begin
      // Alloc middle:
      Size:=(Size+7) and (not 7);
      EnterCriticalSection(Pool.MidLock);
      try
        Result:=MmInternalAllocMiddle(Pool,Size);
      finally
        LeaveCriticalSection(Pool.MidLock);
      end;
    end else begin
      Result:=MmAllocLarge(Pool,Size);
    end;
  end else begin
    // bad pool...
    Result:=nil;
  end;
end;{}
procedure MmAllocMiddle_Bailout(ExcFrame: Pointer);
asm
  // Exception-handler for MmAllocMiddle
  push eax
  //
  lea  eax,[ebx].TMmPool.MidLock
  push eax
  call LeaveCriticalSection
  //
  pop  eax
  //
  jmp  MmTry_Bailout
end;

function MmAllocMiddle(Pool: PMmPool; Size: integer): Pointer; assembler;
asm //                 EAX            EDX           ->EAX
  // try-entrance is fixed for MmTry_Bailout (just ebx above ebp)
  push ebx
  push ebp
  mov  ebp,esp
  //
  mov  ebx,eax
  push edx
  //
  cmp  [eax].TMmPool.Signature,cMmPoolSign
  jne  @bad_pool
  //
  cmp  edx,cMmMaxMiddleSize
  ja   @alloc_large
  //
  //add  edx,7
  //and  edx,not 7
  //
  mov  edx,offset MmAllocMiddle_Bailout
  xor  ecx,ecx
  // note: F8 stepping over this call in debugger does not work !
  call MmPoolMidLock_Try
  //
  mov  eax,ebx
  mov  edx,[ebp-4]
  call MmInternalAllocMiddle
  //
  //push eax
  lea  edx,[ebx].TMmPool.MidLock
  mov  ebx,eax
  push edx
  call LeaveCriticalSection
  //pop  eax
  mov  eax,ebx
  //
  xor  edx,edx
  mov  esp,fs:[edx]
  pop  dword ptr fs:[edx]
  //
@out0:
  {$ifdef MINIMM_DBG}
  call MmDbg_CheckMiddlePage
  {$endif MINIMM_DBG}
  mov  esp,ebp
  pop  ebp
  pop  ebx
  ret
  //
@alloc_large:
  call MmAllocLarge
  jmp  @out0
@bad_pool:
  xor  eax,eax
  jmp  @out0
end;

function MmAllocStartup(Pool: PMmPool; Size: integer): Pointer; assembler;
asm //                  EAX            EDX           ->EAX
  test edx,edx
  jle  @fail
  add  edx,7
  and  edx,not 7
  cmp  edx,cMmMaxMiddleSize
  ja   @large
  //
  // Note: no need for locking during EXE/DLL startup - entry-point is always single-thread...
  push ebx
  push edx
  mov  ebx,eax
  lea  eax,[eax].TMmPool.MidPage
  call MmMidPageAlloc
  test eax,eax
  jz   @start_end
  add  esp,4
  pop  ebx
  ret
  //
@start_end:
  // end of startup-page... switch to normal run-mode...
  mov  eax,ebx
  mov  ecx,offset MmAlloc
  mov  dword ptr [eax].TMmPool.AllocProc,ecx
  pop  edx
  pop  ebx
  //jmp  MmAlloc
  jmp  ecx
  //
@large:
  jmp  MmAllocLarge
  //
@fail:
  xor  eax,eax
end;

{$ifdef MINIMM_PAGEALIGNED}
function MmAllocPageAligned(Pool: PMmPool; Size: integer): Pointer; register; forward;
{$endif MINIMM_PAGEALIGNED}

{function MmAllocLarge(Pool: PMmPool; Size: integer): Pointer;
var Large: PMmLargePage;
    InSize: integer;
begin
  // Size and Pool are already checked...
  //TODO if (Size and $FFF=0) then return MmAllocPageAligned(Pool,Size)
  InSize:=Size;
  Size:=Longint(Cardinal(Size+(16+$FFF)) and $FFFFF000);
  Large:=Pool.VirtPool.AllocRawPage(Pool.VirtPool.Arg,Size);
  if (Large<>nil) then begin
    Large.Signature:=cMmLargePageSign;
    Large.Capacity:=Size;
    Large.Size:=InSize;//Size-16;
    Large.Pool:=Pool;
    Result:=@Large.Data;
  end else begin
    Result:=MmOutOfMemoryError(Size,Pointer(Pool));
  end;
end;{}

function MmAllocLarge(Pool: PMmPool; Size: integer): Pointer; assembler;
asm //                EAX            EDX           ->EAX
  // Size and Pool are already checked...
  //
  // if (Size and $FFF=0) then return MmAllocPageAligned(Pool,Size)
  {$ifdef MINIMM_PAGEALIGNED}
  test edx,$FFF
  jz   @page_aligned
  {$endif MINIMM_PAGEALIGNED}
  //
@no_aligned:
  push ebx
  mov  ebx,eax
  push edx
  //
  add  edx,$FFF+16
  and  edx,$FFFFF000 // Size:=Longint(Cardinal(Size+(16+$FFF)) and $FFFFF000);
  push edx
  //
  //Large:=Pool.VirtPool.AllocRawPage(Pool.VirtPool.Arg,Size);
  mov  eax,[eax].TMmPool.VirtPool.Arg
  call [ebx].TMmPool.VirtPool.AllocRawPage
  //
  //if (Large<>nil) then begin
  test eax,eax
  jz   @failed
  // Large.Signature:=cMmLargePageSign;
  mov  [eax].TMmLargePage.Signature,cMmLargePageSign
  // Large.Capacity:=Size;
  pop  [eax].TMmLargePage.Capacity
  // Large.Size:=InSize;//Size-16;
  pop  [eax].TMmLargePage.&Size
  // Large.Pool:=Pool;
  mov  [eax].TMmLargePage.Pool,ebx
  // Result:=@Large.Data;
  lea  eax,[eax].TMmLargePage.Data
  pop  ebx
  ret
  //
{$ifdef MINIMM_PAGEALIGNED}
@page_aligned:
  // for other pool, we would not be able to free this memory later...
  cmp  eax,g_MainMmPool
  jne  @no_aligned
  nop
  jmp  MmAllocPageAligned
{$endif MINIMM_PAGEALIGNED}
  //
@failed:
  //Result:=MmOutOfMemoryError(Size,Pointer(Pool));
  pop  eax
  mov  edx,ebx
  call MmOutOfMemoryError
  pop  edx
  pop  ebx
  //ret
end;

{$ifdef MINIMM_PAGEALIGNED}
procedure MmAllocAligned_Bailout(ExcFrame: Pointer);
asm
  // exception handler for MmAllocPageAligned
  push eax
  //
  // Free Result:
  xor  edx,edx
  xchg [ebp-8],edx
  test edx,edx
  jz   @no_res
  test edx,$FFF
  jnz  @no_res
  mov  ecx,[ebp-4]
  mov  eax,[ebx].TMmPool.VirtPool.Arg
  call [ebx].TMmPool.VirtPool.ReleaseRawPage
  //
@no_res:
  pop  eax
  //
  jmp  MmAllocMiddle_Bailout
end;

type
  PMmAlignedDesc=^TMmAlignedDesc;
  TMmAlignedDesc=record
    Mem:           Pointer;
    Size:          Longint;
  end;
  PMmAlignedList=^TMmAlignedList;
  TMmAlignedList=array[0..777] of TMmAlignedDesc;

function MmAllocPageAligned(Pool: PMmPool; Size: integer): Pointer; assembler;
asm //                      EAX            EDX           ->EAX
  // try-entrance is fixed for MmTry_Bailout (just ebx above ebp)
  push ebx
  push ebp
  mov  ebp,esp
  //
  // Pool and Size already checked...
  mov  ebx,eax
  push edx // [ebp-4]=Size
  push 0   // [ebp-8]=Result
  push edi // [ebp-12]=SaveEdi
  push esi // [ebp-16]=SaveEsi
  //
  mov  edx,offset MmAllocAligned_Bailout
  xor  ecx,ecx
  call MmPoolMidLock_Try
  //
  // Result:=Pool.VirtPool.AllocRawPage(Pool.VirtPool.Arg,Size);
  mov  eax,[ebx].TMmPool.VirtPool.Arg
  mov  edx,[ebp-4]
  call [ebx].TMmPool.VirtPool.AllocRawPage
  // if Result=nil then exit;
  test eax,eax
  jz   @fail1
  mov  [ebp-8],eax
  //
  // MmReAlloc(Pool,Pool.AlignedList,(Pool.AlignedCount+1)*8);
  lea  edx,[ebx].TMmPool.AlignedList
  mov  ecx,[ebx].TMmPool.AlignedCount
  inc  ecx
  shl  ecx,3
  mov  eax,ebx
  call MmReAlloc
  test al,al
  jz   @fail2
  //
  //
  // Find sorted position:
  mov  edx,[ebx].TMmPool.AlignedList
  xor  esi,esi // l:=0;
  mov  edi,[ebx].TMmPool.AlignedCount
  sub  edi,1 // h:=Pool.AlignedCount-1;
  // Check most common - appending...
  jl   @fbreak
  mov  ecx,[ebp-8]
  cmp  ecx,[edx+8*edi].TMmAlignedDesc.Mem
  jg   @append
  //
  // while (l<=h)
  cmp  esi,edi
  jg   @fbreak
@fcy:
  // i:=(l+h) shr 1;
  lea  eax,[esi+edi]
  shr  eax,1
  // c:=AlignedList[i].Mem-Result
  mov  ecx,[edx+8*eax].TMmAlignedDesc.Mem
  sub  ecx,[ebp-8]
  // if (c<0) then l:=i+1
  jge  @not_after
  inc  eax
  mov  esi,eax
  // while (l<=h)
  cmp  esi,edi
  jle  @fcy
  nop
  jmp  @fbreak
  //
@fail2:
  xor  edx,edx
  xchg [ebp-8],edx
  mov  ecx,[ebp-4]
  mov  eax,[ebx].TMmPool.VirtPool.Arg
  call [ebx].TMmPool.VirtPool.ReleaseRawPage
  jmp  @fail1
  //
@append:
  lea  esi,[edi+1]
  jmp  @fbreak
@not_after:
  // else h:=i-1;
  dec  eax
  mov  edi,eax
  // while (l<=h)
  cmp  esi,edi
  jle  @fcy
@fbreak:
  // Index:=l;
  lea  eax,[edx+8*esi]
  // if (Index<Count) then Move(List[Index],List[Index+1],(Count-Index) shl 3);
  mov  ecx,[ebx].TMmPool.AlignedCount
  sub  ecx,esi
  jle  @no_insert
  //shl  ecx,3
  //lea  esi,[edx+8*esi] ; add  esi,ecx
  lea  esi,[eax+8*ecx-8+4] //x191202: lea  esi,[eax+8*ecx] //zmena240217: +4 !
  lea  edi,[esi+8]
  std
  //shr  ecx,2
  add  ecx,ecx
  rep  movsd
  cld
@no_insert:
  // List[Index].Mem:=Result; List[Index].Size:=Size;
  mov  ecx,[ebp-8]
  mov  edx,[ebp-4]
  mov  [eax].TMmAlignedDesc.Mem,ecx
  mov  [eax].TMmAlignedDesc.&Size,edx
  // inc(Count);
  inc  [ebx].TMmPool.AlignedCount
  //
@fail1:
@out1:
  //push eax
  lea  eax,[ebx].TMmPool.MidLock
  push eax
  call LeaveCriticalSection
  //pop  eax
  //mov  eax,[ebp-8]
  //
  xor  edx,edx
  mov  esp,fs:[edx]
  pop  dword ptr fs:[edx]
  //
  lea  esp,[ebp-16]
  pop  esi
  pop  edi
  pop  eax // result at [ebp-8]
  pop  edx
  pop  ebp
  pop  ebx
  //ret
end;

function MmFreePageAligned(Pool: PMmPool; Page: Pointer; Item: Pointer): Boolean;
asm //                     EAX            EDX            ECX           ->AL
  //
  //if (Pool=nil) then Pool:=g_MainMmPool; if (Pool=nil) then return False;
  test eax,eax
  jnz  @have_pool
  mov  eax,g_MainMmPool
  test eax,eax
  jz   @no_pool
@have_pool:
  //
  push ebx
  push ebp
  mov  ebp,esp
  mov  ebx,eax
  //
  push ecx // [ebp-4]=Item
  //
  mov  edx,offset MmAllocMiddle_Bailout
  xor  ecx,ecx
  call MmPoolMidLock_Try
  //
  push edi
  push esi
  //
  mov  edx,[ebx].TMmPool.AlignedList
  mov  edi,[ebx].TMmPool.AlignedCount
  xor  esi,esi // l:=0;
  dec  edi     // h:=Pool.AlignedCount-1;
  // while (l<=h)
  cmp  esi,edi
  jg   @break
@fcy:
  // i:=(l+h) shr 1;
  lea  eax,[esi+edi]
  shr  eax,1
  // c:=Pool.AlignedList[i].Mem-Item;
  mov  ecx,[edx+8*eax].TMmAlignedDesc.Mem
  sub  ecx,[ebp-4]
  // if (c<0) then l:=i+1
  jl   @not_before
  // else h:=i-1;
  lea  edi,[eax-1]
  // if (c=0) then found
  jz   @found
  // while (l<=h)
@fnext:
  cmp  esi,edi
  jle  @fcy
@break:
  xor  eax,eax
  jmp  @fdone
  //
@not_before:
  // l:=i+1;
  lea  esi,[eax+1]
  jmp  @fnext
  //
@found:
  // this is a known item... Pool.AlignedList[eax].Mem=Item
  mov  esi,eax
  lea  edi,[edx+8*eax]
  xor  edx,edx
  xchg [edi].TMmAlignedDesc.Mem,edx
  mov  ecx,[edi].TMmAlignedDesc.&Size
  mov  eax,[ebx].TMmPool.VirtPool.Arg
  call [ebx].TMmPool.VirtPool.ReleaseRawPage
  //
  // dec(Pool.AlignedCount);
  mov  ecx,[ebx].TMmPool.AlignedCount
  dec  ecx
  mov  [ebx].TMmPool.AlignedCount,ecx
  // if (Index<Count)
  sub  ecx,esi // C:=Count-Index;
  jle  @no_move
  // then Move(List[Index+1],List[Index],(Count-Index) shl 3);
  lea  esi,[edi+8]
  add  ecx,ecx
  rep  movsd
@no_move:
  //
  mov  eax,1
  //
@fdone:
  pop  esi
  pop  edi
  //
@done:
  push eax
  lea  eax,[ebx].TMmPool.MidLock
  push eax
  call LeaveCriticalSection
  pop  eax
  //
  xor  edx,edx
  mov  esp,fs:[edx]
  pop  dword ptr fs:[edx]
  //
@out0:
  mov  esp,ebp
  pop  ebp
  pop  ebx
@no_pool:
  //ret
end;

function MmQueryPageAligned(Pool: PMmPool; Item: Pointer): integer;
var i,l,h,c: Longint;
begin
  Result:=0;
  EnterCriticalSection(Pool.MidLock);
  try
    l:=0;
    h:=Pool.AlignedCount-1;
    while (l<=h) do begin
      i:=(l+h) shr 1;
      c:=Longint(PMmAlignedList(Pool.AlignedList)[i].Mem)-Longint(Item);
      if (c<0) then
        l:=i+1
      else begin
        h:=i-1;
        if (c=0) then begin
          Result:=PMmAlignedList(Pool.AlignedList)[i].Size;
          break;//exit;
        end;
      end;
    end;
  finally
    LeaveCriticalSection(Pool.MidLock);
  end;
end;

function MmReAllocPageAligned(Pool: PMmPool; var P: Pointer; NewSize: integer): Boolean; assembler;
asm //                        EAX               [EDX]        ECX              ->EAX
  push ebx
  push 0   // [esp+8]=OldSize
  //New240218
  add ecx,3
  and ecx,-4
  //
  push ecx // [esp+4]=NewSize
  push edx // [esp]=^P
  mov  ebx,eax
  //
  //OldSize:=MmQueryPageAligned(Pool,P);
  mov  edx,[edx]
  call MmQueryPageAligned
  mov  [esp+8],eax
  // if (OldSize=0) then return False;
  test eax,eax
  jz   @out
  // if (NewSize=0) then MmFree(Extract(P)); return True;
  // New:=MmAlloc(Pool,NewSize);
  mov  eax,ebx
  mov  edx,[esp+4]
  test edx,edx
  jz   @free
  call MmAlloc
  // if (New=nil) then return False;
  test eax,eax
  jz   @out
  // MoveSize:=OldSize;
  mov  ecx,[esp+8]
  // if (MoveSize>NewSize) then MoveSize:=NewSize;
  cmp  ecx,[esp+4]
  jb   @size_ok
  mov  ecx,[esp+4]
@size_ok:
  // Move(P^,New^,MoveSize);
  mov  edx,[esp]
  push edi
  push esi
  mov  esi,[edx]
  mov  edi,eax
  ;//??? add  ecx,3  //New240218 !
  shr  ecx,2
  rep  movsd
  pop  esi
  pop  edi
  //
  // OldP:=P; P:=New; MmFree(nil,P);
  mov  ecx,[esp]
  mov  edx,eax
  xchg [ecx],edx
  xor  eax,eax
  call MmFree
  //
  mov  eax,1
  jmp  @out
  //
@free:
  mov  eax,[esp]
  xor  edx,edx
  xchg [eax],edx
  xor  eax,eax
  call MmFree
  mov  eax,1
  jmp  @out
  //
@out:
  add  esp,12
  pop  ebx
end;
{$endif MINIMM_PAGEALIGNED}

// No checking, discards EDX,ECX, preserves EAX, clears 12 bytes in unlinked entry...
function MmRelUnlinkC12(RelEntry: Pointer): Pointer;
asm //                  EAX               ->EAX
  push ebx
  mov  ecx,[eax] // OffsNext
  mov  edx,[eax+4] // OffsPrior
  mov  ebx,ecx
  sub  ebx,edx // OffsNext-OffsPrior
  mov  [eax+edx],ebx // Prior.OffsNext:=OffsNext-OffsPrior;
  neg  ebx
  mov  [eax+ecx+4],ebx // Next.OffsPrior:=OffsPrior-OffsNext;
  // clear 12 bytes:
  xor  ebx,ebx
  mov  [eax],ebx
  mov  [eax+4],ebx
  mov  [eax+8],ebx
  //
  pop  ebx
end;

var
  g_Mm_BadPtr: Pointer; // for debugging - pointer, that is failing FreeMem...

type
  PMmFreeExcFrame=^TMmFreeExcFrame;
  
  TExceptHandlerProc_MM=function(excPtr: PExceptionRecord; errPtr: PMmFreeExcFrame; ctxPtr,dspPtr: Pointer): Longint; stdcall;
        { ExceptHandler: }
        { ->    [ESP+ 4] excPtr: PExceptionRecord       }
        {       [ESP+ 8] errPtr: PExcFrame              }
        {       [ESP+12] ctxPtr: Pointer                }
        {       [ESP+16] dspPtr: Pointer                }
        { <-    EAX return value - always one           }
  
  TMmFreeExcFrame = record
    next:          Pointer; //PExcFrame;
    ExcHandler:    TExceptHandlerProc_MM; //desc: PExcDesc;
    pContinue:     Pointer;
    SavedRegs:     Pointer;
  end;

{$ifndef INSIDE_SYSTEM}
{$ifndef RTLUNWIND}
{$define RTLUNWIND}
const
  kernel = 'kernel32.dll';

procedure RtlUnwind;                    external kernel name 'RtlUnwind';
{$endif RTLUNWIND}
{$else ->Inside System}

//procedure IsBadCodePtr; external kernel name 'IsBadCodePtr';
{$endif INSIDE_SYSTEM}

(*
//function Exc_MmFreeMiddleHandler(excPtr: PExceptionRecord; errPtr: PGuardExcFrame; ctxPtr,dspPtr: Pointer): Longint; stdcall;
procedure Exc_MmFreeMiddle_Handler;
const
  cUnwinding          = 2;
  cUnwindingForExit   = 4;
  cUnwindInProgress   = cUnwinding or cUnwindingForExit;
asm
        { ->    [ESP+ 4] excPtr: PExceptionRecord       }
        {       [ESP+ 8] errPtr: PExcFrame              }
        {       [ESP+12] ctxPtr: Pointer                }
        {       [ESP+16] dspPtr: Pointer                }
        { <-    EAX return value - always one   }
        //
        MOV     EAX,[ESP+4]
        TEST    [EAX].TExceptionRecord.ExceptionFlags,cUnwindInProgress
        JNE     @@exit
        //
  //...
@@GoUnwind:
  cld
        OR      [EAX].TExceptionRecord.ExceptionFlags,cUnwinding
        mov     edx,[esp+8]
  // call RtlUnwind:
        PUSH    0
        PUSH    EAX
        PUSH    offset @@returnAddress
        PUSH    EDX
        CALL    RtlUnwind
  // RtlUnwind returns here...
@@returnAddress:
  // return to original code @continue, execute finally block:
  mov  esp,[esp+8] // ESP=PExcFrame
  xor  eax,eax
  pop  dword ptr fs:[eax] // pop prev-frame
  // execute Finally:
  mov  eax,[esp+8] // original Item
  and  eax,$FFFFF000
  cmp  [eax].TMmMiddlePageHdr.Signature,cMmMiddlePageSign
  jne  @bad_page
  mov  eax,[eax+cMmMidPageTailOffset].TMmMiddlePageTail.Pool
  lea  eax,[eax].TMmPool.MidLock
  push eax
  call LeaveCriticalSection
@bad_page:
  // return EAX=0
  xor  eax,eax
  ret // go @continue
@@exit:
  push 1
  pop  eax
end;

function MmFreeMiddle_v1(Pool: PMmPool; Page: Pointer; Item: Pointer): Boolean; register;
asm //                EAX            EDX            ECX           ->AL
  //xp: note: EAX=Pool=NIL !
  test ecx,$FFF
  jz   @page_aligned
  // Check, if Item is from a middle-page or large-page...
  cmp  [edx].TMmMiddlePageHdr.Signature,cMmMiddlePageSign
  jne  @not_middle
  // a middle-page:
@middle_page:
  test ecx,7
  jnz  @not_aligned
  //test ecx,$FFF -- is above...
  //jz   @not_aligned
  //
  {$ifdef MINIMM_DBG}
  call MmDbg_CheckMiddlePage
  {$endif MINIMM_DBG}
  //
  // write-check the page before entering the lock...
  xor  eax,eax
  //x: mov  [ecx],eax
  //
  // try-finally:
  push ebx   //!!!TODO - use "smart" except handler instead... (and SAFE! - now for unlocking pool a page-pointer is used, which may be gone with the page...!)
  push edi
  push esi
  push ebp
  push ecx // needed in except-handler for unlocking pool...
  push offset @continue
  push offset Exc_MmFreeMiddle_Handler
  //xor  eax,eax -- already...
  push dword ptr fs:[eax] // prev-handler-rec
  mov  fs:[eax],esp
  mov  ebp,esp
  //
  push eax // replaced below by Pool...
  push ecx
  mov  eax,[edx+cMmMidPageTailOffset].TMmMiddlePageTail.Pool
  mov  [esp+4],eax
  lea  eax,[eax].TMmPool.MidLock
  push eax
  call EnterCriticalSection
  //
  pop  ecx
  //
  movsx edx,[ecx-8].TMmMiddlePrefix.NextSizeFlags
  test edx,edx
  js   @already_free
  and  edx,$1FF
  mov  eax,edx
  shl  edx,3
  cmp  [ecx+edx].TMmMiddlePrefix.PriorSizeFlags,ax
  jne  @damaged
  //
  xor  eax,eax
  mov  [ecx],eax // at least clear first dword (object VMT)
  //x: mov  [ecx+12],eax // clear possible cMmMiddleFreeSign  -- ! not sure, that at least 16 bytes allocated!
  cmp  dword ptr [ecx+12],cMmMiddleFreeSign
  je   @do_clear_12
@in_clr:
  test ecx,$FF7 // Must NOT clear Page.Signature !
  jz   @skip_clearTag
  mov  [ecx-8].TMmMiddlePrefix.dwTag,eax // clear tag
@skip_clearTag:
  //
  {x:
  test [ecx+edx].TMmMiddlePrefix.NextFlags,$80
  jnz  @merge_next
  movsx eax,[ecx-8].TMmMiddlePrefix.PriorSizeFlags
  test eax,eax
  jl   @merge_prior{}
  movsx eax,[ecx-8].TMmMiddlePrefix.PriorSizeFlags
  test eax,eax
  jl   @merge_prior
  test [ecx+edx].TMmMiddlePrefix.NextFlags,$80
  jnz  @merge_next
  //
@single_free:
  // make stand-alone free item:
  or   [ecx-8].TMmMiddlePrefix.NextFlags,$80
  or   [ecx+edx].TMmMiddlePrefix.PriorFlags,$80
  // link to free-list:
@link_free:
  {$ifdef MINIMM_DBG}
  mov  eax,ecx
  call MmDbg_CheckMiddlePage
  {$endif MINIMM_DBG}
  mov  edx,ecx
  mov  eax,ecx
  and  eax,$FFFFF000
  mov  eax,[eax+cMmMidPageTailOffset].TMmMiddlePageTail.Pool
  call MmLinkMiddleFree
  //
  // unlock:
@middle_ok:
  mov  eax,1
@done_middle:
  xchg [esp],eax // take Pool, store Result
  lea  eax,[eax].TMmPool.MidLock
  push eax
  call LeaveCriticalSection
  //
  pop  eax
  //
  // leave try-frame...
  mov  esp,ebp
  xor  edx,edx
  pop  dword ptr fs:[edx]
  add  esp,8
@continue:
  pop  ecx
  pop  ebp
  pop  esi
  pop  edi
  pop  ebx
  //
  ret
  //
@do_clear_12:
  mov  [ecx+12],eax
  jmp  @in_clr
  //
@merge_next:
  // merge this item with next (which could be linked in free-list)  ECX=This, ECX+EDX=Next
  cmp  [ecx+edx+8].TMmMiddleFree.Signature,cMmMiddleFreeSign
  jne  @next_not_linked
  // unlink next item:
  push ebx
  push ecx
  //x:push edx
  lea  eax,[ecx+edx+8].TMmMiddleFree.OffsNext
  mov  ecx,[eax] // OffsNext
  mov  edx,[eax+4] // OffsPrior
  mov  ebx,ecx
  sub  ebx,edx // OffsNext-OffsPrior
  mov  [eax+edx],ebx // Prior.OffsNext:=This.OffsNext-This.OffsPrior
  neg  ebx
  mov  [eax+ecx+4],ebx // Next.OffsPrior:=This.OffsPrior-This.OffsNext
  xor  ebx,ebx
  mov  [eax+8],ebx // clear cMmMiddleFreeSign
  // merge blocks: EAX=Next
@merge_next_2:
  sub  eax,offset TMmMiddleFree.OffsNext
  movzx edx,[eax-8].TMmMiddlePrefix.NextSizeFlags
  and  edx,$1FF
  //x:pop  ecx
  //x:neg  ecx
  //x:add  ecx,eax // EDX=This
  pop  ecx // EDX=This
  mov  ebx,edx // EBX=Next.TileCount
  shl  edx,3   // ECX=SizeNext
  add  bx,[ecx-8].TMmMiddlePrefix.NextSizeFlags // EBX=This.TileCount+Next.TileCount
  inc  bx // EBX=This.TileCount+Next.TileCount+1  (+1 for consumed middle separator)
  or   bx,$8000
  mov  [ecx-8].TMmMiddlePrefix.NextSizeFlags,bx
  mov  [eax+edx].TMmMiddlePrefix.PriorSizeFlags,bx
  pop  ebx
  //
  jmp  @link_free
  //
@next_not_linked:
  // possibly the next item is Pool.MidPage ... ?
  cmp  [ecx+edx+8].TMmMiddleFree.Signature,cMmCurAllocSign
  jne  @single_free
  //
  mov  eax,ecx
  and  eax,$FFFFF000
  mov  eax,[eax+cMmMidPageTailOffset].TMmMiddlePageTail.Pool
  lea  edx,[ecx+edx+8]
  cmp  [eax].TMmPool.MidPage,edx
  je   @unstep_CurAlloc
  sub  edx,ecx
  sub  edx,8
  jmp  @single_free
  //
@unstep_CurAlloc:
  // Next is CurAlloc, which can be unstepped to this: ECX=This,EDX=ThisSize+...
  sub  edx,ecx
  sub  edx,8
  mov  [ecx+edx+8].TMmMiddleFree.Signature,edx // clear previous curAlloc signature
  push ebx
  mov  ebx,eax
  movzx eax,[ecx+edx].TMmMiddlePrefix.NextSizeFlags
  and  eax,$1FF
  //inc  eax; shl  eax,3; add  eax,edx
  lea  eax,[edx+8*eax+8]
  mov  edx,eax
  shr  edx,3
  or   edx,$8000
  mov  [ecx-8].TMmMiddlePrefix.NextSizeFlags,dx
  mov  [ecx+eax].TMmMiddlePrefix.PriorSizeFlags,dx
  mov  [ecx].TMmMiddleFree.Signature,cMmCurAllocSign
  mov  [ebx].TMmPool.MidPage,ecx
  pop  ebx
  jmp  @middle_ok
  //
@merge_prior:
  // merge this item with prior (which could need to get re-linked to another free-list)  ECX=This,ECX+EDX=Next, EAX=SizePrior or $8000
  //
  test [ecx+edx].TMmMiddlePrefix.NextFlags,$80
  jnz  @merge_both
  //
@prior_merge0:
  and  eax,$1FF
  shr  edx,3
  inc  eax
  add  edx,eax // EDX=CountPrior+CountThis+8
  shl  eax,3
  neg  eax
  add  eax,ecx // EAX=Prior
  //
  // merge blocks:
@prior_merge2: // EAX=Prior, EDX=NewTotalCount
  mov  ecx,edx
  or   ecx,$8000
  shl  edx,3
  mov  [eax+edx].TMmMiddlePrefix.PriorSizeFlags,cx
  mov  [eax-8].TMmMiddlePrefix.NextSizeFlags,cx
  //
  // check, if needs unlinking:
  cmp  [eax].TMmMiddleFree.Signature,cMmMiddleFreeSign
  jne  @prior_not_linked
  //
  // unlink:
  push ebx
  lea  eax,[eax].TMmMiddleFree.OffsNext
  mov  edx,[eax] // edx=OffsNext
  mov  ecx,[eax+4] // ecx=OffsPrior
  mov  ebx,edx
  sub  ebx,ecx // ebx=OffsNext-OffsPrior
  mov  [eax+ecx],ebx // Prior.OffsNext:=This.OffsNext-This.OffsPrior
  neg  ebx
  mov  [eax+edx+4],ebx // Next.OffsPrior:=This.OffsPrior-This.OffsNext
  xor  ebx,ebx
  mov  [eax+8],ebx
  pop  ebx
  // link back with new size:
  lea  edx,[eax-offset TMmMiddleFree.OffsNext]
  and  eax,$FFFFF000
  mov  eax,[eax+cMmMidPageTailOffset].TMmMiddlePageTail.Pool
  call MmLinkMiddleFree
  jmp  @middle_ok
  //
@prior_not_linked:
  nop
  jmp  @middle_ok
  //
@merge_both: // ECX=This,EDX=This.Size,EAX=PriorCount or $FFFF8000
  cmp  [ecx+edx+8].TMmMiddleFree.Signature,cMmMiddleFreeSign
  jne  @merge_prior_next_not_linked //@prior_merge0 // next not linked, just merge with prior...
  // unlink next:
  push ebx
  movzx ebx,[ecx+edx].TMmMiddlePrefix.NextSizeFlags
  and  ebx,$1FF
  and  eax,$1FF
  //add  ebx,eax
  //add  ebx,2
  lea  ebx,[eax+ebx+2]
  shl  ebx,3
  add  ebx,edx
  shr  ebx,3
  push eax
  push ebx//edx // (new)EDX=Prior.Count+This.Count+Next.Count+2
  push ecx
  lea  eax,[ecx+edx+8].TMmMiddleFree.OffsNext
  mov  edx,[eax] // edx=OffsNext
  mov  ecx,[eax+4] // ecx=OffsPrior
  mov  ebx,edx
  sub  ebx,ecx // ebx=OffsNext-OffsPrior
  mov  [eax+ecx],ebx // Prior.OffsNext:=This.OffsNext-This.OffsPrior
  neg  ebx
  mov  [eax+edx+4],ebx // Next.OffsPrior:=This.OffsPrior-This.OffsNext
  xor  ebx,ebx
  mov  [eax+8],ebx
  pop  ecx
  pop  edx
  pop  eax
  pop  ebx
  // merge with prior:
  //and  eax,$1FF
  shl  eax,3
  add  eax,8
  neg  eax
  add  eax,ecx // eax=Prior
  jmp  @prior_merge2
  //
@merge_prior_next_not_linked:
  cmp  [ecx+edx+8].TMmMiddleFree.Signature,cMmCurAllocSign
  jne  @prior_merge0 // next not linked, just merge with prior...
  push eax
  push edx
  push ecx
  mov  eax,ecx
  and  eax,$FFFFF000
  mov  eax,[eax+cMmMidPageTailOffset].TMmMiddlePageTail.Pool
  lea  edx,[ecx+edx+8]
  //cmp  [eax].TMmPool.MidPage,edx
  xchg eax,edx
  xor  ecx,ecx
  lock cmpxchg [edx].TMmPool.MidPage,ecx
  //
  pop  ecx
  pop  edx
  pop  eax
  jne  @prior_merge0
  // Next is CurAlloc, Prior is also free... EAX=CountPrior or $8000,ECX=This,EDX=ThisSize
  push ebx
  and  eax,$1FF
  mov  [ecx+edx+8].TMmMiddleFree.Signature,eax // clear CurAllocSign...
  movzx ebx,[ecx+edx].TMmMiddlePrefix.NextSizeFlags
  and  ebx,$1FF
  shl  eax,3
  add  edx,eax
  add  eax,8
  neg  eax
  add  eax,ecx // EAX=Prior
  //add  edx,ebx ; add  edx,16 // EDX=PriorSize+ThisSize+NextSize+16
  lea  edx,[edx+8*ebx+16] // EDX=PriorSize+ThisSize+NextSize+16
  // merge blocks:
  mov  ebx,edx
  shr  ebx,3
  or   ebx,$8000
  mov  [eax-8].TMmMiddlePrefix.NextSizeFlags,bx
  mov  [eax+edx].TMmMiddlePrefix.PriorSizeFlags,bx
  // unlink old prior:
  lea  eax,[eax].TMmMiddleFree.OffsNext
  call MmRelUnlinkC12
  test eax,eax
  jz   @middle_ok_pb
  sub  eax,offset TMmMiddleFree.OffsNext
  mov  edx,eax
  and  edx,$FFFFF000
  mov  edx,[edx+cMmMidPageTailOffset].TMmMiddlePageTail.Pool
  mov  [edx].TMmPool.MidPage,eax
  //jmp  @middle_ok_pb
@middle_ok_pb:
  pop  ebx
  jmp  @middle_ok
  //
@already_free:
@damaged:
  mov  g_Mm_BadPtr,ecx
  xor  eax,eax
  jmp  @done_middle
  //
@not_middle:
  // check, if it is a large-page:
  cmp  [edx].TMmPage.Signature,cMmLargePageSign
  jne  @bad_page
  // a stand-alone item (large-page)
@large_page:
  //xp: cmp  [edx].TMmLargePage.Pool,eax
  //xp: jne  @diff_pool
  mov  eax,[edx].TMmLargePage.Pool //xp: 
  push ebx
  mov  ebx,eax
  mov  ecx,[edx].TMmLargePage.Capacity
  mov  eax,[eax].TMmPool.VirtPool.Arg
  call [ebx].TMmPool.VirtPool.ReleaseRawPage
  pop  ebx
  //
  ret
  //
@page_aligned:
  {$ifdef MINIMM_PAGEALIGNED}
  nop
  jmp  MmFreePageAligned
  {$endif MINIMM_PAGEALIGNED}
  // else bail to Result:=False
  //
@not_aligned:
@diff_pool:
@bad_page:
  mov  g_Mm_BadPtr,ecx
  xor  eax,eax
end;
//*)

function MmFreeMiddle(Pool: PMmPool; Page: Pointer; Item: Pointer): Boolean; register;
asm //                EAX            EDX            ECX           ->AL
  //xp: note: EAX=Pool=NIL !
  test ecx,$FFF
  jz   @page_aligned
  // Check, if Item is from a middle-page or large-page...
  cmp  [edx].TMmMiddlePageHdr.Signature,cMmMiddlePageSign
  jne  @not_middle
  // a middle-page:
@middle_page:
  test ecx,7
  jnz  @not_aligned
  //test ecx,$FFF -- is above...
  //jz   @not_aligned
  //
  {$ifdef MINIMM_DBG}
  call MmDbg_CheckMiddlePage
  {$endif MINIMM_DBG}
  //
  // write-check the page before entering the lock...
  //xor  eax,eax
  //x: mov  [ecx],eax
  //
  // try-finally:
  push ebx
  push ebp
  mov  ebp,esp
  mov  ebx,[edx+cMmMidPageTailOffset].TMmMiddlePageTail.Pool
  mov  eax,ebx
  //
  push ecx // [ebp-4]=Item
  //
  xor  ecx,ecx
  mov  edx,offset MmAllocMiddle_Bailout
  call MmPoolMidLock_Try
  //
  mov  ecx,[ebp-4]
  //
  movsx edx,[ecx-8].TMmMiddlePrefix.NextSizeFlags
  test edx,edx
  js   @already_free
  and  edx,$1FF
  mov  eax,edx
  shl  edx,3
  cmp  [ecx+edx].TMmMiddlePrefix.PriorSizeFlags,ax
  jne  @damaged
  //
  xor  eax,eax
  mov  [ecx],eax // at least clear first dword (object VMT)
  //x: mov  [ecx+12],eax // clear possible cMmMiddleFreeSign  -- ! not sure, that at least 16 bytes allocated!
  cmp  dword ptr [ecx+12],cMmMiddleFreeSign
  je   @do_clear_12
@in_clr:
  test ecx,$FF7 // Must NOT clear Page.Signature !
  jz   @skip_clearTag
  mov  [ecx-8].TMmMiddlePrefix.dwTag,eax // clear tag
@skip_clearTag:
  //
  {x:
  test [ecx+edx].TMmMiddlePrefix.NextFlags,$80
  jnz  @merge_next
  movsx eax,[ecx-8].TMmMiddlePrefix.PriorSizeFlags
  test eax,eax
  jl   @merge_prior{}
  movsx eax,[ecx-8].TMmMiddlePrefix.PriorSizeFlags
  test eax,eax
  jl   @merge_prior
  test [ecx+edx].TMmMiddlePrefix.NextFlags,$80
  jnz  @merge_next
  //
@single_free:
  // make stand-alone free item:
  or   [ecx-8].TMmMiddlePrefix.NextFlags,$80
  or   [ecx+edx].TMmMiddlePrefix.PriorFlags,$80
  // link to free-list:
@link_free:
  {$ifdef MINIMM_DBG}
  mov  eax,ecx
  call MmDbg_CheckMiddlePage
  {$endif MINIMM_DBG}
  mov  edx,ecx
  {mov  eax,ecx
  and  eax,$FFFFF000
  mov  eax,[eax+cMmMidPageTailOffset].TMmMiddlePageTail.Pool{}
  mov  eax,ebx
  call MmLinkMiddleFree
  //
  // unlock:
@middle_ok:
  mov  eax,1
@done_middle:
  xchg eax,ebx // take Pool, save Result
  lea  eax,[eax].TMmPool.MidLock
  push eax
  call LeaveCriticalSection
  //
  mov  eax,ebx
  //
  // leave try-frame...
  xor  edx,edx
  mov  esp,fs:[edx]
  pop  dword ptr fs:[edx]
  //
  mov  esp,ebp
  pop  ebp
  pop  ebx
  //
  ret
  //
@do_clear_12:
  mov  [ecx+12],eax
  jmp  @in_clr
  //
@merge_next:
  // merge this item with next (which could be linked in free-list)  ECX=This, ECX+EDX=Next
  cmp  [ecx+edx+8].TMmMiddleFree.Signature,cMmMiddleFreeSign
  jne  @next_not_linked
  // unlink next item:
  push edi
  push ecx
  //x:push edx
  lea  eax,[ecx+edx+8].TMmMiddleFree.OffsNext
  mov  ecx,[eax] // OffsNext
  mov  edx,[eax+4] // OffsPrior
  mov  edi,ecx
  sub  edi,edx // OffsNext-OffsPrior
  mov  [eax+edx],edi // Prior.OffsNext:=This.OffsNext-This.OffsPrior
  neg  edi
  mov  [eax+ecx+4],edi // Next.OffsPrior:=This.OffsPrior-This.OffsNext
  xor  edi,edi
  mov  [eax+8],edi // clear cMmMiddleFreeSign
  // merge blocks: EAX=Next
@merge_next_2:
  sub  eax,offset TMmMiddleFree.OffsNext
  movzx edx,[eax-8].TMmMiddlePrefix.NextSizeFlags
  and  edx,$1FF
  //x:pop  ecx
  //x:neg  ecx
  //x:add  ecx,eax // EDX=This
  pop  ecx // EDX=This
  mov  edi,edx // edi=Next.TileCount
  shl  edx,3   // ECX=SizeNext
  add  di,[ecx-8].TMmMiddlePrefix.NextSizeFlags // edi=This.TileCount+Next.TileCount
  inc  di // EDI=This.TileCount+Next.TileCount+1  (+1 for consumed middle separator)
  or   di,$8000
  mov  [ecx-8].TMmMiddlePrefix.NextSizeFlags,di
  mov  [eax+edx].TMmMiddlePrefix.PriorSizeFlags,di
  pop  edi
  //
  jmp  @link_free
  //
@next_not_linked:
  // possibly the next item is Pool.MidPage ... ?
  cmp  [ecx+edx+8].TMmMiddleFree.Signature,cMmCurAllocSign
  jne  @single_free
  //
  {mov  eax,ecx
  and  eax,$FFFFF000
  mov  eax,[eax+cMmMidPageTailOffset].TMmMiddlePageTail.Pool{}
  mov  eax,ebx
  lea  edx,[ecx+edx+8]
  cmp  [eax].TMmPool.MidPage,edx
  je   @unstep_CurAlloc
  sub  edx,ecx
  sub  edx,8
  jmp  @single_free
  //
@unstep_CurAlloc:
  // Next is CurAlloc, which can be unstepped to this: ECX=This,EDX=ThisSize+...
  sub  edx,ecx
  sub  edx,8
  mov  [ecx+edx+8].TMmMiddleFree.Signature,edx // clear previous curAlloc signature
  movzx eax,[ecx+edx].TMmMiddlePrefix.NextSizeFlags
  and  eax,$1FF
  //inc  eax; shl  eax,3; add  eax,edx
  lea  eax,[edx+8*eax+8]
  mov  edx,eax
  shr  edx,3
  or   edx,$8000
  mov  [ecx-8].TMmMiddlePrefix.NextSizeFlags,dx
  mov  [ecx+eax].TMmMiddlePrefix.PriorSizeFlags,dx
  mov  [ecx].TMmMiddleFree.Signature,cMmCurAllocSign
  mov  [ebx].TMmPool.MidPage,ecx
  jmp  @middle_ok
  //
@merge_prior:
  // merge this item with prior (which could need to get re-linked to another free-list)  ECX=This,ECX+EDX=Next, EAX=SizePrior or $8000
  //
  test [ecx+edx].TMmMiddlePrefix.NextFlags,$80
  jnz  @merge_both
  //
@prior_merge0:
  and  eax,$1FF
  shr  edx,3
  inc  eax
  add  edx,eax // EDX=CountPrior+CountThis+8
  shl  eax,3
  neg  eax
  add  eax,ecx // EAX=Prior
  //
  // merge blocks:
@prior_merge2: // EAX=Prior, EDX=NewTotalCount
  mov  ecx,edx
  or   ecx,$8000
  shl  edx,3
  mov  [eax+edx].TMmMiddlePrefix.PriorSizeFlags,cx
  mov  [eax-8].TMmMiddlePrefix.NextSizeFlags,cx
  //
  // check, if needs unlinking:
  cmp  [eax].TMmMiddleFree.Signature,cMmMiddleFreeSign
  jne  @prior_not_linked
  //
  // unlink:
  push edi
  lea  eax,[eax].TMmMiddleFree.OffsNext
  mov  edx,[eax] // edx=OffsNext
  mov  ecx,[eax+4] // ecx=OffsPrior
  mov  edi,edx
  sub  edi,ecx // edi=OffsNext-OffsPrior
  mov  [eax+ecx],edi // Prior.OffsNext:=This.OffsNext-This.OffsPrior
  neg  edi
  mov  [eax+edx+4],edi // Next.OffsPrior:=This.OffsPrior-This.OffsNext
  xor  edi,edi
  mov  [eax+8],edi
  pop  edi
  // link back with new size:
  lea  edx,[eax-offset TMmMiddleFree.OffsNext]
  //and  eax,$FFFFF000
  //mov  eax,[eax+cMmMidPageTailOffset].TMmMiddlePageTail.Pool
  mov  eax,ebx
  call MmLinkMiddleFree
  jmp  @middle_ok
  //
@prior_not_linked:
  nop
  jmp  @middle_ok
  //
@merge_both: // ECX=This,EDX=This.Size,EAX=PriorCount or $FFFF8000
  cmp  [ecx+edx+8].TMmMiddleFree.Signature,cMmMiddleFreeSign
  jne  @merge_prior_next_not_linked //@prior_merge0 // next not linked, just merge with prior...
  // unlink next:
  push edi
  movzx edi,[ecx+edx].TMmMiddlePrefix.NextSizeFlags
  and  edi,$1FF
  and  eax,$1FF
  //add  edi,eax
  //add  edi,2
  lea  edi,[eax+edi+2]
  shl  edi,3
  add  edi,edx
  shr  edi,3
  push eax
  push edi//edx // (new)EDX=Prior.Count+This.Count+Next.Count+2
  push ecx
  lea  eax,[ecx+edx+8].TMmMiddleFree.OffsNext
  mov  edx,[eax] // edx=OffsNext
  mov  ecx,[eax+4] // ecx=OffsPrior
  mov  edi,edx
  sub  edi,ecx // edi=OffsNext-OffsPrior
  mov  [eax+ecx],edi // Prior.OffsNext:=This.OffsNext-This.OffsPrior
  neg  edi
  mov  [eax+edx+4],edi // Next.OffsPrior:=This.OffsPrior-This.OffsNext
  xor  edi,edi
  mov  [eax+8],edi
  pop  ecx
  pop  edx
  pop  eax
  pop  edi
  // merge with prior:
  //and  eax,$1FF
  shl  eax,3
  add  eax,8
  neg  eax
  add  eax,ecx // eax=Prior
  jmp  @prior_merge2
  //
@merge_prior_next_not_linked:
  cmp  [ecx+edx+8].TMmMiddleFree.Signature,cMmCurAllocSign
  jne  @prior_merge0 // next not linked, just merge with prior...
  push eax
  push ecx
  lea  eax,[ecx+edx+8]
  xor  ecx,ecx
  //x: lock cmpxchg [ebx].TMmPool.MidPage,ecx -- pool is locked, does not need "lock" here...
  cmpxchg [ebx].TMmPool.MidPage,ecx
  //
  pop  ecx
  pop  eax
  jne  @prior_merge0
  // Next is CurAlloc, Prior is also free... EAX=CountPrior or $8000,ECX=This,EDX=ThisSize
  push edi
  and  eax,$1FF
  mov  [ecx+edx+8].TMmMiddleFree.Signature,eax // clear CurAllocSign...
  movzx edi,[ecx+edx].TMmMiddlePrefix.NextSizeFlags
  and  edi,$1FF
  shl  eax,3
  add  edx,eax
  add  eax,8
  neg  eax
  add  eax,ecx // EAX=Prior
  //add  edx,edi ; add  edx,16 // EDX=PriorSize+ThisSize+NextSize+16
  lea  edx,[edx+8*edi+16] // EDX=PriorSize+ThisSize+NextSize+16
  // merge blocks:
  mov  edi,edx
  shr  edi,3
  or   edi,$8000
  mov  [eax-8].TMmMiddlePrefix.NextSizeFlags,di
  mov  [eax+edx].TMmMiddlePrefix.PriorSizeFlags,di
  // unlink old prior:
  lea  eax,[eax].TMmMiddleFree.OffsNext
  call MmRelUnlinkC12
  test eax,eax
  jz   @middle_ok_pb
  sub  eax,offset TMmMiddleFree.OffsNext
  {mov  edx,eax
  and  edx,$FFFFF000
  mov  edx,[edx+cMmMidPageTailOffset].TMmMiddlePageTail.Pool
  mov  [edx].TMmPool.MidPage,eax{}
  mov  [ebx].TMmPool.MidPage,eax
  //jmp  @middle_ok_pb
@middle_ok_pb:
  pop  edi
  jmp  @middle_ok
  //
@already_free:
@damaged:
  mov  g_Mm_BadPtr,ecx
  xor  eax,eax
  jmp  @done_middle
  //
@not_middle:
  // check, if it is a large-page:
  cmp  [edx].TMmPage.Signature,cMmLargePageSign
  jne  @bad_page
  // a stand-alone item (large-page)
@large_page:
  //xp: cmp  [edx].TMmLargePage.Pool,eax
  //xp: jne  @diff_pool
  mov  eax,[edx].TMmLargePage.Pool //xp: 
  push ebx
  mov  ebx,eax
  mov  ecx,[edx].TMmLargePage.Capacity
  mov  eax,[eax].TMmPool.VirtPool.Arg
  call [ebx].TMmPool.VirtPool.ReleaseRawPage
  pop  ebx
  //
  ret
  //
@page_aligned:
  {$ifdef MINIMM_PAGEALIGNED}
  nop
  jmp  MmFreePageAligned
  {$endif MINIMM_PAGEALIGNED}
  // else bail to Result:=False
  //
@not_aligned:
@diff_pool:
@bad_page:
  mov  g_Mm_BadPtr,ecx
  xor  eax,eax
end;

function MmCheckFreePageBase_v1(Item: Pointer): Boolean; register;
begin
  // Freeing Item and $FFF=0  -- can be VeryLarge item allocated at page-start?
  {$ifdef MINIMM_PAGEALIGNED}
  if (Item<>nil)
  and (Cardinal(Item) and $FFF=0)
  then begin
    Result:=MmFreePageAligned(g_MainMmPool,Item,Item);
    exit;
  end;
  {$endif MINIMM_PAGEALIGNED}
  Result:=False;
end;

function MmCheckFreePageBase(Item: Pointer): Boolean; assembler;
asm
  // Freeing Item and $FFF=0  -- can be VeryLarge item allocated at page-start?
  {$ifdef MINIMM_PAGEALIGNED}
  //if (Item<>nil)
  test eax,$FFFF0000
  jz   @bad
  //and (Cardinal(Item) and $FFF=0)
  test eax,$FFF
  jnz  @bad
  //then Result:=MmFreePageAligned(g_MainMmPool,Item,Item); exit;
  mov  edx,eax
  mov  ecx,eax
  mov  eax,g_MainMmPool
  jmp  MmFreePageAligned
  {$endif MINIMM_PAGEALIGNED}
@bad:
  //Result:=False;
  xor  eax,eax
end;

function MmQuerySizeMiddle(Item: Pointer; Page: Pointer): integer; register;
asm //                     EAX            EDX           ->EAX
  // Check, if Item is from a middle-page or large-page...
  cmp  [edx].TMmMiddlePageHdr.Signature,cMmMiddlePageSign
  jne  @not_middle
  // a middle-page:
  test eax,7
  jnz  @not_aligned
  test eax,$FFF
  jz   @not_aligned
  movsx eax,[eax-8].TMmMiddlePrefix.NextSizeFlags
  test eax,eax
  jl   @not_used
  shl  eax,3 //New240215 !
  ret
@not_used:
@not_aligned:
  xor  eax,eax
  ret
  //
@not_middle:
  // check, if it is a large-page:
  cmp  [edx].TMmPage.Signature,cMmLargePageSign
  jne  @bad_page
  // a stand-alone item (large-page)
  lea  ecx,[edx].TMmLargePage.&Data
  cmp  ecx,eax
  jne  @not_exact
  mov  eax,[edx].TMmLargePage.&Size
  sub  eax,16 // subtract header size!
  ret
@not_exact:
@bad_page:
  xor  eax,eax
end;

procedure MmLinkFreePage(Pool: PMmPool; Page: PMmPage); register; forward;
function MmCheckItemFree(Item: Pointer; bnFreeing: Boolean): Pointer; register; forward;
procedure MmCheckReleasePage(Pool: PMmPool; Page: PMmPage); register; forward;
//function MmCheckFreePageBase(Item: Pointer): Boolean; register; forward;
procedure MmCollectFree(Pool: PMmPool); register; forward; //New200822

function MmFree(Pool: PMmPool; Item: Pointer): Boolean; assembler;
asm //          EAX            EDX           ->AL
  //xp: push eax -- Pool arg unused...
  mov  ecx,edx
  mov  eax,edx
  and  edx,$FFFFF000
  jz   @bad_ptr
  test eax,$FFF
  jz   @page_base
  cmp  [edx].TMmPage.Signature,cMmPageSign
  jne  @diff_page
  // check 16-byte alignment:
  test eax,$F
  jnz  @not_exact
  // check alignment to type-size:
  mov  ecx,[edx].TMmPage.Typ
  call [ecx].TMmType.CheckAligned
  test eax,eax
  jz   @not_exact
  // check, if already free:
  cmp  [eax].TMmFreeItem.wSign,cMmFreeSign //x200822: cmp  [eax].TMmFreeItem.bSign,cMmFreeSign
  je   @check_free
@free_ok:
  {$ifdef MINIMM_DBG}
  call MmDbg_CheckSmallPage
  {$endif MINIMM_DBG}
  // clear vmt:
  xor  ecx,ecx
  mov  [eax],ecx
  mov  [eax+4],ecx
  mov  [eax+8],ecx
  mov  [eax+12],ecx
  //
  push ebx
  mov  ebx,eax
  //
  // convert to TileIndex
  mov  ecx,eax
  and  ecx,$FF0
  shr  ecx,4
  //
  mov  [ebx].TMmFreeItem.wSign,cMmFreeSign //x200822: mov  [ebx].TMmFreeItem.bSign,cMmFreeSign
@retry:
  mov  al,[edx].TMmPage.FreeList
  mov  [ebx].TMmFreeItem.NextFree,al
  lock cmpxchg [edx].TMmPage.FreeList,cl
  jne  @retry_f
  //
  mov  cl,1
  xadd [edx].TMmPage.FreeCount,cl
  cmp  cl,4 // re-link with 5 free items...
  //? cmp  cl,9 // re-link with 10 free items... (not much difference)
  je   @check_link_free
  inc  cl //New200822 -- needed after xadd...
  cmp  cl,[edx].TMmPage.MaxFreeCount
  jae  @empty_page
  //
@done_ok:
  {$ifdef MINIMM_DBG}
  mov  eax,ebx
  call MmDbg_CheckSmallPage
  {$endif MINIMM_DBG}
  mov  eax,1
  pop  ebx
  //xp: mov  [esp],eax // clear Pool from stack...
  //xp: pop  edx
  ret
  //
@diff_page:
  // could there be a middle-page or large-page with different rules?
  mov  ecx,eax
  //xp: pop  eax
  xor  eax,eax
  jmp  MmFreeMiddle //xp: Pool arg unused...
  //
@page_base:
  nop
  jmp  MmCheckFreePageBase
  //
@retry_f:
  db $F3; nop
  jmp  @retry
  //
@empty_page:
  // the whole page seems empty, check to release it...
  //xp: mov  eax,[esp+4]
  mov  eax,[edx].TMmPage.Typ
  mov  eax,[eax].TMmType.Pool
  call MmCheckReleasePage
  jmp  @done_ok
  //
@check_free:
  // item seems already free, check it:
  mov  dl,1
  call MmCheckItemFree
  mov  edx,eax
  and  edx,$FFFFF000
  jnz  @free_ok
  xor  eax,eax
  jmp  @bad_ptr
  //
@check_link_free:
  cmp  [edx].TMmPage.NextFreePage,0
  jnz  @done_ok
@link_free:
  // link page to free-list:
  //xp: mov  eax,[esp+4] // Pool...
  {$ifdef MINIMM_DBG}
  mov  eax,edx
  call MmDbg_CheckSmallPage
  {$endif MINIMM_DBG}
  mov  eax,[edx].TMmPage.Typ
  mov  eax,[eax].TMmType.Pool
  call MmLinkFreePage
  jmp  @done_ok
  //
@not_exact:
@bad_ptr:
  mov g_Mm_BadPtr,eax
  xor  eax,eax
  //note: ebx pushed later, not in this branch...
  //xp: mov  [esp],eax // clear Pool from stack...
  //xp: pop  edx
end;

function MmQuerySizeFast(Item: Pointer): integer; assembler;
asm //                   EAX           ->EAX
  mov  edx,eax
  and  edx,$FFFFF000
  jz   @bad_ptr
  cmp  [edx].TMmPage.Signature,cMmPageSign
  jne  @diff_page
  movzx eax,[edx].TMmPage.ItemSize
  shl  eax,4
  ret
@diff_page:
  nop
  jmp  MmQuerySizeMiddle
@bad_ptr:
  xor  eax,eax
end;

function MmQuerySizeCheck(Item: Pointer): integer; assembler;
asm //                    EAX           ->EAX
  mov  edx,eax
  and  edx,$FFFFF000
  jz   @bad_ptr
  test eax,$FFF
  jz   @page_aligned
  cmp  [edx].TMmPage.Signature,cMmPageSign
  jne  @diff_page
  // check 16-byte rounding:
  test eax,$F
  jnz  @not_exact
  // check alignment to type-size:
  mov  ecx,[edx].TMmPage.Typ
  call [ecx].TMmType.CheckAligned
  test eax,eax
  jz   @not_exact
  //
  cmp  [eax].TMmFreeItem.wSign,cMmFreeSign //x200822: cmp  [eax].TMmFreeItem.bSign,cMmFreeSign
  je   @check_free
@item_ok:
  //
  movzx eax,[edx].TMmPage.ItemSize
  shl  eax,4
  ret
  //
@check_free:
  xor  edx,edx
  call MmCheckItemFree
  mov  edx,eax
  and  edx,$FFFFF000
  jnz  @item_ok
  xor  eax,eax
  jmp  @bad_ptr
  //
@diff_page:
  nop
  jmp  MmQuerySizeMiddle
  //
@page_aligned:
  {$ifdef MINIMM_PAGEALIGNED}
  mov  edx,eax
  mov  eax,g_MainMmPool
  jmp  MmQueryPageAligned
  {$endif MINIMM_PAGEALIGNED}
@bad_ptr:
@not_exact:
  xor  eax,eax
end;

function MmGetItemPool(Item: Pointer): PMmPool; assembler;
asm //                 EAX           ->EAX
  mov  edx,eax
  and  edx,$FFFFF000
  jz   @bad_item
  cmp  eax,edx
  je   @bad_item //!!! is Large allocated at page-start?
  cmp  [edx].TMmPage.Signature,cMmPageSign
  jne  @not_small_page
  mov  eax,[edx].TMmPage.Typ
  mov  eax,[eax].TMmType.Pool
  ret
@not_small_page:
  cmp  [edx].TMmMiddlePageHdr.Signature,cMmMiddlePageSign
  jne  @not_middle
  mov  eax,[edx+cMmMidPageTailOffset].TMmMiddlePageTail.Pool
  ret
@not_middle:
  cmp  [edx].TMmLargePage.Signature,cMmLargePageSign
  jne  @bad_page
  mov  eax,[edx].TMmLargePage.Pool
  ret
@bad_item:
@bad_page:
  xor  eax,eax
end;

function MmReAllocLarge(Pool: PMmPool; var P: Pointer; NewSize: integer): Boolean; register;
asm //                  EAX               [EDX]        ECX              ->AL
  //!!! VirtPool.ReAllocRawPage:function(Arg: Pointer; var Page: Pointer; OldSize,NewCapacity,NewCommit: integer): Boolean register;
  push ebx
  push ecx // [esp+4]=NewSize
  push edx // [esp]=^P
  mov  ebx,eax
  //
  add  ecx,$FFF+16
  and  ecx,$FFFFF000
  mov  edx,[edx]
  and  edx,$FFFFF000
  cmp  [edx].TMmLargePage.Signature,cMmLargePageSign
  jne  @bad_page
  cmp  ecx,[edx].TMmLargePage.Capacity
  ja   @growing_capa
  mov  eax,[edx].TMmLargePage.&Size
  add  eax,$FFF
  and  eax,$FFFFF000
  cmp  ecx,eax//[edx].TMmLargePage.&Size
  ja   @growing_commit
  je   @no_change1
  // shrinking commit:
  add  ecx,ecx
  cmp  ecx,[edx].TMmLargePage.Capacity
  jb   @shrink_half
  shr  ecx,1
  // shrinking to more than half current size, just decommit...
  push MEM_DECOMMIT
  sub  eax,ecx
  push eax
  lea  eax,[edx+ecx]
  push eax
  mov  eax,[esp+4+12]
  mov  [edx].TMmLargePage.&Size,eax
  call VirtualFree
@no_change:
@done_ok:
  mov  eax,1
  add  esp,8
  pop  ebx
  ret
  //
@no_change1:
  // still within same 4kb page...
  // just change Size to actual used size:
  mov  eax,[esp+4]
  mov  [edx].TMmLargePage.&Size,eax
  jmp  @done_ok
  //
@shrink_half:
  //x: shr  ecx,1; mov edx,ecx
  // shrinking to less than half current size - realloc in new...
  mov  edx,[esp+4]
  mov  eax,ebx
  call [eax].TMmPool.AllocProc
  mov  ecx,[esp+4]
  //
@after_realloc: // ECX=MoveSize, EAX=New, [esp]=^P
  test eax,eax
  jz   @alloc_failed
  //
  mov  edx,[esp]
  push edi
  push esi
  add  ecx,3
  shr  ecx,2
  mov  esi,[edx]
  mov  edi,eax
  //
  {//!!! ?VeryLargeMove
  mov  ebx,eax
  xor  ebx,esi
  and  ebx,$F // if both Src&Dst are same 16-byte aligned, can use XMM for large-move...
  je   @very_large_move
  {}
  //
  rep  movsd
  pop  esi
  pop  edi
  //
  xchg [edx],eax
  //
  mov  edx,eax
  xor  eax,eax
  call MmFree
  //
  jmp  @done_ok
  //
@growing_commit:
  // Growing CommitSize, can retain current Capacity... EAX=PageRound(OldSize),ECX=PageRound(NewSize),EDX=Page
  //x:mov  eax,[edx].TMmLargePage.&Size
  sub  ecx,eax // ecx=AddSize
  push edx
  //
  push PAGE_READWRITE
  push MEM_COMMIT
  push ecx
  add  eax,edx
  push eax
  call VirtualAlloc
  test eax,eax
  pop  edx
  jz   @alloc_failed
  mov  eax,[esp+4]
  mov  [edx].TMmLargePage.&Size,eax
  jmp  @done_ok
  //
@growing_capa:
  // re-allocate in new space:
  //!!! use smarter ReAllocRawPage with some reserve in advance ?!
  mov  edx,[esp+4]//ecx
  mov  eax,ebx
  call [eax].TMmPool.AllocProc
  test eax,eax
  jz   @alloc_failed
  //
  mov  edx,[esp]
  mov  edx,[edx]
  and  edx,$FFFFF000
  mov  ecx,[edx].TMmLargePage.&Size // MoveSize=Old.Size
  jmp  @after_realloc
  //
@bad_page:
@alloc_failed:
  xor  eax,eax
  add  esp,8
  pop  ebx
  //ret
end;

function InternalMmMiddleShrinkInplace(Item: Pointer; NewSize: integer): Boolean; assembler;
asm //                                 EAX            EDX              ->AL
  add  edx,7
  and  edx,not 7
  jle  @invalid_size
  movsx ecx,[eax-8].TMmMiddlePrefix.NextSizeFlags
  test ecx,ecx
  jl   @not_used
  and  ecx,$1FF
  shl  ecx,3
  cmp  ecx,edx
  jbe  @no_shrink
  //
  push ebx
  push esi
  mov  ebx,ecx
  sub  ebx,edx
  cmp  ebx,32
  jb   @no_spare
  // now ecx=CurrentSize,edx=NewSize,ebx=RestSize,eax=Item
  mov  esi,edx
  shr  esi,3
  mov  [eax+edx].TMmMiddlePrefix.PriorSizeFlags,si
  mov  [eax-8].TMmMiddlePrefix.NextSizeFlags,si
  mov  esi,ebx
  sub  esi,8
  shr  esi,3
  or   esi,$8000
  mov  [eax+ecx].TMmMiddlePrefix.PriorSizeFlags,si
  mov  [eax+edx].TMmMiddlePrefix.NextSizeFlags,si
  //??? xor  esi,esi
  //??? mov  [eax+edx].TMmMiddlePrefix.dwTag,0 // clear new dwTag -- x: can ignore tags on free space...?
  //
  //!!!TODO  - check MergeNext ?!
  // add new space to free-list:
  add  edx,eax
  add edx,8
  //mov  eax,edx
  and  eax,$FFFFF000
  mov  eax,[eax+cMmMidPageTailOffset].TMmMiddlePageTail.Pool
  call MmLinkMiddleFree
  //
@no_spare:
  pop  esi
  pop  ebx
  mov  eax,1
  ret
  //  
@no_shrink:
  sete al
  ret
@invalid_size:
@not_used:
  xor  eax,eax
end;

function InternalMmMiddleGrowInplace(Item: Pointer; NewSize: integer): Boolean; assembler;
asm //                               EAX            EDX              ->AL
  add  edx,7
  and  edx,not 7
  jle  @invalid_size
  movsx ecx,[eax-8].TMmMiddlePrefix.NextSizeFlags
  test ecx,ecx
  jl   @not_used
  and  ecx,$1FF
  shl  ecx,3
  cmp  ecx,edx
  jae  @no_change
  //
  push ebx
  //
  movsx ebx,[eax+ecx].TMmMiddlePrefix.NextSizeFlags
  test ebx,ebx
  jge  @next_not_free
  //
  cmp  [eax+ecx+8].TMmMiddleFree.Signature,cMmMiddleFreeSign
  jne  @next_not_free // if the next is not linked, it may be curAlloc...
  //
  and  ebx,$1FF
  shl  ebx,3
  add  ebx,8
  add  ebx,ecx
  cmp  ebx,edx
  jne  @not_exact
@exact:
  // next space can be used exactly...
  // Unlink next space:
  push eax
  lea  eax,[eax+ecx+8].TMmMiddleFree.OffsNext
  call MmRelUnlinkC12
  pop  eax
  //
  mov  ecx,ebx
  shr  ecx,3
  mov  [eax+ebx].TMmMiddlePrefix.PriorSizeFlags,cx
  mov  [eax-8].TMmMiddlePrefix.NextSizeFlags,cx
  //
@ok_p:
  pop  ebx
@no_change:
  mov  eax,1
  ret
  //
@not_exact:
  jb   @not_enough
  // there will be some rest:
  //
  // Unlink next space:
  push edx
  push eax
  lea  eax,[eax+ecx+8].TMmMiddleFree.OffsNext
  call MmRelUnlinkC12
  pop  eax
  pop  edx
  //
  mov  ecx,ebx
  shr  ecx,3
  mov  [eax+ebx].TMmMiddlePrefix.PriorSizeFlags,cx
  mov  [eax-8].TMmMiddlePrefix.NextSizeFlags,cx
  pop  ebx
  jmp  InternalMmMiddleShrinkInplace
  //
@not_enough:
@next_not_free:
  pop  ebx
@not_used:
@invalid_size:
  xor  eax,eax
end;

{function MmMiddleGrowInplace(Pool: PMmPool; Item: Pointer; NewSize: integer): Boolean; register;
begin
  EnterCriticalSection(Pool.MidLock);
  try
    Result:=InternalMmMiddleGrowInplace(Item,NewSize);
  finally
    LeaveCriticalSection(Pool.MidLock);
  end;
end;

function MmMiddleShrinkInplace(Pool: PMmPool; Item: Pointer; NewSize: integer): Boolean; register;
begin
  EnterCriticalSection(Pool.MidLock);
  try
    Result:=InternalMmMiddleShrinkInplace(Item,NewSize);
  finally
    LeaveCriticalSection(Pool.MidLock);
  end;
end;{}

procedure MmPoolMidLock_HandleException; forward;

type
  PMidLockTryFrame=^TMidLockTryFrame;
  TMidLockTryFrame=record
    next:          Pointer; //PExcFrame;
    ExcHandler:    TExceptHandlerProc_MM; //desc: PExcDesc;
    BailProc:      Pointer;
    RegEsi:        Pointer;
    RegEdi:        Pointer;
    RegEbx:        Pointer;
    RegEbp:        Pointer;
    FailResult:    Longint;
  end;

// try-enter inspired by XP system dlls ...
procedure MmPoolMidLock_Try(Pool: PMmPool; BailProc: Pointer; FailResult: Longint);
asm //                      EAX            EDX                ECX
  //pop  ecx // ECX=ret-address
  xchg [esp],ecx
  //
  push ebp
  push ebx
  push edi
  push esi
  //
  push edx
  push offset MmPoolMidLock_HandleException
  xor  edx,edx
  push dword ptr fs:[edx]
  mov  fs:[edx],esp
  //
  push ecx // ret-address
  //
  lea  eax,[eax].TMmPool.MidLock
  push eax
  call EnterCriticalSection
  //
  //ret
end;

procedure MmPoolMidLock_HandleException;
const
  cUnwinding          = 2;
  cUnwindingForExit   = 4;
  cUnwindInProgress   = cUnwinding or cUnwindingForExit;
asm
//function GuardExcHandler(excPtr: PExceptionRecord; errPtr: PGuardExcFrame; ctxPtr,dspPtr: Pointer): Longint; stdcall;
        { ->    [ESP+ 4] excPtr: PExceptionRecord       }
        {       [ESP+ 8] errPtr: PExcFrame              }
        {       [ESP+12] ctxPtr: Pointer                }
        {       [ESP+16] dspPtr: Pointer                }
        { <-    EAX return value - always one   }
        //
        MOV     EAX,[ESP+4]
        TEST    [EAX].TExceptionRecord.ExceptionFlags,cUnwindInProgress
        JNE     @@exit
        //
  //...
@@GoUnwind:
  cld
        OR      [EAX].TExceptionRecord.ExceptionFlags,cUnwinding
        mov     edx,[esp+8]
  // call RtlUnwind:
        PUSH    0
        PUSH    EAX
        PUSH    offset @@returnAddress
        PUSH    EDX
        CALL    RtlUnwind
  // RtlUnwind returns here...
@@returnAddress:
  mov  eax,[esp+8] // ESP=PExcFrame
  xor  edx,edx
  mov  ecx,[eax].TMidLockTryFrame.next
  mov  fs:[edx],ecx // pop prev-frame
  //
  mov  esi,[eax].TMidLockTryFrame.RegEsi
  mov  edi,[eax].TMidLockTryFrame.RegEdi
  mov  ebx,[eax].TMidLockTryFrame.RegEbx
  mov  ebp,[eax].TMidLockTryFrame.RegEbp
  //
  jmp  [eax].TMidLockTryFrame.BailProc // go to MmMidGrowInplace_Bailout
  //
@@exit:
  push 1
  pop  eax
end;

procedure MmTry_Bailout(ExcFrame: Pointer);
asm
  test eax,eax
  jz   @skip_res
  mov  eax,[eax].TMidLockTryFrame.FailResult
@skip_res:
  //
  // undo stack-frame entrance in MmMiddleGrowInplace and return to original caller:
  mov  esp,ebp
  pop  ebp
  pop  ebx
  //ret
end;

procedure MmMidGrowInplace_Bailout(ExcFrame: Pointer);
asm
  // Exception-handler for MmMiddleGrowInplace and MmMiddleShrinkInplace
  push eax
  //
  lea  eax,[ebx].TMmPool.MidLock
  push eax
  call LeaveCriticalSection
  //
  pop  eax
  //
  jmp  MmTry_Bailout
end;

function MmMiddleGrowInplace(Pool: PMmPool; Item: Pointer; NewSize: integer): Boolean; register;
asm //                       EAX            EDX            ECX
  push ebx
  push ebp
  mov  ebp,esp
  //
  mov  ebx,eax
  push edx
  push ecx
  //
  {$ifdef MINIMM_DBG}
  mov  eax,edx
  call MmDbg_CheckMiddlePage
  mov  eax,ebx
  {$endif MINIMM_DBG}
  //
  mov  edx,offset MmMidGrowInplace_Bailout
  xor  ecx,ecx
  // note: F8 stepping over this call in debugger does not work !
  call MmPoolMidLock_Try
  //
  mov  eax,[ebp-4]
  mov  edx,[ebp-8]
  call InternalMmMiddleGrowInplace
  //
  push eax
  lea  eax,[ebx].TMmPool.MidLock
  push eax
  call LeaveCriticalSection
  pop  eax
  //
  xor  edx,edx
  mov  esp,fs:[edx]
  pop  dword ptr fs:[edx]
  //
  mov  esp,ebp
  pop  ebp
  pop  ebx
end;

function MmMiddleShrinkInplace(Pool: PMmPool; Item: Pointer; NewSize: integer): Boolean; register;
asm //                         EAX            EDX            ECX
  push ebx
  push ebp
  mov  ebp,esp
  //
  mov  ebx,eax
  push edx
  push ecx
  //
  {$ifdef MINIMM_DBG}
  mov  eax,edx
  call MmDbg_CheckMiddlePage
  mov  eax,ebx
  {$endif MINIMM_DBG}
  //
  mov  edx,offset MmMidGrowInplace_Bailout
  xor  ecx,ecx
  // note: F8 stepping over this call in debugger does not work !
  call MmPoolMidLock_Try
  //
  mov  eax,[ebp-4]
  mov  edx,[ebp-8]
  call InternalMmMiddleShrinkInplace
  //
  push eax
  lea  eax,[ebx].TMmPool.MidLock
  push eax
  call LeaveCriticalSection
  pop  eax
  //
  xor  edx,edx
  mov  esp,fs:[edx]
  pop  dword ptr fs:[edx]
  //
  mov  esp,ebp
  pop  ebp
  pop  ebx
end;

function MmReAlloc(Pool: PMmPool; var P: Pointer; NewSize: integer): Boolean; assembler;
asm //             EAX               [EDX]        ECX              ->EAX
  push eax // [esp+8]=Pool
  push ecx // [esp+4]=NewSize
  push edx // [esp]=^P
  mov  edx,[edx]
  mov  eax,edx
  test edx,$FFF
  jz   @page_base
  and  edx,$FFFFF000
  jz   @alloc_new
  test ecx,ecx
  jle  @free_old
  cmp  [edx].TMmPage.Signature,cMmPageSign
  jne  @diff_page
  //
@small_page:
  // round-up
  add  ecx,$F
  and  ecx,not $F
  mov  [esp+4],ecx
  //
  //x: mov  eax,edx
  call MmQuerySizeCheck
  test eax,eax
  jle  @bad_item
  mov  ecx,[esp+4]
  cmp  eax,ecx
  jae  @no_grow
  // Growing small-item - alloc a new one and move contents:
@realloc_small_g:
  mov  edx,eax
  xchg [esp+4],edx // put OldSize, take NewSize
  {x:
  mov  eax,[esp]
  mov  eax,[eax]
  and  eax,$FFFFF000
  mov  eax,[eax].TMmPage.Typ
  mov  eax,[eax].TMmType.Pool
  {}
@realloc_small2d:
  mov  eax,[esp+8]
  call [eax].TMmPool.AllocProc
  test eax,eax
  jz   @alloc_failed
@after_small_alloc:
  mov  ecx,[esp+4] // take the smaller of (OldSize,NewSize)
  mov  edx,[esp]
  push edi
  push esi
  mov  edi,eax
  mov  esi,[edx]
  add  ecx,3
  shr  ecx,2
  rep  movsd
  pop  esi
  pop  edi
  xchg [edx],eax
  mov  edx,eax
  //xp: and  eax,$FFFFF000
  //xp: mov  eax,[eax].TMmPage.Typ
  //xp: mov  eax,[eax].TMmType.Pool
  xor  eax,eax
  call MmFree
  jmp  @ok
  //
@page_base:
  test edx,$FFFFF000
  jz   @alloc_new
  {$ifdef MINIMM_PAGEALIGNED}
  pop  edx
  pop  ecx
  pop  eax
  jmp  MmReAllocPageAligned
  {$else -->!MINIMM_PAGEALIGNED}
  jmp  @bad_item
  {$endif !MINIMM_PAGEALIGNED}
  //
@no_grow:
  //je   @no_change
  // shrinking small-item: // EAX=OldSize, ECX=NewSize   ECX<EAX
  add  ecx,$20
  cmp  ecx,eax // if (NewSize<=OldSize) and (NewSize+$20>=OldSize) return True;
  jge  @no_change
  //
  sub  ecx,$20
  mov  edx,ecx
  jmp  @realloc_small2d //n2
  {x: mov  eax,[esp]
  mov  eax,[eax]
  and  eax,$FFFFF000
  mov  eax,[eax].TMmPage.Typ
  mov  eax,[eax].TMmType.Pool
  {}
  {x2: mov  eax,[esp+8]
  //
  call [eax].TMmPool.AllocProc
  test eax,eax
  jz   @alloc_failed
  //
  nop // in [esp+4] there is still NewSize, which is <OldSize
  jmp  @after_small_alloc{}
  //
@no_change:
@ok:
  mov  eax,1
  add  esp,12
  push eax // clear Pool from stack...
  pop  edx
  ret
  //
@alloc_new: // (P=nil) ...
  test ecx,ecx
  jle  @bad_request
  //
  mov  eax,[esp+8]
  mov  edx,[esp+4]
  call [eax].TMmPool.AllocProc
  test eax,eax
  jz   @alloc_failed
  mov  edx,[esp]
  mov  [edx],eax
  jmp  @ok
  //
@free_old: // (NewSize<=0)   EAX=Pool,
  jl   @bad_request
  mov  eax,[esp]
  xor  edx,edx
  xchg [eax],edx
  mov  eax,[esp+8]
  call MmFree
  jmp  @ok
  //
@diff_page:
  cmp  [edx].TMmMiddlePageHdr.Signature,cMmMiddlePageSign
  jne  @no_midPage
@middle_page:
  // ReAlloc in MiddlePage:
  mov  edx,[esp]
  mov  edx,[edx]
  movsx eax,[edx-8].TMmMiddlePrefix.NextSizeFlags
  test eax,eax
  jl   @mid_is_free
  and  eax,$1FF
  shl  eax,3
  add  ecx,7
  and  ecx,not 7
  cmp  eax,ecx
  je   @no_change
  jg   @mid_shrink
  // MiddlePage growing:
  test [edx+eax].TMmMiddlePrefix.NextFlags,$80
  jz   @next_not_free
  cmp  [edx+eax+8].TMmMiddleFree.Signature,cMmMiddleFreeSign
  jne  @next_not_free // if next is not linked, it may be cur-alloc...
  // can grow into following free space
  // MiddlePages may need locking...
  mov  eax,[esp+8]
  call MmMiddleGrowInplace
  test al,al
  jnz  @ok
  //
@next_not_free:
  // alloc new, move contents, free old:
  mov  eax,[esp+8]
  mov  edx,[esp+4]
  call [eax].TMmPool.AllocProc
  test eax,eax
  jz   @alloc_failed
  mov  edx,[esp]
  mov  edx,[edx]
  movsx ecx,[edx-8].TMmMiddlePrefix.NextSizeFlags
  and  ecx,$1FF
  //xs32: shl  ecx,3
  push edi
  push esi
  mov  esi,edx
  mov  edi,eax
  //xs32: shr  ecx,2
  add  ecx,ecx //"xs32" instead...
  rep  movsd
  pop  esi
  pop  edi
  mov  edx,[esp]
  xchg [edx],eax
  mov  edx,eax
  xor  eax,eax
  call MmFree
  //
  jmp  @ok
  //
@mid_shrink: // eax=OldSize < ecx=NewSize,edx=Item
  // if the new is very small, move to small-page...
  cmp  ecx,cMmMaxSmallSize/2
  jb   @realloc_middle_to_small
  //!!!TODO: if the next is free, it's always worth merging...
  sub  eax,cMmMaxSmallSize
  cmp  eax,ecx
  jb   @ok // not worth shrinking...
  //add  eax,cMmMaxSmallSize
  mov  eax,[esp+8]
  call MmMiddleShrinkInplace
  jmp  @ok
  //
@realloc_middle_to_small:
  mov  edx,ecx
  jmp  @realloc_small2d
  //
@no_midPage:
  cmp  [edx].TMmLargePage.Signature,cMmLargePageSign
  jne  @no_largePage
@large_page:
  // ReAlloc LargePage:
  cmp  [edx].TMmLargePage.&Size,ecx
  je   @no_change
  {add  ecx,$FFF+16
  and  ecx,$FFFFF000
  cmp  [edx].TMmLargePage.Capacity,ecx
  je   @large_in_capa{}
  //
  pop  edx
  pop  ecx
  pop  eax
  jmp  MmReAllocLarge
  //
@mid_is_free:
@no_largePage:
@bad_item:
@bad_request:
@alloc_failed:
  xor  eax,eax
  add  esp,12
  push eax // clear Pool from stack...
  pop  edx
end;

{function MmAllocWithTag(Pool: PMmPool; Size: integer; Tag: Cardinal): Pointer;
begin
  if (Size<cMmMaxMiddleSize)
  and (Size>0)
  then begin
    Result:=MmAllocMiddle(Pool,Size);
    if (Result<>nil) then
      PLongint(Longint(Result)-4)^:=Tag;
  end else
    Result:=nil;
end;{}
function MmAllocWithTag(Pool: PMmPool; Size: integer; Tag: Cardinal): Pointer; assembler;
asm //                  EAX            EDX            ECX           ->EAX
  cmp  edx,cMmMaxMiddleSize
  ja   @fail
  test edx,edx
  jle  @fail
  push edx
  push ecx
  call MmAllocMiddle
  test eax,eax
  jz   @skip_tag
  //
  //mov  edx,eax
  //and  edx,$FFF
  //cmp  edx,8
  //je   @page_header
  test eax,$FF7
  jz   @page_header
  //
@pop_tag:
  pop  dword ptr [eax-4]
  add  esp,4
  ret
@skip_tag:
  add  esp,8
@fail:
  xor  eax,eax
  ret
  //
@page_header:
  // if we get space just after PageHeader, the preceeding dwTag is Page.Signature instead...
  // must move the allocation further...
  push eax // now [esp]=NewPlace, [esp+4]=Tag, [esp+8]=Size
  mov  edx,eax
  and  eax,$FFFFF000
  mov  eax,[eax+cMmMidPageTailOffset].TMmMiddlePageTail.Pool
  mov  ecx,[esp+8]
  add  ecx,16
  call MmMiddleGrowInplace
  test al,al
  jz   @failed_grow
  // move new space, make preceeding space free:
@page_header2a:
  mov  eax,[esp]
  add  eax,16
  mov  [eax-8].TMmMiddlePrefix.PriorSizeFlags,$8001
  mov  ecx,[esp+8]
  mov  edx,ecx
  shr  ecx,3
  mov  [eax-8].TMmMiddlePrefix.NextSizeFlags,cx
  mov  [eax+ecx].TMmMiddlePrefix.PriorSizeFlags,cx
  mov  [eax-16].TMmMiddlePrefix.NextSizeFlags,$8001  //!!!TODO - test this...
  //
  add  esp,4
  pop  dword ptr [eax-4]
  add  esp,4
  ret
  //
@failed_grow:
  pop  edx
  mov  eax,edx
  and  eax,$FFFFF000
  mov  eax,[eax+cMmMidPageTailOffset].TMmMiddlePageTail.Pool
  push eax
  xor  eax,eax
  call MmFree
  //
  pop  eax
  mov  edx,[esp+4]
  add  edx,16
  call MmAllocMiddle
  //
  test eax,eax
  jz   @alloc_failed2
  //
  //mov  edx,eax
  //and  edx,$FFF
  //cmp  edx,8
  //je   @page_header2b
  test eax,$FF7
  jz   @page_header2b
  //
  add  esp,4
  jmp  @pop_tag
  //
@page_header2b:
  mov  [esp],eax
  jmp  @page_header2a
@alloc_failed2:
  add  esp,12
  xor  eax,eax
  //ret
end;

{pascal:
function MmCheckItemFree(Item: Pointer; bnFreeing: Boolean): Pointer; register;
var Walk: PMmFreeItem;
    C,idx: integer;
    Page: Pointer;
begin
  // PMmFreeItem(Item).bSign=$AB - check, if it is a free item... return NIL if bad, return Item if ok
  C:=256;
  Page:=Pointer(Longint(Item) and $FFFFF000);
  idx:=PMmPage(Page).FreeList;
  while (idx<>0) do begin
    dec(C);
    if (C=0) then
      break;
    Walk:=Pointer(Longint(Page)+(idx shl 4));
    if (Walk=Item) then begin
      // this is a linked free item...
      Result:=nil;
      exit;
    end;
    idx:=Walk.NextFree;
  end;
  // not a linked free item...
  Result:=Item;
end;{}
function MmCheckItemFree(Item: Pointer; bnFreeing: Boolean): Pointer; assembler;
asm //                   EAX            DL                 ->EAX
  // PMmFreeItem(Item).bSign=$AB - check, if it is a free item... return NIL if bad, return Item if ok
  push ebx
  mov  ebx,eax
  {$ifdef MINIMM_DBG}
  call MmDbg_CheckSmallPage
  {$endif MINIMM_DBG}
  and  ebx,$FFFFF000
  mov  ecx,256
  movzx edx,[ebx].TMmPage.FreeList
  // while (idx<>0)
  test edx,edx
  jz   @break
@cy:
  // if (--C==0) break;
  dec  ecx
  jz   @break
  // Walk:=@Page[idx shl 4]
  add  edx,edx
  lea  edx,[ebx+edx*8]
  // if (Walk=Item) then return nil;
  cmp  edx,eax
  je   @found
  // idx:=Walk.NextFree
  movzx edx,[edx].TMmFreeItem.NextFree
  // while (idx<>0)
  test edx,edx
  jnz  @cy
  // not a linked free item...
@break:
  pop  ebx
  ret
@found:
  xor  eax,eax
  pop  ebx
  //ret
end;

procedure MmLinkFreePage(Pool: PMmPool; Page: PMmPage); assembler;
asm //                   EAX            EDX
  push ebx
  mov  ebx,edx
  {$ifdef MINIMM_DBG}
  mov  eax,edx
  call MmDbg_CheckSmallPage
  {$endif MINIMM_DBG}
  cmp  [edx].TMmPage.Signature,cMmPageSign
  jne  @bad_page
  cmp  [edx].TMmPage.NextFreePage,0
  jne  @already_linked
  //
  mov  ecx,[edx].TMmPage.Typ
  cmp  ecx,$10000
  jbe  @bad_type
  //
@retry_link:
  mov  edx,[ecx].TMmType.FreePages
  xor  eax,eax
  lock cmpxchg [ebx].TMmPage.NextFreePage,edx
  jnz  @thread_problem // Page was linked by another thread?
  //
  mov  eax,edx
  lock cmpxchg [ecx].TMmType.FreePages,ebx
  jnz  @link_failed // other thread linked another free-page, so we will retry...
  //
  pop  ebx
  ret
  //
@link_failed:
  db $F3; nop
  jmp     @retry_link
  //
@done:
@bad_page:
@bad_type:
@thread_problem:
@already_linked:
  pop  ebx
end;

procedure MmCheckReleasePage(Pool: PMmPool; Page: PMmPage);
var Typ: PMmType;
begin
  // The Page seems empty (Page.FreeCount>=Page.MaxFreeCount)
  // Remove from Typ.FreePages
  // Lock the pool, check, if it is really empty, and release it or return to FreePages list
  // (since MmAlloc does not use locking, its vain to lock the pool... Add page to PageFreeSchedule list, and release later, if it is still free...)
  Typ:=Page.Typ;
  if (Typ.FreePages=Page) then // always retain top FreePage...
    exit;
  //
  if (Page.NextFreePage=nil) then
    exit;
  //
  //x200822: Pool.bnCollectNeeded:=True; //!!! where is it linked?
  //New200822
  inc(Pool.iCollectNeeded);
  if (Pool.iCollectNeeded>cMmMaxFreeSmallPages) then begin
    dec(Pool.iCollectNeeded,8);
    MmCollectFree(Pool);
    //inc(Pool.iCollectNeeded,8);
  end;
end;

function UnalignedCheck(P: Pointer): Pointer; forward;

function CheckAligned1(P: Pointer): Pointer; assembler;
asm
  // and $F already checked...
  test eax,$FF0
  jz   UnalignedCheck // item 0 is not available...
end;

function CheckAligned2(P: Pointer): Pointer; assembler;
asm
  test eax,$1F
  jnz  UnalignedCheck
  test eax,$FF0
  jz   UnalignedCheck // item 0 is not available...
end;

type
  PByteBitTable=^TByteBitTable;
  TByteBitTable=array[0..31] of Byte;

var
  Div3Table: TByteBitTable;
  Div5Table: TByteBitTable;
  Div7Table: TByteBitTable;
  Div9Table: TByteBitTable;
  Div11Table: TByteBitTable;
  Div13Table: TByteBitTable;
  Div15Table: TByteBitTable;

procedure InitAlignTables(); assembler;
asm
  mov  eax,offset Div3Table
  push 3
  pop  edx
  call @InitTable
  //
  mov  eax,offset Div5Table
  push 5
  pop  edx
  call @InitTable
  //
  mov  eax,offset Div7Table
  push 7
  pop  edx
  call @InitTable
  //
  mov  eax,offset Div9Table
  push 9
  pop  edx
  call @InitTable
  //
  mov  eax,offset Div11Table
  push 11
  pop  edx
  call @InitTable
  //
  mov  eax,offset Div13Table
  push 13
  pop  edx
  call @InitTable
  //
  mov  eax,offset Div15Table
  push 15
  pop  edx
  call @InitTable
  //
  ret
  //
@InitTable:
  // in EAX=^Table, EDX=ItemSize...
  xor  ecx,ecx
  // Item 0 should not be set...
  add  ecx,edx
@cy:
  bts  [eax],ecx
  add  ecx,edx
  cmp  ecx,$FF
  jbe  @cy
  ret
end;

function CheckAligned3(P: Pointer): Pointer; assembler;
asm
  mov  ecx,eax
  shr  ecx,4
  and  ecx,$FF
  bt   dword ptr [Div3Table],ecx
  jnc  UnalignedCheck
end;

function CheckAligned4(P: Pointer): Pointer; assembler;
asm
  test eax,$3F
  jnz  UnalignedCheck
  test eax,$FF0
  jz   UnalignedCheck // item 0 is not available...
end;

function CheckAligned5(P: Pointer): Pointer; assembler;
asm
  mov  ecx,eax
  shr  ecx,4
  and  ecx,$FF
  bt   dword ptr [Div5Table],ecx
  jnc  UnalignedCheck
end;

function CheckAligned6(P: Pointer): Pointer; assembler;
asm
  mov  ecx,eax
  and  ecx,$FFF
  shr  ecx,5
  jc   UnalignedCheck
  //and  ecx,$FF
  bt   dword ptr [Div3Table],ecx
  jnc  UnalignedCheck
end;

function CheckAligned7(P: Pointer): Pointer; assembler;
asm
  mov  ecx,eax
  shr  ecx,4
  and  ecx,$FF
  bt   dword ptr [Div7Table],ecx
  jnc  UnalignedCheck
end;

function CheckAligned8(P: Pointer): Pointer; assembler;
asm
  test eax,$7F
  jnz  UnalignedCheck
  test eax,$FF0
  jz   UnalignedCheck // item 0 is not available...
end;

function CheckAligned9(P: Pointer): Pointer; assembler;
asm
  mov  ecx,eax
  shr  ecx,4
  and  ecx,$FF
  bt   dword ptr [Div9Table],ecx
  jnc  UnalignedCheck
end;

function CheckAligned10(P: Pointer): Pointer; assembler;
asm
  mov  ecx,eax
  and  ecx,$FFF
  shr  ecx,5
  jc   UnalignedCheck
  //and  ecx,$FF
  bt   dword ptr [Div5Table],ecx
  jnc  UnalignedCheck
end;

function CheckAligned11(P: Pointer): Pointer; assembler;
asm
  mov  ecx,eax
  shr  ecx,4
  and  ecx,$FF
  bt   dword ptr [Div11Table],ecx
  jnc  UnalignedCheck
end;

function CheckAligned12(P: Pointer): Pointer; assembler;
asm
  mov  ecx,eax
  test ecx,$30
  jnz  UnalignedCheck
  and  ecx,$FFF
  shr  ecx,6
  //and  ecx,$FF
  bt   dword ptr [Div3Table],ecx
  jnc  UnalignedCheck
end;

function CheckAligned13(P: Pointer): Pointer; assembler;
asm
  mov  ecx,eax
  shr  ecx,4
  and  ecx,$FF
  bt   dword ptr [Div13Table],ecx
  jnc  UnalignedCheck
end;

function CheckAligned14(P: Pointer): Pointer; assembler;
asm
  mov  ecx,eax
  and  ecx,$FFF
  shr  ecx,5
  jc   UnalignedCheck
  //and  ecx,$FF
  bt   dword ptr [Div7Table],ecx
  jnc  UnalignedCheck
end;

function CheckAligned15(P: Pointer): Pointer; assembler;
asm
  mov  ecx,eax
  shr  ecx,4
  and  ecx,$FF
  bt   dword ptr [Div15Table],ecx
  jnc  UnalignedCheck
end;

function CheckAligned16(P: Pointer): Pointer; assembler;
asm
  test eax,$FF
  jnz  UnalignedCheck
  test eax,$FF0
  jz   UnalignedCheck // item 0 is not available...
end;

function GetAlignCheckCode(Index: integer): TCheckAlignedProc;
asm //                     EAX            ->EAX
  test eax,eax
  //jle  @not_avail
  jl   @not_avail
  cmp  eax,16
  ja   @not_avail
  mov  eax,dword ptr @CodeTable[eax*4]
  //
//???   cmp  dword ptr [Div3Table],0
//???   je   @init_tables
  //
  ret
@not_avail:
  xor  eax,eax
  ret
  //
//??? @init_tables:
//???   push eax
//???   call InitAlignTables
//???   pop  eax
//???   ret
  //
@CodeTable:
  dd offset CheckAligned1
  dd offset CheckAligned1
  dd offset CheckAligned2
  dd offset CheckAligned3
  dd offset CheckAligned4
  dd offset CheckAligned5
  dd offset CheckAligned6
  dd offset CheckAligned7
  dd offset CheckAligned8
  dd offset CheckAligned9
  dd offset CheckAligned10
  dd offset CheckAligned11
  dd offset CheckAligned12
  dd offset CheckAligned13
  dd offset CheckAligned14
  dd offset CheckAligned15
  dd offset CheckAligned16
end;

function UnalignedCheck(P: Pointer): Pointer;
asm
  // Freeing unaligned pointer...
  mov  g_Mm_BadPtr,eax
  {$ifdef MINIMM_DBG}
  call DbgBreak
  {$endif MINIMM_DBG}
  xor  eax,eax
end;

function MmOutOfMemoryError(Size: integer; Typ: PMmType): Pointer;
begin
  // Out of memory:  Typ may be PMmType or PMmPool... with PMmType, Size=0 ...
  {$ifdef MINIMM_DBG}
  DbgBreak(Typ);
  {$endif MINIMM_DBG}
  Result:=nil;
end;

function MmAllocSmallPage(Pool: PMmPool; Typ: PMmType): Pointer; //New200907
var Page: PMmPage;
begin
  // AllocRawPage uses locking anyway, so we'll link the page inside another level of the lock...
  EnterCriticalSection(Pool.MidLock);
  try
    Page:=Pool.VirtPool.AllocRawPage(Pool.VirtPool.Arg,$1000);
    if (Page<>nil) then begin
      MmInsertTailList(@Pool.SmallPageList,@Page.SmallPageLink);
    end;
  finally
    LeaveCriticalSection(Pool.MidLock);
  end;
  Result:=Page;
end;

function MmAcquirePage(Pool: PMmPool; Typ: PMmType): Pointer; assembler;
asm //                 EAX            EDX          ->EAX
  // Need another free page of this type:
  cmp  [eax].TMmPool.Signature,cMmPoolSign
  jne  @badPool
  push edx
  //x200907:
  //lea  ecx,[eax].TMmPool.VirtPool
  //mov  eax,[eax].TMmPool.VirtPool.Arg
  //mov  edx,$1000
  //call [ecx].TMmVirtPager.AllocRawPage
  //New200907 - AllocRawPage uses locking anyway, so we'll link the page inside another level of the lock...
  call MmAllocSmallPage
  //
  test eax,eax
  jz   @OutOfMemory
  mov  edx,[esp]
  //
  and  eax,$FFFFF000
  cmp  dword ptr [eax],0
  jnz  @not_empty
  //
  mov  [eax].TMmPage.Typ,edx
  mov  [eax].TMmPage.Signature,cMmPageSign
  mov  ecx,dword ptr [edx].TMmType.ItemSize // copy (ItemSize,FreeCount,FreeList,MaxFreeCount) as 1 dword...
  mov  dword ptr [eax].TMmPage.ItemSize,ecx
  //
  // format free-list:
  push ebx
  push eax
  lea  ecx,[eax+$1000]
  movzx edx,[edx].TMmType.ItemSize
  shl  edx,4
  sub  ecx,edx // last possible item...
  //New200831
  cmp  edx,32 // TMmPage header size
  jb   @more_reserve
  //
  lea  eax,[eax+edx] // first item...
  //
@fmtcy:
  lea  ebx,[eax+edx]
  cmp  ebx,ecx
  ja   @last_item
  shl  ebx,4
  and  ebx,cMmNextFreeMask_DW
  or   ebx,cMmFreeSign_DW
  mov  [eax+8],ebx
  add  eax,edx
  jmp  @fmtcy
@last_item:
  mov  dword ptr [eax+8],cMmFreeSign_DW
  //
  pop  eax
  pop  ebx
  //
  // add to Typ.FreePages:
  mov  ecx,eax
  mov  edx,[esp]
@retry_linkFree:
  mov  eax,[edx].TMmType.FreePages
  //
  test eax,eax
  jz   @first_page
  //
  mov  [ecx].TMmPage.NextFreePage,eax
  lock cmpxchg [edx].TMmType.FreePages,ecx
  jne  @thread_problem
  mov  eax,ecx
  pop  edx
  ret
  //
  //New200831
@more_reserve:
  push ecx
  lea  ecx,[eax+32] // TMmPage header size
@more_res2:
  add  eax,edx // with edx=0 it would block here...?
  cmp  eax,ecx
  jb   @more_res2
  pop  ecx
  jmp  @fmtcy
  //
@first_page:
  mov  [ecx].TMmPage.NextFreePage,edx
  lock cmpxchg [edx].TMmType.FreePages,ecx
  jne  @thread_problem
  mov  eax,ecx
  pop  edx
  ret
  //
@thread_problem:
  db $F3; nop
  jmp  @retry_linkFree
  //
@OutOfMemory:
  pop  edx
  mov  eax,$1000
  jmp  MmOutOfMemoryError
  //
@not_empty:
  pop  edx
@badPool:
  xor  eax,eax
  //ret
end;

{//x200831:
function LinkPoolFreeSched_BAD(var FreePageSched; Page: PMmPage): Boolean; //New200822
asm //                             EAX            EDX
  push edi
  push esi
  xor  esi,esi // initially checking there is (Page.NextFreePage=nil)
  mov  edi,edx // Page
  mov  ecx,eax // ^FreePageSched
  //
@retry:
  // if (Page.NextFreePage=what_should_be_there) then Page.NextFreePage:=FreePageSched head;
  mov  eax,esi // what should be in Page.NextFreePage
  mov  edx,[ecx] // FreePageSched head...
  lock cmpxchg [edi].TMmPage.NextFreePage,edx
  jne  @page_not_free
  // now there is original FreePageSched
  mov  esi,edx
  // and then FreePageSched:=Page
  mov  eax,edx
  lock cmpxchg [ecx],edi
  jne  @retry_f
  // ok
  xor  eax,eax // return True below...
@page_not_free: // return False from "jne" above...
  sete al
  //
  pop  esi
  pop  edi
  ret
  //
@retry_f:
  nop
  jmp  @retry
end;
{}

function MmSmallPage_IsEmpty(Page: PMmPage; MaxFreeCount: integer): Boolean; //New200822
asm //                       EAX            EDX
  // from MmCheckItemFree above...
  push ebx
  push esi
  push edx
  mov  ebx,eax
  test ebx,$FFF
  jnz  @bad_ptr
  xor  esi,esi // actual count...
  mov  ecx,255 // max count...
  movzx edx,[ebx].TMmPage.FreeList
  test edx,edx
  jz   @break
@cy:
  //
  dec  ecx
  jz   @infinite
  // Walk:=@Page[idx shl 4]
  add  edx,edx
  lea  edx,[ebx+edx*8]
  //
  movzx eax,[edx].TMmFreeItem.NextFree
  cmp  [edx].TMmFreeItem.wSign,cMmFreeSign
  jne  @free_damaged
  //
  inc  esi
  mov  edx,eax
  test eax,eax
  jnz  @cy
  // end of free-list...
@break:
  mov  eax,esi
  cmp  al,[ebx].TMmPage.FreeCount // FreeCount == Page.FreeCount ?
  sete dl
  cmp  esi,[esp] // FreeCount == MaxFreeCount ?
  sete al
  and  al,dl
  jmp  @done
  //
@bad_ptr:
@infinite:
@free_damaged:
  xor  eax,eax
  //
@done:
  pop  edx
  pop  esi
  pop  ebx
end;

function LinkPoolFreeSched_v2(Base: PListEntry; Page: PMmPage): Boolean; //New200831
begin
  if (Base.BLink<>nil) then begin
    // InsertTailList(Base,@Page.FreeLink);
    MmInsertTailList(Base,@Page.FreeLink);
    Result:=True;
  end else
    Result:=False;
end;

procedure MmFreeSmallPageUnlinked(Page: PMmPage; Typ: PMmType); //New200822
begin
  // now Page.NextFreePage=nil, Page.Typ=Typ, Page.Signature=cMmPageSign
  // Page.FreeCount>=Page.NextFreeCount - so verify it is empty...
  // there should be no contention on freeing from this page,
  // but other thread may have started to allocate there...
  // Called inside MidLock...
  dec(Typ.Pool.iCollectNeeded);
  if (Page.FreeCount>=Page.MaxFreeCount) then begin
    if MmSmallPage_IsEmpty(Page,Typ.MaxFreeCount)
    and (Page.FreeLink.FLink=nil) //New200906b - sometimes it is already linked ?!
    and (Page<>Typ.FreePages) //New200906b - sometimes the list is circular and we get back to top-page...?!
    then begin
      // schedule freeing this page...
      if LinkPoolFreeSched_v2(@Typ.Pool.FreePageList,Page) then //x200831: if LinkPoolFreeSched(Typ.Pool.FreePageSched,Page) then
        inc(Typ.Pool.FreeSchedCount);
      exit;
    end;
  end;
  //
  // try to return it to free-list:
  if (Page.Signature=cMmPageSign)
  and (Page.Typ=Typ)
  and (Page.NextFreePage=nil)
  then begin
    MmLinkFreePage(Typ.Pool,Page);
    inc(Typ.Pool.iCollectNeeded);
  end;
end;

procedure MmType_CollectFree(Typ: PMmType; Pool: PMmPool); //New200822
asm //                       EAX           EDX
  // cannot use locking, other threads in MmAlloc/MmFree don't use locking either...
  push ebp
  mov  ebp,esp
  //
  push edx // [ebp-4]=^Pool
  push ebx // [ebp-8]
  push edi // [ebp-12]
  push esi // [ebp-16]
  push 0   // [ebp-20] = Count
  //
  mov  esi,eax // PMmType
@restart:
  lea  ebx,[esi].TMmType.FreePages
  // skip TopPage
@cy0:
  mov  eax,[ebx]
  test eax,eax
  jle  @done
  //
  cmp  eax,esi
  je   @typ_empty
  //
  cmp  [eax].TMmPage.Typ,esi //cmp  [eax].TMmPage.Signature,cMmPageSign
  jne   @first_bad
  lea  ebx,[eax].TMmPage.NextFreePage
  //
  //New200831 - skip second also...
  mov  eax,[ebx]
  test eax,eax
  jle  @done
  cmp  eax,esi
  je   @typ_empty
  cmp  [eax].TMmPage.Typ,esi
  jne  @second_bad
  lea  ebx,[eax].TMmPage.NextFreePage
  //
@cy1:
  mov  eax,[ebx]
  test eax,eax
  jle  @done
  //
  cmp  eax,esi // last page in free-list has NextFreePage=Typ
  je   @done
  //
  cmp  [ebx-(offset TMmPage.NextFreePage)+(offset TMmPage.Typ)],esi
  jne  @done
  //
  cmp  [eax].TMmPage.Typ,esi //cmp  [eax].TMmPage.Signature,cMmPageSign
  jne  @diff_page
  mov  cl,[eax].TMmPage.FreeCount
  cmp  cl,[eax].TMmPage.MaxFreeCount
  je   @empty_page
  //
  cmp  eax,[ebx]
  jne  @chain_modified
  //
  lea  ebx,[eax].TMmPage.NextFreePage
  //
  //mov  edx,[ebp-20]
  //inc  edx
  //mov  [ebp-20],edx
  // //cmp  edx,127 ; jl   @cy1
  //cmp  dh,$FE // walk no more than 256 Mb...
  //jb   @cy1
  add  dword ptr [ebp-20],1
  //cmp  byte ptr [ebp-18],0 ; jz   @cy1
  cmp  byte ptr [ebp-19],$10 // 16Mb...
  jb   @cy1
  // infinite chain suspected, or walked enough... -- yes, it may get circular ?!
  nop
  jmp  @done
  //
@second_bad:
@first_bad:
  // is it still there?
  cmp  [ebx],eax
  jne  @restart
  // first page is bad...
  nop
  jmp  @done
  //
@empty_page:
  // this page seems empty...
  // unlink it - first take NextFreePage, then check if it was valid:
  mov  edx,[eax].TMmPage.NextFreePage
  cmp  [eax].TMmPage.Typ,esi //cmp  [eax].TMmPage.Signature,cMmPageSign
  jne  @chain_modified
  //
  mov  ecx,eax // for case cmpxchg fails, eax is modified...
  lock cmpxchg [ebx],edx
  jne  @unlink_failed
  // now the page was "acquired"... check it...
  // first also remove original NextFreePage
  mov  eax,edx
  xor  edx,edx
  lock cmpxchg [ecx].TMmPage.NextFreePage,edx
  jne  @unlink_failed
  // check it more thoroughly, if it is really free...
  //
  mov  eax,ecx
  mov  edx,esi
  call MmFreeSmallPageUnlinked
  //
  and dword ptr [ebp-20],0 // reset Count...
  //
  // now see, if the chain is still valid...
  mov  edx,ebx
  and  edx,$FFFFF000
  cmp  [edx].TMmPage.Typ,esi //cmp  [edx].TMmPage.Signature,cMmPageSign
  je   @cy1
  // previous page is gone...
  nop
  jmp  @done
  //
@unlink_failed:
  mov  eax,ecx
@diff_page:
@chain_modified:
  // is it still there in FreePages or NextFreePage?
  cmp  eax,[ebx]
  je   @chain_bad
  // this page in EAX was unlinked by someone else...
  //cmp  [ebp-20],0
  //je   @restart
  // either this or previous page was unlinked:
  mov  edx,ebx
  and  edx,$FFFFF000
  cmp  [edx].TMmPage.Typ,esi //cmp  [edx].TMmPage.Signature,cMmPageSign
  je   @cy1 // previous is ok...
  // else previous page is gone...
@chain_bad:
  nop
  //jmp  @done
  //
@typ_empty:
@done:
  lea  esp,[ebp-16]
  pop  esi
  pop  edi
  pop  ebx
  // clear Pool from stack...
  pop  edx
  push eax
  pop  eax
  pop  ebp
  //ret
end;

//x200831:
{procedure MmBgCollectWork_x200831(Pool: PMmPool);
asm
  push ebx
  push esi
  mov  ebx,eax
  // if (Pool.Signature<>cMmPoolSign) then return
  cmp  [ebx].TMmPool.Signature,cMmPoolSign
  jne  @bad_pool
  // while (Pool.FreePageSched<>nil) and (Pool.FreePageSched<>@Pool.FreePageSched) do begin
@next_page:
@retry_unlink:
  lea  edx,[ebx].TMmPool.FreePageSched
  mov  eax,[edx]
  test eax,eax
  jz   @bad_FreePageSched
  cmp  eax,edx
  je   @empty
  test eax,$FFF
  jnz  @bad_FreePageSched
  // Page:=MmUnlinkFreePageSched(Pool.FreePageSched);
  // if (Page=nil) or (Page=@Pool.FreePageSched) or (Longint(Page) and $FFF<>0) then break;
  mov  esi,eax
  mov  ecx,[eax].TMmPage.NextFreePage
  cmp  [eax].TMmPage.Signature,cMmPageSign
  jne  @bad_FreePageSched
  lock cmpxchg [edx],ecx
  jne  @retry_unlink
  sub  [ebx].TMmPool.FreeSchedCount,1
  mov  eax,ecx
  xor  ecx,ecx
  lock cmpxchg [esi].TMmPage.NextFreePage,ecx
  jne  @next_page
  //
  // if MmSmallPage_IsEmpty(Page,Typ.MaxFreeCount) then
  mov  eax,esi
  mov  edx,[eax].TMmPage.Typ
  movzx edx,[edx].TMmType.MaxFreeCount
  call MmSmallPage_IsEmpty
  test al,al
  jz   @next_page
  //New200830 - yet again check the page is not TopPage of MmType...
  mov  eax,esi
  mov  ecx,[eax].TMmPage.Typ
  cmp  [ecx].TMmType.FreePages,eax
  je   @top_page
  //
@will_release:
  // Page.Signature:=0;
  xor  ecx,ecx
  mov  eax,cMmPageSign
  lock cmpxchg [esi].TMmPage.Signature,ecx
  jne  @next_page
  // Pool.VirtPool.ReleaseRawPage(Pool.VirtPool.Arg,Page,$1000)
  mov  ecx,$1000
  mov  edx,esi
  mov  eax,[ebx].TMmPool.VirtPool.Arg
  call [ebx].TMmPool.VirtPool.ReleaseRawPage
  //
  jmp  @next_page
  //
  //New200830: may not release Top FreePage of MmType...
@top_page:
  cmp  esi,[ebx].TMmPool.FreePageSched
  jne  @next_page
  //
@bad_FreePageSched:
  // empty the whole list:
  lea  edx,[ebx].TMmPool.FreePageSched
  mov  [edx],edx
  //
@bad_pool:
@empty:
@done:
  pop  esi
  pop  ebx
  //ret
end;{}

function MmCheckFreePages(Typ: PMmType): Boolean; //New200907
asm //                    EAX (ebx)
  push ebx
  push edi
  mov  ebx,eax
  //
@restart:
  xor  ecx,ecx
  lea  edx,[eax].TMmType.FreePages
  cmp  [edx],ecx
  je   @empty 
@cy:
  mov  eax,[edx]
  //
  cmp  eax,ebx
  je   @done_ok
  test eax,eax
  je   @done_bad
  //
  cmp  [eax].TMmPage.FreeLink.FLink,0
  jnz  @unlink_free
  //
@continue:
  cmp  [edx],eax
  jne  @restart
  lea  edx,[eax].TMmPage.NextFreePage
  inc  ecx
  cmp  ecx,16384
  jb   @cy
  // chain is circular ?!
  // (we could cut it here, but if there is really so much pages, we would loose the rest...)
  mov  [eax].TMmPage.NextFreePage,ebx
@done_bad:
  xor  eax,eax
  jmp  @done
  //
@unlink_free:
  // RemoveEntryListAndClear(@eax.FreeLink)
  and  eax,$FFFFF000
  lea  eax,[eax].TMmPage.FreeLink
  mov  edi,[eax].TListEntry.FLink
  mov  ecx,[eax].TListEntry.BLink
  cmp  [edi].TListEntry.BLink,eax
  jne  @bad_link
  cmp  [ecx].TListEntry.FLink,eax
  jne  @bad_link
  mov  [edi].TListEntry.BLink,ecx
  mov  [ecx].TListEntry.FLink,edi
@bad_link:
  xor  ecx,ecx
  mov  [eax].TListEntry.FLink,ecx
  mov  [eax].TListEntry.BLink,ecx
  and  eax,$FFFFF000
  jmp  @continue
  //
@empty:
@done_ok:
  mov  al,1
@done:
  pop  edi
  pop  ebx
end;

procedure MmBgCollectWork_v2(Pool: PMmPool); //New200831
var Base,Entry,Prev: PListEntry;
    Page: PMmPage;
    i: integer;
    bnDamaged,bnRetry: Boolean;
Label _Retry;
begin
  //New200907 first walk all Typ.FreePages links (but may be circular)
  //          and all pages in Typ.FreePages chain unschedule from this releasing
  //          or unlink from FreePages chain...?
_Retry:
  bnDamaged:=False;
  i:=1;
  repeat
    if not MmCheckFreePages(Pool.SmallTypes[i]) then
      bnDamaged:=True;
    inc(i);
  until (i>High(Pool.SmallTypes));
  //
  // free pages in FreePageList, if they are really still empty...
  Base:=@Pool.FreePageList;
  while (Base.FLink<>nil) and (Base.FLink<>Base) and (Base.FLink.BLink=Base) do begin
    Entry:=Base.FLink;
    if (MmRemoveEntryListAndClear(Entry)=nil) then begin
      // Chain is damaged?!
      if (Base.FLink=Entry) then begin
        bnDamaged:=True;
        break;
      end;
      continue; //New200907
    end;
    //
    Page:=Pointer(Longint(Entry) and not $FFF);
    if (Entry=@Page.FreeLink)
    and ((Entry.FLink=Entry) or (Entry.FLink=nil))
    and (Page.NextFreePage=nil)
    and (Page.FreeCount>=Page.MaxFreeCount)
    and (Page.Typ<>nil)
    //and not IsBadWritePtr(Page.Typ,SizeOf(TMmType))
    and ((Longint(Page.Typ) xor Longint(Pool)) and (not $FFF)=0) // all Typ are in Pool-page...
    and (Page<>Page.Typ.FreePages)
    and (Page.Typ.MaxFreeCount=Page.MaxFreeCount)
    and MmSmallPage_IsEmpty(Page,Page.MaxFreeCount)
    then begin
      // Free this page now...
      dec(Pool.FreeSchedCount);
      MmRemoveEntryListAndClear(@Page.SmallPageLink); //New200907
      Pool.VirtPool.ReleaseRawPage(Pool.VirtPool.Arg,Page,$1000)
    end;
  end;
  MmInitializeListHead(@Pool.FreePageList);
  Pool.FreeSchedCount:=0;
  //
  //New200907
  if bnDamaged then begin
    // Try to re-collect other free pages, if somewhere link is broken:
    // this would swap-in whole process... (could have the link elsewhere in a "descriptor" ?)
    bnRetry:=False;
    Base:=@Pool.SmallPageList;
    Entry:=Base.FLink;
    if (Entry<>Base) and (Entry<>nil) and (Entry.BLink=Base) then
      while (Entry<>nil) and (Entry<>Base) do begin
        Page:=Pointer(Longint(Entry)-Longint(@PMmPage(nil).SmallPageLink));
        if (Cardinal(Page) and $FFF<>0)
        or IsBadWritePtr(Page,4)
        or (Page.Signature<>cMmPageSign)
        then
          break;
        //
        if (Page.FreeCount>=Page.MaxFreeCount)
        and (Page.NextFreePage=nil)
        and (Page.FreeLink.FLink=nil)
        and ((Cardinal(Page.Typ) xor Cardinal(Pool)) and $FFFFF000=0)
        and (Page<>Page.Typ.FreePages)
        and MmSmallPage_IsEmpty(Page,Page.Typ.MaxFreeCount)
        then begin
          inc(Pool.FreeSchedCount);
          MmInsertTailList(@Pool.FreePageList,@Page.FreeLink);
          bnRetry:=True;
        end;
        //
        Prev:=@Page.SmallPageLink;
        Entry:=Prev.FLink;
        if (Entry=nil) or (Entry.BLink<>Prev) then
          break;
      end;
    if bnRetry then
      goto _Retry;
  end;
  //
end;

procedure MmInternalCollectThread(Pool: PMmPool); //New200822
var H: THandle;
    i: integer;
Label _Retry;
begin
  //New200907 (try once more)
  i:=0;
_Retry:
  //
  Sleep(300); //zmena200823: 100->300 ...
  //
  if (Pool.Signature=cMmPoolSign) then begin
    EnterCriticalSection(Pool.MidLock);
    try
      MmBgCollectWork_v2(Pool); //zmena200831: +v2
      //
      if (GetCurrentThreadId=Pool.idCollectThread) then begin
        H:=Pool.hCollectThread;
        Pool.hCollectThread:=Pool.hCollectThread and 0;
        Pool.idCollectThread:=Pool.idCollectThread and 0;
        if (H<>0) then
          CloseHandle(H);
      end;
      //
    finally
      LeaveCriticalSection(Pool.MidLock);
    end;
  end;
  //New200907 - try once more soon, if some other pages were released just after our LeaveCriticalSection...
  inc(i);
  if (i=1) then
    goto _Retry;
  //
end;

function MmCollectThread(Param: Pointer): Longint; stdcall; //New200822
begin
  try
    MmInternalCollectThread(Param);
  except
    // Main thread catcher...
  end;
  Result:=0;
end;

procedure MmCollectFree(Pool: PMmPool); //New200822
var i,StartCount: integer;
    Typ: PMmType;
    H: THandle;
    ExCode: Cardinal;
Label _Out;
begin
  // Lock does not protect small-pages,
  // but protects multiple threads making this collect
  // and protects Pool.FreePageSched list...
  try
    //New200825
    if (Pool.Signature<>cMmPoolSign) then
      goto _Out;
    //
    EnterCriticalSection(Pool.MidLock);
    try
      StartCount:=Pool.FreeSchedCount;
      //
      Pool.iCollectNeeded:=Pool.iCollectNeeded and 0; // otherwise it gets negative a lot...
      //
      for i:=1 to High(Pool.SmallTypes) do begin
        Typ:=Pool.SmallTypes[i];
        if (Typ.FreePages<>nil) then
          MmType_CollectFree(Typ,Pool);
      end;
      if (Pool.iCollectNeeded<0) then
        Pool.iCollectNeeded:=Pool.iCollectNeeded and 0; // otherwise it gets negative a lot...
      //
      if (Pool.FreeSchedCount>StartCount) then begin
        // actually free the pages after some small time in other thread...
        // if they are still free...
        H:=Pool.hCollectThread;
        if (H<>0)
        and GetExitCodeThread(H,ExCode)
        then begin
          if (ExCode<>STILL_ACTIVE) then begin
            Pool.hCollectThread:=0;
            CloseHandle(H);
            H:=0;
          end else
            H:=Longint(Pool);
        end else begin
          H:=0;
          Pool.hCollectThread:=H;
        end;
        //
        if (H=0) then begin
          // Start collect thread now:
          H:=CreateThread(nil,0,@MmCollectThread,Pool,0,Pool.idCollectThread);
          Pool.hCollectThread:=H;
        end;
      end;
      //
    finally
      LeaveCriticalSection(Pool.MidLock);
    end;
    //
  _Out:
  except
    //???
    if (Pool<>nil) then ;;;
  end;
end;

{ IMemPool }

function MmSameGuid(const G1,G2: TGuid): Boolean; //New171113
asm
  push ebx
  xor ebx,ebx
  mov ecx,[eax]
  xor ecx,[edx]
  or  ebx,ecx
  mov ecx,[eax+4]
  xor ecx,[edx+4]
  or  ebx,ecx
  jnz @diff
  mov ecx,[eax+8]
  xor ecx,[edx+8]
  or  ebx,ecx
  mov ecx,[eax+12]
  xor ecx,[edx+12]
  or  ebx,ecx
@diff:
  setz al
  pop ebx
end;

//function MmIsGuidInList(const G: TGuid; GuidList: Pointer{PPointerList}): Boolean; //New171113
//var i: integer;
//begin
//  if (GuidList<>nil) then begin
//    i:=0;
//    while (GuidList[i]<>nil) do begin
//      if MmSameGuid(G,PGuid(GuidList[i])^) then begin
//        Result:=True;
//        exit;
//      end;
//      inc(i);
//    end;
//  end;
//  Result:=False;
//end;

function MmIsGuidInList(const G: TGuid; GuidList: Pointer{PPointerList}): Boolean; //New171113
asm //                        EAX       EDX                             ->AL
  push eax
  test edx,edx
  jz   @none
@cy:
  push edx
  mov  edx,[edx]
  test edx,edx
  jz   @none
  mov  eax,[esp+4]
  call MmSameGuid
  test al,al
  pop  edx
  jnz  @res
  add  edx,4
  jmp  @cy
@none:
  xor  eax,eax
@res:
  pop  edx
end;//ret

// LAddRef
{$ifndef LADDREF} {$define LADDREF} {$define LRELEASE}
function LAddRef(var Slot): Boolean; // returns true if ok (initial Slot > 0 ! )
asm //              [EAX] ->AL
  mov  edx,eax
@retry:
  mov  eax,[edx]
  test eax,eax
  jle  @fail
  lea  ecx,[eax+1]
  cmpxchg [edx],ecx
  jnz  @retry_f
  mov  al,1
  jmp  @ret // = ret
@retry_f:
  nop
  jmp  @retry
@fail:
  xor  eax,eax
@ret:
end;

function LRelease(var Slot): Boolean; // returns True if Slot released to 0...
asm //               [EAX] ->AL
  mov  edx,eax
@retry:
  mov  eax,[edx]
  test eax,eax
  jle  @fail
  lea  ecx,[eax-1]
  cmpxchg [edx],ecx
  jnz  @retry_f
  test ecx,ecx
  setz al
  jmp  @ret // = ret
@retry_f:
  nop
  jmp  @retry
@fail:
  xor  eax,eax
@ret:
end;
{$endif LADDREF}

function Mm_IMemPool_Vmt(): Pointer; assembler; //New171113
asm //                    ->EAX
  mov eax,offset @Vmt
  ret
  nop
  nop
  dd  0 //New200825
  //
@Vmt:
  // IUnknown
  dd @QueryInterface
  dd @AddRef
  dd @Release
  // IMemPool
  dd @Alloc
  dd @Free
  dd @ReAlloc
  dd @Query
  dd 0
  db 'MmPool',0,0 // just cosmetic, for debugging...
  //
@QueryInterface: // [esp+4]=[ebp+8]=Self, [esp+8]=[ebp+12]=Guid, [esp+12]=[ebp+16]=^Out
  push ebp
  mov  ebp,esp
  //
  push 0
  push offset guid_IUnknown
  push offset guid_IMemPool
  mov  eax,[ebp+12]
  mov  edx,esp
  call MmIsGuidInList
  test al,al
  mov  esp,ebp
  jz   @qi_2 //x181204: jne  @qi_2
@qi_this:
  //x200825:
  //push dword ptr [ebp+8]
  //call @AddRef
  //New200825
  mov  eax,[ebp+8]
  push eax
  mov  eax,[eax]
  call dword ptr [eax+4] // _AddRef()
  //
  mov  eax,[ebp+8]
@qi_retEax:
  mov  ecx,[ebp+16] //x181204: [ebp+12]
  mov  [ecx],eax
  xor  eax,eax
@qi_out:
  mov  esp,ebp
  pop  ebp
  retn 12
  //
@qi_IMalloc:
  mov  eax,[ebp+8]
  lea  eax,[eax+(offset TMmPool.Vmt_IMalloc)-(offset TMmPool.Vmt)]
  mov  [ebp+8],eax
  jmp  @qi_this
  //
@qi_2:
  //New200825
  push 0
  push offset guid_IMalloc
  mov  eax,[ebp+12]
  mov  edx,esp
  call MmIsGuidInList
  test al,al
  mov  esp,ebp
  jnz  @qi_IMalloc
  //
  mov  eax,E_NOINTERFACE
  jmp  @qi_out
  //
@AddRef: // [esp+4]=Self
  mov eax,[esp+4]
  sub eax,offset TMmPool.Vmt
  cmp [eax].TMmPool.Signature,cMmPoolSign
  jne @bad_ref
  lea eax,[eax].TMmPool.RefCount
  call LAddRef
@BoolRef:
  // for True return 1, for False return 0, 1-param stdcall proc...
  and eax,1
  retn 4
  //
@Release: // [esp+4]=Self
  mov eax,[esp+4]
  sub eax,offset TMmPool.Vmt
  cmp [eax].TMmPool.Signature,cMmPoolSign
  jne @bad_ref
  lea eax,[eax].TMmPool.RefCount
  call LRelease
  test al,al
  mov  al,1
  jz   @BoolRef
  // released...
  mov  eax,[esp+4]
  sub eax,offset TMmPool.Vmt
  cmp [eax].TMmPool.RefCount,0
  jne @BoolRef
  sub [eax].TMmPool.RefCount,127
  mov edx,[eax].TMmPool.Owner //New181204
  call MmDestroyPool
  //
@bad_ref:
  xor  eax,eax
  retn 4
  //
@Alloc: // eax=^VMT, edx=Size, ->eax
  sub eax,offset TMmPool.Vmt
  jmp MmAlloc
  //
@Free: // eax=^VMT, edx=Item, ->al
  sub eax,offset TMmPool.Vmt
  jmp MmFree
  //
@ReAlloc: // eax=^VMT, edx=^Item, ecx=NewSize, ->al
  sub eax,offset TMmPool.Vmt
  jmp MmReAlloc
  //
@Query: // eax=^VMT, edx=Item, ->eax
  mov eax,edx
  jmp MmQuerySizeCheck
end;

function Mm_Middle_IMemPool_Vmt(): Pointer; assembler; //New171113
asm //                           ->EAX
  mov eax,offset @Vmt
  ret
  nop
  nop
  //
  dd  0 //New200825
@Vmt:
  // IUnknown
  dd @QueryInterface
  dd @AddRef
  dd @Release
  // IMemPool
  dd @Alloc
  dd @Free
  dd @ReAlloc
  dd @Query
  dd 0
  db 'MmPoolM',0 // just cosmetic, for debugging...
  //
@QueryInterface: // [esp+4]=[ebp+8]=Self, [esp+8]=[ebp+12]=Guid, [esp+12]=[ebp+16]=^Out
  push ebp
  mov  ebp,esp
  //
  // Fix AllocProc to prevent switching to small-page MmAlloc...
  mov  eax,[ebp+8]
  sub  eax,offset TMmPool.Vmt
  cmp  [eax].TMmPool.Signature,cMmPoolSign
  jne  @qi_eFail // return E_POINTER
  cmp  dword ptr [eax].TMmPool.AllocProc,offset MmAllocStartup
  je   @FixAlloc
  cmp  dword ptr [eax].TMmPool.AllocProc,offset MmAlloc
  je   @FixAlloc
@qi_FixOk:
  //
  push 0
  push offset guid_IUnknown
  push offset guid_IMemPool
  mov  eax,[ebp+12]
  mov  edx,esp
  call MmIsGuidInList
  test al,al
  mov  esp,ebp
  jne  @qi_2
@qi_this:
  push dword ptr [ebp+8]
  call @AddRef
  mov  eax,[ebp+8]
@qi_retEax:
  mov  ecx,[ebp+12]
  mov  [ecx],eax
  xor  eax,eax
@qi_out:
  mov  esp,ebp
  pop  ebp
  retn 12
  //
@qi_2:
  mov  eax,E_NOINTERFACE
  jmp  @qi_out
@qi_eFail:
  mov  eax,E_POINTER
  jmp  @qi_out
  //
@FixAlloc:
  mov  [eax].TMmPool.AllocProc,offset MmAllocMiddle
  jmp  @qi_FixOk
  //
@AddRef: // [esp+4]=Self
  mov eax,[esp+4]
  sub eax,offset TMmPool.Vmt
  cmp [eax].TMmPool.Signature,cMmPoolSign
  jne @bad_ref
  lea eax,[eax].TMmPool.RefCount
  call LAddRef
@BoolRef:
  // for True return 1, for False return 0, 1-param stdcall proc...
  and eax,1
  retn 4
  //
@Release: // [esp+4]=Self
  mov eax,[esp+4]
  sub eax,offset TMmPool.Vmt
  cmp [eax].TMmPool.Signature,cMmPoolSign
  jne @bad_ref
  lea eax,[eax].TMmPool.RefCount
  call LRelease
  test al,al
  mov  al,1
  jz   @BoolRef
  // released...
  mov  eax,[esp+4]
  sub eax,offset TMmPool.Vmt
  cmp [eax].TMmPool.RefCount,0
  jne @BoolRef
  sub [eax].TMmPool.RefCount,127
  mov edx,[eax].TMmPool.Owner //New181204
  call MmDestroyPool
  //
@bad_ref:
  xor  eax,eax
  retn 4
  //
@Alloc: // eax=VMT, edx=Size, ->eax
  sub eax,offset TMmPool.Vmt
  jmp MmAllocMiddle
  //
@Free: // eax=VMT, edx=Item, ->al
  sub eax,offset TMmPool.Vmt
  jmp MmFreeMiddle //? MmFree
  //
@ReAlloc: // eax=VMT, edx=^Item, ecx=NewSize, ->al
  sub eax,offset TMmPool.Vmt
  jmp MmReAlloc
  //
@Query: // eax=VMT, edx=Item, ->eax
  mov eax,edx
  jmp MmQuerySizeCheck
end;

function MmIMemPool(Pool: PMmPool; out IMemPool): Boolean; register;
asm //              EAX               [EDX]     ->AL
  test eax,eax
  jz   @out
  cmp  [eax].TMmPool.Signature,cMmPoolSign
  jne  @bad_pool
  lea  eax,[eax].TMmPool.Vmt
  cmp  dword ptr [eax],0
  jz   @no_vmt
  mov  [edx],eax
  //mov  al,1
  lea  eax,[eax+(offset TMmPool.RefCount)-(offset TMmPool.Vmt)]
  call LAddRef
  ret
@out:
@bad_pool:
@no_vmt:
  xor  eax,eax
end;

function Mm_IMalloc_Vmt(): Pointer; assembler; //New200825
asm
  mov eax,offset @Vmt
  ret
  nop
  nop
  //
  dd  0 //New200825
@Vmt:
  // IUnknown:
  dd @QueryInterface
  dd @AddRef
  dd @Release
  // IMalloc:
  dd @Im_Alloc
  dd @Im_Realloc
  dd @Im_Free
  dd @Im_GetSize
  dd @Im_DidAlloc
  dd @Im_HeapMinimize
  dd 0
  //
@QueryInterface:
  push ebp
  mov  ebp,esp
  //
  push 0
  push offset guid_IUnknown
  push offset guid_IMalloc
  mov  eax,[ebp+12]
  mov  edx,esp
  call MmIsGuidInList
  test al,al
  mov  esp,ebp
  jz   @qi_2
  //
@qi_this:
  mov  eax,[ebp+8]
  push eax
  mov  eax,[eax]
  call dword ptr [eax+4] // _AddRef()
  //
  mov  eax,[ebp+8]
@qi_retEax:
  mov  ecx,[ebp+16]
  mov  [ecx],eax
  xor  eax,eax
@qi_out:
  mov  esp,ebp
  pop  ebp
  retn 12
  //
@qi_2:
  // could we query back IMemPool from IMalloc ?
  //
  mov  eax,E_NOINTERFACE
  jmp  @qi_out
  //
@AddRef: // [esp+4]=Self
  mov  eax,[esp+4]
  sub  eax,offset TMmPool.Vmt_IMalloc
  cmp  [eax].TMmPool.Signature,cMmPoolSign
  jne  @bad_ref
  lea  eax,[eax].TMmPool.RefCount
  call LAddRef
@BoolRef:
  and  eax,1
  retn 4
  //
@Release:
  mov  eax,[esp+4]
  sub  eax,offset TMmPool.Vmt_IMalloc
  cmp  [eax].TMmPool.Signature,cMmPoolSign
  jne  @bad_ref
  lea  eax,[eax].TMmPool.RefCount
  call LRelease
  test al,al
  mov  al,1
  jz   @BoolRef
  // IMalloc released... do nothing...
@bad_ref:
  xor  eax,eax
  retn 4
  //
@Im_Alloc: // [esp+4]=^Vmt_IMalloc, [esp+8]=cb
  mov  eax,[esp+4]
  sub  eax,offset TMmPool.Vmt_IMalloc
  mov  edx,[esp+8]
  pop  dword ptr [esp+4] // pop caller, store where needed... pop uses esp after popping...
  pop  ecx // discard...
  jmp  MmAlloc
  //
@Im_Free: // [esp+4]=^Vmt_IMalloc, [esp+8]=pv
  mov  eax,[esp+4]
  sub  eax,offset TMmPool.Vmt_IMalloc
  mov  edx,[esp+8]
  pop  dword ptr [esp+4] // pop caller, store where needed... pop uses esp after popping...
  pop  ecx // discard...
  jmp  MmFree
  //
@Im_ReAlloc: // [esp+4]=^Vmt_IMalloc, [esp+8]=pv, [esp+12]=cb
  mov  eax,[esp+4]
  sub  eax,offset TMmPool.Vmt_IMalloc
  mov  edx,[esp+8]
  mov  ecx,[esp+12]
  pop  dword ptr [esp+8] // pop caller, store where needed... pop uses esp after popping...
  add  esp,8
  jmp  MmReAlloc
  //
@Im_GetSize: // [esp+4]=^Vmt_IMalloc, [esp+8]=pv
  mov  eax,[esp+8]
  call MmQuerySizeCheck
  retn 8
  //
@Im_DidAlloc: // [esp+4]=^Vmt_IMalloc, [esp+8]=pv
  mov  eax,[esp+8]
  call MmQuerySizeCheck
  test eax,eax
  setnz al // 1 = our pointer, 0 = other pointer
  and  eax,1
  retn 8
  //
@Im_HeapMinimize: // [esp+4]=^Vmt_IMalloc
  mov  eax,[esp+4]
  sub  eax,offset TMmPool.Vmt_IMalloc
  call MmCollectFree
  retn 8
end;

{ MainMmPool }

function MmGetMem(Size: integer): Pointer; assembler;
asm //            EDX
  mov  edx,eax
  mov  eax,g_MainMmPool
  jmp  [eax].TMmPool.AllocProc
end;

procedure Mm_FreeMem_Error; forward;

function MmFreeMem(Ptr: Pointer): integer; assembler;
asm //             EAX
  mov  edx,eax
  xor  eax,eax
  call MmFree
  // return 0 if ok...
  and  eax,1
  setz al
  jz   Mm_FreeMem_Error
end;

procedure Mm_FreeMem_Error;
asm
  {$ifdef MINIMM_DBG}
  call DbgBreak
  {$endif MINIMM_DBG}
end;

function MmReallocMem(P: Pointer; Size: integer): Pointer; assembler;
asm
  push eax // put P to stack-var
  mov  ecx,edx
  mov  edx,esp
  mov  eax,g_MainMmPool
  call MmReAlloc
  xor  edx,edx
  test eax,eax
  pop  eax // take new value of P
  setz dl  // 1 if bad, 0 if ok
  dec  edx // 0 if bad, -1 if ok
  and  eax,edx // mask result
end;

type
  PMmVirtPool=^TMmVirtPool;
    PMmVirtSpace=^TMmVirtSpace;
  
  TMmVirtPool=record
    Signature:     Cardinal;
    //Note: MUST NOT have separate lock from main pool... "ABAB" lock problem...
    //x! Lock:          TRtlCriticalSection;
    lpLock:        Pointer;//PRtlCriticalSection;
    Spaces:        TListEntry;
    DefSpaceSize:  integer;
    MemPool:       Pointer;
    PrepSpace:     Pointer;
  end;

  TMmVirtSpace=record
    Entry:         TListEntry;
    Base:          Pointer;
    Capacity:      Cardinal;
    //
    PageCount:     Word;
    FreePageCount: Word;
    LoMap:         Pointer;
    HiMap:         Pointer;
    //Signature:     Cardinal;//dwPad...
    Signature:     Word;
    wMinFree:      Word; // lowest free page index...
  end;

const
  cMmVirtPoolSign=$84E0D36F;
  //cMmVirtSpaceSign=$DF9BEE1E;
  cMmVirtSpaceSign=$BA85;

const
  cMmInitVirtPool=$D457A8E3;

function MmVirtPoolPreCreate(): Pointer;
begin
  // wait for MmVirtPoolReInit
  Result:=Pointer(cMmInitVirtPool);
end;

procedure MmInitializeListHead(Base: PListEntry); assembler;
asm
  mov [eax],eax
  mov [eax+4],eax
end;

function MmFastNextEntry(Entry: PListEntry): Pointer; assembler;
asm //                   EAX               ->EAX
  mov  edx,eax
  mov  eax,[eax].TListEntry.FLink
  cmp  eax,edx
  je   @empty
  cmp  [eax].TListEntry.BLink,edx
  jne  @bad_link
  ret
@bad_link:
@empty:
  xor  eax,eax
  //ret
end;

function MmInsertHeadList(Base,Entry: PListEntry): Pointer; assembler;
asm //                    EAX  EDX               ->EAX   (returns Entry or NIL)
  mov  ecx,[eax].TListEntry.FLink
  cmp  [ecx].TListEntry.BLink,eax
  jne  @bad_link
  mov  [edx].TListEntry.BLink,eax
  mov  [edx].TListEntry.FLink,ecx
  mov  [eax].TListEntry.FLink,edx
  mov  [ecx].TListEntry.BLink,edx
  mov  eax,edx
  ret
@bad_link:
  xor  eax,eax
end;

function MmInsertTailList(Base,Entry: PListEntry): Pointer; assembler;
asm //                    EAX  EDX               ->EAX   (returns Entry or NIL)
  mov  ecx,[eax].TListEntry.BLink
  cmp  [ecx].TListEntry.FLink,eax
  jne  @bad_link
  mov  [edx].TListEntry.FLink,eax
  mov  [edx].TListEntry.BLink,ecx
  mov  [eax].TListEntry.BLink,edx
  mov  [ecx].TListEntry.FLink,edx
  mov  eax,edx
  ret
@bad_link:
  xor  eax,eax
end;

function MmRemoveEntryListAndClear(Entry: PListEntry): Pointer; assembler;
asm //                             EAX               ->EAX
  mov  edx,[eax].TListEntry.FLink
  mov  ecx,[eax].TListEntry.BLink
  test edx,edx
  jz   @not_linked
  test ecx,ecx
  jz   @not_linked
  cmp  [edx].TListEntry.BLink,eax
  jne  @bad_link
  cmp  [ecx].TListEntry.FLink,eax
  jne  @bad_link
  // unlink:
  mov  [edx].TListEntry.BLink,ecx
  mov  [ecx].TListEntry.FLink,edx
  xor  edx,edx
  mov  [eax].TListEntry.FLink,eax
  mov  [eax].TListEntry.BLink,eax
  ret
@bad_link:
@not_linked:
  xor eax,eax
end;

{$ifndef SETBIT} {$define SETBIT}
procedure SetBit(Dst: Pointer; Index: integer); assembler;
asm //           EAX           EDX
  bts [eax],edx
end;
{$endif SETBIT}

procedure SetNBitsAligned(Bits: Pointer; Count: integer); //New200907
asm //                    EAX            EDX
  cmp edx,8
  jb  @small
  or  ecx,-1
  sub edx,32
  jl  @after32
@cy32:
  mov [eax],ecx
  add eax,4
  sub edx,32
  jnl @cy32
@after32:
  add edx,32
  jz  @done
  sub edx,8
  jl  @after8
@cy8:
  mov [eax],cl
  inc eax
  sub edx,8
  jnl @cy8
@after8:
  add edx,8
  jz  @done
@small:
  xor ecx,ecx
@cy1:
  bts [eax],ecx
  inc ecx
  dec edx
  jnz @cy1
@done:
end;

function MmVirtAllocFromNewSpace(Pool: PMmVirtPool; Size: integer): Pointer; register;
var NewSize,PageCount,MapSize,DescSize: integer;
    Space,Other: PMmVirtSpace;
    pdw: PLongint; //pb: PByte;
Label _Fail,_HaveSpace;
begin
  // Pool is already locked, Size is rounded up...
  NewSize:=Size;
  if (Size<Pool.DefSpaceSize) then
    NewSize:=Pool.DefSpaceSize;
  //
  PageCount:=NewSize shr 12;
  MapSize:=(PageCount+7) shr 3;
  //Note: there may be no space left, so AllocProc could make an infinite recursion here...
  // So we must have 1 DefSpaceSize descriptor pre-allocated...
  DescSize:=SizeOf(TMmVirtSpace)+2*MapSize;
  if (Pool.PrepSpace<>nil)
  and (PLongint(Pool.PrepSpace)^=DescSize)
  then begin
    Space:=Pool.PrepSpace;
    Pool.PrepSpace:=nil;
    goto _HaveSpace;
  end;
  //
  //Space:=TMmPtrAlloc(PMmPool(Pool.MemPool).AllocProc)(PMmPool(Pool.MemPool),SizeOf(TMmVirtSpace)+2*MapSize);
  Space:=TMmPtrAlloc(PMmPool(Pool.MemPool).AllocProc)(PMmPool(Pool.MemPool),DescSize);
_HaveSpace:
  FillChar(Space^,DescSize,0); //New200822
  Space.Signature:=cMmVirtSpaceSign;
  Space.LoMap:=Pointer(Longint(Space)+SizeOf(TMmVirtSpace));
  Space.HiMap:=Pointer(Longint(Space.LoMap)+MapSize);
  Space.Capacity:=NewSize;
  Space.PageCount:=PageCount;
  Space.FreePageCount:=PageCount;
  Space.Base:=VirtualAlloc(nil,NewSize,MEM_RESERVE,PAGE_NOACCESS);
  if (Space.Base=nil) then begin
  _Fail:
    MmFree(Pool.MemPool,Space);
    Result:=MmOutOfMemoryError(Size,Pool.MemPool);
    exit;
  end;
  Result:=VirtualAlloc(Space.Base,Size,MEM_COMMIT,PAGE_READWRITE);
  if (Result=nil) then begin
    VirtualFree(Space.Base,0,MEM_RELEASE);
    goto _Fail;
  end;
  PageCount:=Size shr 12;
  dec(Space.FreePageCount,PageCount);
  Space.wMinFree:=PageCount;
  SetBit(Space.HiMap,0);
  {//x200907:
  if (PageCount<32) then begin
    for i:=0 to PageCount-1 do begin
      SetBit(Space.LoMap,i);
    end;
  end else begin
    i:=0;
    pdw:=Space.LoMap;
    while (PageCount>=32) do begin
      pdw^:=-1;
      dec(PageCount,32);
      inc(pdw);
      inc(i,4);
    end;
    pb:=Pointer(pdw);
    while (PageCount>8) do begin
      pb^:=$FF;
      dec(PageCount,8);
      inc(pb);
      inc(i);
    end;
    i:=i shl 3;
    while (PageCount>0) do begin
      dec(PageCount);
      SetBit(Space.LoMap,i);
      inc(i);
    end;
  end;{}
  //zmena200907 - code-size optimization:
  SetNBitsAligned(Space.LoMap,PageCount);
  //
  Other:=MmFastNextEntry(@Pool.Spaces);
  if (Other<>nil)
  //and (Other<>@Pool.Spaces)
  and (Other.FreePageCount>0)
  then
    MmInsertTailList(@Pool.Spaces,@Space.Entry)
  else
    MmInsertHeadList(@Pool.Spaces,@Space.Entry);
  //
  if (Pool.PrepSpace=nil) then begin
    // Pre-allocate a DefSpaceSize descriptor, now when we have a Space ready...
    // Note: locking problem, if we did not share a Lock with the MemPool...
    PageCount:=Pool.DefSpaceSize shr 12;
    MapSize:=(PageCount+7) shr 3;
    DescSize:=SizeOf(TMmVirtSpace)+2*MapSize;
    pdw:=TMmPtrAlloc(PMmPool(Pool.MemPool).AllocProc)(PMmPool(Pool.MemPool),DescSize);
    if (pdw<>nil) then begin
      pdw^:=DescSize;
      Pool.PrepSpace:=pdw;
    end;
  end;
end;

function MmVirtInternalPageAlloc(Pool: PMmVirtPool; Size: integer): Pointer; register;
asm //                           EAX                EDX           ->EAX
  // Pool is already locked, Size is rounded up...
  push ebx
  push edi
  push esi
  push edx    // [esp] = Size
  shr  edx,12 // PageCount
  jz   @none
@1page:
  mov  ebx,eax
  lea  eax,[eax].TMmVirtPool.Spaces
  mov  ecx,[eax].TListEntry.FLink // Space     !!!TODO - remove first space from head and link to tail, if completelly full... (spare searching)
  jmp  @cy_in
@none:
  //New240217
  pop  edx
  push $1000
  //
  push 1
  pop  edx
  jmp  @1page
@cy:
  cmp  dx,[ecx].TMmVirtSpace.FreePageCount
  ja   @next
  //
  // find free bits:
  mov  edi,[ecx].TMmVirtSpace.LoMap
  //xor  eax,eax
  movzx eax,[ecx].TMmVirtSpace.wMinFree
  cmp  ax,[ecx].TMmVirtSpace.PageCount
  ja   @next_lda //x240217: ja   @next !
  //
@bit_cy1a:
  bt   [edi],eax
  jc   @bit_skip1a
  // first bit found...
  mov  [ecx].TMmVirtSpace.wMinFree,ax
  jmp  @first_found
  //
@bit_skip1a:
  inc  eax
  cmp  ax,[ecx].TMmVirtSpace.PageCount
  jb   @bit_cy1a
  //lea  eax,[ebx].TMmVirtPool.Spaces
  //jmp  @next
  //zmena240217:
  nop
  jmp  @next_lda
  //
@bit_cy1b:
  bt   [edi],eax
  jc   @bit_skip1b
  // first bit found...
@first_found:
  mov  esi,1
  cmp  edx,esi//1
  je   @found
  inc  eax
  // check other bits:        //!!!TODO: faster bit-skipping for FFFFFFFF...
@bit_cy2:
  bt   [edi],eax
  jc   @bit_skip1b
  inc  eax
  inc  esi
  cmp  esi,edx
  jae  @found0
  cmp  ax,[ecx].TMmVirtSpace.PageCount
  jb   @bit_cy2
  //
  //lea  eax,[ebx].TMmVirtPool.Spaces
  //jmp  @next
  //zmena240217:
  nop
  jmp  @next_lda
  //
@bit_skip1b:
  inc  eax
  cmp  ax,[ecx].TMmVirtSpace.PageCount
  jb   @bit_cy1b
@next_lda:
  lea  eax,[ebx].TMmVirtPool.Spaces
  jmp  @next
  // free bits found, mark them
@found0:
  //New240218 - may have found just after the last bit, when searching for 2 from last bit...!
  cmp  ax,[ecx].TMmVirtSpace.PageCount
  ja   .next_lda
  //
  sub  eax,esi
@found:
  // now EAX=BitIndex of free space...
  cmp  ax,[ecx].TMmVirtSpace.wMinFree
  jne  @no_move_minFree
  add  eax,esi
  mov  [ecx].TMmVirtSpace.wMinFree,ax
  sub  eax,esi
@no_move_minFree:
  //
  push eax
  mov  edi,[ecx].TMmVirtSpace.HiMap
  bts  [edi],eax
  mov  edi,[ecx].TMmVirtSpace.LoMap
  mov  esi,edx
@mark_cy:
  bts  [edi],eax
  inc  eax
  sub  esi,1
  jg   @mark_cy
  pop  eax
  //
  sub  [ecx].TMmVirtSpace.FreePageCount,dx
  // commit memory:
  shl  eax,12
  add  eax,[ecx].TMmVirtSpace.Base
  shl  edx,12
  mov  edi,ecx
  push PAGE_READWRITE
  push MEM_COMMIT
  push edx
  push eax
  call VirtualAlloc
  mov  ecx,edi
  test eax,eax
  jnz  @out
  mov  edx,[esp]
  shr  edx,12 //New240217 !
@skip:
  lea  eax,[ebx].TMmVirtPool.Spaces
@next:
  mov  ecx,[ecx].TListEntry.FLink
@cy_in:
  // while (Space<>nil) and (Space<>@Pool.Spaces)
  test ecx,ecx
  jz   @break
  cmp  ecx,eax
  jne  @cy
@break:
  // not found, start a new space...
  mov  eax,ebx
  shl  edx,12
  call MmVirtAllocFromNewSpace
@out:
  add  esp,4
  pop  esi
  pop  edi
  pop  ebx
end;

function MmVirtInternalPageFree(Pool: PMmVirtPool; Page: Pointer; Size: integer): Boolean; register;
asm //                          EAX                EDX            ECX           ->AL
  // Pool is already locked, Page is checked to be a page-pointer...
  push ebx
  push edi
  push esi
  push edx // [esp]=Page
  //x140820: mov  ebx,eax
  lea  eax,[eax].TMmVirtPool.Spaces
  mov  ecx,[eax]
  cmp  ecx,eax
  je   @break
@fcy:
  mov  edx,[esp]
  sub  edx,[ecx].TMmVirtSpace.Base
  cmp  edx,[ecx].TMmVirtSpace.Capacity
  jb   @this_space
  mov  esi,[ecx].TListEntry.FLink
  test esi,esi
  jz   @list_damaged
  cmp  [esi].TListEntry.BLink,ecx
  jne  @list_damaged
  //mov  ecx,[ecx].TListEntry.FLink
  mov  ecx,esi
  cmp  esi,eax
  jne  @fcy
  // space not found...
@fail:
  xor  eax,eax
@out:
  add  esp,4
  pop  esi
  pop  edi
  pop  ebx
  ret
  //
@ok:
  mov  eax,1
  jmp  @out
  //
@list_damaged:
  // Some repair?
  //!!!
  jmp @fail
  //
@this_space:
  shr  edx,12 // PageIndex
  mov  eax,[ecx].TMmVirtSpace.HiMap
  mov  edi,[ecx].TMmVirtSpace.LoMap
  //
  cmp  dx,[ecx].TMmVirtSpace.wMinFree
  ja   @not_min_free
  mov  [ecx].TMmVirtSpace.wMinFree,dx
@not_min_free:
  //
  btr  [eax],edx
  jnc  @not_start
  xor  esi,esi
@cy:
  btr  [edi],edx
  jnc  @break
  //inc  [ecx].TMmVirtSpace.FreePageCount
  inc  esi
  inc  edx
  cmp  dx,[ecx].TMmVirtSpace.PageCount
  jae  @break
  bt   [eax],edx
  jnc  @cy
@break:
  add  [ecx].TMmVirtSpace.FreePageCount,si
  mov  ebx,ecx //New140820
  //New200907 - if there is error and esi=0 at start of space, that would decommit whole space
  test esi,esi
  jz   @skip_decommit
  //
  // free pages:
  mov  edx,[esp]
  push MEM_DECOMMIT
  shl  esi,12
  push esi
  push edx
  call VirtualFree
  //
@skip_decommit:
  //
  mov  ecx,ebx //New140820
  //
  // Check, if whole page released:
  movzx eax,[ecx].TMmVirtSpace.FreePageCount
  cmp  ax,[ecx].TMmVirtSpace.PageCount
  jb   @ok
  //
  // check, that really free:
  xor  edx,edx
  mov  eax,[ecx].TMmVirtSpace.LoMap
@cy2:
  bt   [eax],edx
  jc   @sync_FreeCount // not all free?
  inc  edx
  cmp  dx,[ecx].TMmVirtSpace.PageCount
  jb   @cy2
  //
  // Drop whole space:
  mov  edi,ecx
  // unlink:
  mov  eax,ecx
  call MmRemoveEntryListAndClear
  test eax,eax
  jz   @ok
  // free memory:
  xor  eax,eax
  mov  [edi].TMmVirtSpace.Signature,ax//eax
  xchg [edi].TMmVirtSpace.Base,eax
  //test eax,eax;jz   @skip_free
  push MEM_RELEASE
  push 0
  push eax
  call VirtualFree
  //
  // free space desc
  xor  eax,eax
  mov  edx,edi
  call MmFree
  //
  jmp  @ok
  //
@sync_FreeCount:
  inc  esi
  jmp  @cy3_in
@cy3:
  bt   [eax],edx
  jc   @sync_FreeCount
@cy3_in:
  inc  edx
  cmp  dx,[ecx].TMmVirtSpace.PageCount
  jb   @cy3
  //
  neg  esi
  add  si,[ecx].TMmVirtSpace.PageCount // PageCount-UsedPageCount
  mov  [ecx].TMmVirtSpace.FreePageCount,si
  jmp  @ok
  //
@not_start:
  xor  eax,eax
  jmp  @fail
end;

procedure MmVirtPoolSetInitSpace(Pool: PMmVirtPool; InitPage,DefSpace: Pointer);
var Space: PMmVirtSpace;
begin
  // Use the page, allocated in MmVirtPageAlloc with cMmInitVirtPool... (64k)
  Space:=Pointer(DefSpace);
  Space.LoMap:=Pointer(Longint(Space)+SizeOf(TMmVirtSpace));
  Space.HiMap:=Pointer(Longint(Space.LoMap)+2);
  Space.PageCount:=16;
  Space.FreePageCount:=15;
  Space.Capacity:=$10000;
  SetBit(Space.HiMap,0);
  SetBit(Space.LoMap,0);
  Space.Base:=InitPage;
  Space.Signature:=cMmVirtSpaceSign;
  MmInsertHeadList(@Pool.Spaces,@Space.Entry);
end;

function MmVirtPageAlloc(Arg: Pointer; Size: integer): Pointer;
var Space: Pointer;
    Pool: PMmVirtPool;
begin
  if (Size<=0) then begin
    Result:=nil;
    exit;
  end;
  //
  if (Arg=Pointer(cMmInitVirtPool))
  and (Size=$1000)
  then begin
    // Start a new Pool, will get MmVirtPoolReInit later:
    Space:=VirtualAlloc(nil,$10000,MEM_RESERVE,PAGE_NOACCESS);
    if (Space<>nil) then begin
      Result:=VirtualAlloc(Space,$1000,MEM_COMMIT,PAGE_READWRITE);
      if (Result<>nil) then
        exit;
      VirtualFree(Space,0,MEM_RELEASE);
    end;
  end else
  if (Arg<>nil) then begin
    Size:=(Size+$FFF) and (not $FFF);
    Pool:=Arg;
    if (Pool.Signature=cMmVirtPoolSign) then begin
      EnterCriticalSection(TRtlCriticalSection(Pool.lpLock^));
      try
        Result:=MmVirtInternalPageAlloc(Pool,Size);
      finally
        LeaveCriticalSection(TRtlCriticalSection(Pool.lpLock^));
      end;
      exit;
    end;
  end;
  Result:=nil;
end;

function MmVirtPageFree(Arg: Pointer; Page: Pointer; Size: integer): Boolean;
var Pool: PMmVirtPool;
begin
  //New200830 - wrong release of a page?
  //if (Page=Pointer($00B09000)) then begin
  //  if (Page<>nil) then ;;;
  //end;
  //
  Pool:=Arg;
  if (Pool<>nil)
  and (Pool.Signature=cMmVirtPoolSign)
  and (Longint(Page) and $FFF=0)
  and (Page<>nil)
  then begin
    EnterCriticalSection(TRtlCriticalSection(Pool.lpLock^));
    try
      Result:=MmVirtInternalPageFree(Pool,Page,Size);
    finally
      LeaveCriticalSection(TRtlCriticalSection(Pool.lpLock^));
    end;
  end else
    Result:=False;
end;

procedure MmVirtPoolReInit(Pager: PMmVirtPager; AllocProc: TMmPtrAlloc; AllocArg: Pointer; lpLockRtl: Pointer);
var Pool: PMmVirtPool;
    PageCount,MapSize,DescSize: integer;
    pdw: PLongint;
begin
  if (Cardinal(Pager.Arg)=cMmInitVirtPool) then begin
    DescSize:=SizeOf(TMmVirtPool);
    if (lpLockRtl=nil) then
      inc(DescSize,SizeOf(TRtlCriticalSection));
    Pool:=AllocProc(AllocArg,DescSize);
    Pool.Signature:=cMmVirtPoolSign;
    Pool.MemPool:=AllocArg;
    //x: InitializeCriticalSection(Pool.Lock);
    if (lpLockRtl<>nil) then
      Pool.lpLock:=lpLockRtl
    else begin
      Pool.lpLock:=Pointer(Longint(Pool)+SizeOf(TMmVirtPool));
      InitializeCriticalSection(TRtlCriticalSection(Pool.lpLock^));
    end;
    MmInitializeListHead(@Pool.Spaces);
    Pager.Arg:=Pool;
    MmVirtPoolSetInitSpace(Pool,Pointer(Cardinal(Pool) and $FFFFF000),AllocProc(AllocArg,SizeOf(TMmVirtSpace)+4)); //zmena240217: 2->4 !
    if IsLibrary then
      Pool.DefSpaceSize:=$10000 // 64kb in DLL
    else
      Pool.DefSpaceSize:=$400000; // 4M in EXE
    // Pre-allocate a DefSpaceSize descriptor, now when we have a Space ready...
    PageCount:=Pool.DefSpaceSize shr 12;
    MapSize:=(PageCount+7) shr 3;
    DescSize:=SizeOf(TMmVirtSpace)+2*MapSize;
    pdw:=TMmPtrAlloc(PMmPool(Pool.MemPool).AllocProc)(PMmPool(Pool.MemPool),DescSize);
    if (pdw<>nil) then begin
      pdw^:=DescSize;
      Pool.PrepSpace:=pdw;
    end;
  end;
end;

function MmVirtPoolShutdown(Arg,MemPool: Pointer): Boolean; assembler;
asm
  // if (Pool.MemPool=MemPool) then begin
  cmp  [eax].TMmVirtPool.MemPool,edx
  jne  @fail
  // drop spaces:
  push ebp
  mov  ebp,esp
  //
  push 0
  push 0
  //
  lea  ecx,[eax].TMmVirtPool.Spaces
  mov  eax,ecx
@cy1:
  mov  edx,[eax].TListEntry.FLink
  cmp  [edx].TListEntry.BLink,eax
  jne  @damaged
  cmp  edx,ecx
  je   @break1
  mov  eax,edx
  cmp  [eax].TMmVirtSpace.Signature,cMmVirtSpaceSign
  jne  @damaged
  //
  cmp  [eax].TMmVirtSpace.Base,0
  je   @cy1
  push [eax].TMmVirtSpace.Capacity
  push [eax].TMmVirtSpace.Base
  jmp  @cy1
@break1:
@damaged:
  //
@cy2:
  pop  eax
  pop  edx
  //
  test eax,eax
  jz   @break2
  //
  push MEM_RELEASE
  push 0
  push eax
  push MEM_DECOMMIT
  push edx
  push eax
  call VirtualFree
  call VirtualFree
  //
  jmp  @cy2
  //
@break2:
  mov  eax,1
  mov  esp,ebp
  pop  ebp
  ret
  //
@fail:
  xor  eax,eax
  //ret
end;

procedure MmGetDefVirtPager(VirtPool: PMmVirtPager);
begin
  VirtPool.AllocRawPage:=MmVirtPageAlloc;
  VirtPool.ReleaseRawPage:=MmVirtPageFree;
  VirtPool.ReInit:=MmVirtPoolReInit;
  VirtPool.Shutdown:=MmVirtPoolShutdown;
  VirtPool.Arg:=MmVirtPoolPreCreate();
end;

function MmGetMemoryManager(var Mmgr: TMemoryManager; lpVirtPool: PMmVirtPager): Boolean;
var VirtPool: TMmVirtPager;
begin
  if (g_MainMmPool=nil) then begin
    if (lpVirtPool=nil) then begin
      MmGetDefVirtPager(@VirtPool);
      lpVirtPool:=@VirtPool;
    end;
    g_MainMmPool:=MmCreatePool(lpVirtPool,@g_MainMmPool);
    if (g_MainMmPool=nil) then begin
      Result:=False;
      exit;
    end;
    g_MainMmPool.RefCount:=$40000000; //New171113
  end;
  Mmgr.GetMem:=MmGetMem;
  Mmgr.FreeMem:=MmFreeMem;
  Mmgr.ReallocMem:=MmReallocMem;
  Result:=True;
end;

{x:
procedure MmShutdown_Bailout;
asm
  push eax
  //
  lea  eax,[ebx].TMmPool.MidLock
  push eax
  call LeaveCriticalSection
  //
  pop  eax
  //
  jmp  MmTry_Bailout
end;

//x - note: may not lock the pool arround destroying it !
procedure MmShutdownMainPool;
asm
  mov  eax,g_MainMmPool
  test eax,eax
  jz   @not_initialized
  //
  push ebx
  push ebp
  mov  ebp,esp
  //
  mov  ebx,eax
  xor  ecx,ecx
  mov  edx,offset MmShutdown_Bailout
  call MmPoolMidLock_Try
  //
  mov  eax,ebx
  lea  edx,g_MainMmPool
  call MmDestroyPool
  //
  xor  edx,edx
  mov  esp,fs:[edx]
  pop  dword ptr fs:[edx]
  //
  mov  esp,ebp
  pop  ebp
  pop  ebx
  //
  ret
  //
@not_initialized:
  nop
  //ret
end;{}
procedure MmShutdownMainPool;
asm
  xor  eax,eax
  xchg g_MainMmPool,eax
  test eax,eax
  jnz  MmDestroyPool
end;

function MmCreatePool(VirtPool: PMmVirtPager; Owner: Pointer): PMmPool;
var Pool: PMmPool;
    Page: PMmPage;
    Space: Pointer;
    Typ: PMmType;
    i: integer;
begin
  //New200822 - prevent false-positives on small pages...
  if not bnMmWasInit then begin
    bnMmWasInit:=True;
    VirtualAlloc(Pointer($00AB0000),$10000,MEM_RESERVE,PAGE_NOACCESS);
  end;
  //
  Page:=VirtPool.AllocRawPage(VirtPool.Arg,$1000);
  if (Page=nil) then begin
    Result:=nil;
    exit;
  end;
  // Format middle-size page:
  Space:=MmFormatMidPage(Page,Pointer(Longint(Page)+8));
  Pool:=MmMidPageAlloc(Space,SizeOf(TMmPool));
  PMmMiddlePageTail(Longint(Page)+cMmMidPageTailOffset).Pool:=Pool;
  Pool.Signature:=cMmPoolSign;
  Pool.VirtPool:=VirtPool^; // copy record...
  InitializeCriticalSection(Pool.MidLock);
  Pool.MidPage:=Space;
  Pool.AllocProc:=@MmAllocStartup;
  Pool.Owner:=Owner;
  Pool.SmallTypes[0]:=nil;
  //x200831: Pool.FreePageSched:=@Pool.FreePageSched; //New200822
  MmInitializeListHead(@Pool.FreePageList); //New200831
  MmInitializeListHead(@Pool.SmallPageList); //New200907
  //
  Typ:=MmMidPageAlloc(Pool.MidPage,16*SizeOf(TMmType));
  PLongint(Longint(Typ)-8)^:=-1; // Prevent freeing types...
  //
  if (PLongint(@Div3Table)^=0) then
    InitAlignTables;
  //
  for i:=1 to High(Pool.SmallTypes) do begin
    Pool.SmallTypes[i]:=Typ;
    //
    Typ.Signature:=cMmTypeSign;
    Typ.FreePages:=nil;
    Typ.CheckAligned:=GetAlignCheckCode(i);
    Typ.Pool:=Pool;
    Typ.ItemSize:=i;
    //x200831: Typ.DefFreeList:=i;//(i shl 4);
    //x200831: Typ.DefFreeCount:=(4096-(i shl 4)) div (i shl 4);
    //New200831:
    if (i>1) then begin
      Typ.DefFreeList:=i; // probably unused?
      Typ.DefFreeCount:=(4096-(i shl 4)) div (i shl 4);
    end else begin
      Typ.DefFreeList:=2;
      Typ.DefFreeCount:=(4096-32) div 16;
    end;
    //
    Typ.MaxFreeCount:=Typ.DefFreeCount;
    //
    inc(Longint(Typ),SizeOf(TMmType));
  end;
  // Initialize Pool.MidTypes:
  MmInitMidTypes(Pool);
  //
  // Re-initialize VirtPool (re-allocate from the new pool)
  if Assigned(Pool.VirtPool.ReInit) then begin
    Pool.VirtPool.ReInit(@Pool.VirtPool,Pool.AllocProc,Pool,@Pool.MidLock);
  end;
  //
  // Prevent freeing the pool:
  PLongint(Longint(Pool)-8)^:=-1;
  //
  //New171113
  Pool.Vmt:=Mm_IMemPool_Vmt();
  Pool.RefCount:=1;
  //New200825
  Pool.Vmt_IMalloc:=Mm_IMalloc_Vmt();
  //
  Result:=Pool;
end;

function MmDestroyPool(Pool: PMmPool; Owner: Pointer): Boolean;
begin
  if (Pool.Signature<>cMmPoolSign)
  or (Pool.Owner<>Owner)
  then begin
    Result:=False;
    exit;
  end;
  //xx: Note: no pointer to small pages, that are completelly full... will not release them...
  //xx:       can release only those small pages, that contain some free items and are linked in Typ.FreePages ...
  //
  // Just drop all acquired pages in VirtPool:
  DeleteCriticalSection(Pool.MidLock);
  if Assigned(Pool.VirtPool.Shutdown) then
    Result:=Pool.VirtPool.Shutdown(Pool.VirtPool.Arg,Pool)
  else
    Result:=False;
end;

{$ifdef DSMINIMM_END} {$warnings off}end.{$endif DSMINIMM_END}
{$endif DSMINIMM_INC}{$endif IMPLEMENTATION}
